@@ -27,10 +27,6 @@
 #include <ctype.h>
 #include <errno.h>
 
-#ifdef HAVE_LIBSTOKEN
-#include <stoken.h>
-#endif
-
 #ifdef HAVE_LIBOATH
 #include <liboath/oath.h>
 #endif
@@ -41,65 +37,56 @@
 #include "openconnect-internal.h"
 
 static int xmlpost_append_form_opts(struct openconnect_info *vpninfo,
-				    struct oc_auth_form *form, char *body, int bodylen);
+				    struct oc_auth_form *form, struct oc_text_buf *body);
 static int can_gen_tokencode(struct openconnect_info *vpninfo,
 			     struct oc_auth_form *form, struct oc_form_opt *opt);
 static int do_gen_tokencode(struct openconnect_info *vpninfo, struct oc_auth_form *form);
 
-static int append_opt(char *body, int bodylen, char *opt, char *name)
+int openconnect_set_option_value(struct oc_form_opt *opt, const char *value)
 {
-	int len = strlen(body);
+	if (opt->type == OC_FORM_OPT_SELECT) {
+		struct oc_form_opt_select *sopt = (void *)opt;
+		int i;
 
-	if (len) {
-		if (len >= bodylen - 1)
-			return -ENOSPC;
-		body[len++] = '&';
-	}
-
-	while (*opt) {
-		if (isalnum((int)(unsigned char)*opt)) {
-			if (len >= bodylen - 1)
-				return -ENOSPC;
-			body[len++] = *opt;
-		} else {
-			if (len >= bodylen - 3)
-				return -ENOSPC;
-			sprintf(body+len, "%%%02x", *opt);
-			len += 3;
+		for (i=0; i<sopt->nr_choices; i++) {
+			if (value == sopt->choices[i]->name) {
+				opt->_value = sopt->choices[i]->name;
+				return 0;
+			}
 		}
-		opt++;
+		return -EINVAL;
 	}
 
-	if (len >= bodylen - 1)
-		return -ENOSPC;
-	body[len++] = '=';
+	opt->_value = strdup(value);
+	if (!opt->_value)
+		return -ENOMEM;
 
-	while (name && *name) {
-		if (isalnum((int)(unsigned char)*name)) {
-			if (len >= bodylen - 1)
-				return -ENOSPC;
-			body[len++] = *name;
-		} else {
-			if (len >= bodylen - 3)
-				return -ENOSPC;
-			sprintf(body+len, "%%%02X", *name);
-			len += 3;
-		}
-		name++;
-	}
-	body[len] = 0;
+	return 0;
+}
+
+static int append_opt(struct oc_text_buf *body, char *opt, char *name)
+{
+	if (buf_error(body))
+		return buf_error(body);
+
+	if (body->pos)
+		buf_append(body, "&");
+
+	buf_append_urlencoded(body, opt);
+	buf_append(body, "=");
+	buf_append_urlencoded(body, name);
 
 	return 0;
 }
 
 static int append_form_opts(struct openconnect_info *vpninfo,
-			    struct oc_auth_form *form, char *body, int bodylen)
+			    struct oc_auth_form *form, struct oc_text_buf *body)
 {
 	struct oc_form_opt *opt;
 	int ret;
 
 	for (opt = form->opts; opt; opt = opt->next) {
-		ret = append_opt(body, bodylen, opt->name, opt->value);
+		ret = append_opt(body, opt->name, opt->_value);
 		if (ret)
 			return ret;
 	}
@@ -110,7 +97,7 @@ static void free_opt(struct oc_form_opt *opt)
 {
 	/* for SELECT options, opt->value is a pointer to oc_choice->name */
 	if (opt->type != OC_FORM_OPT_SELECT)
-		free(opt->value);
+		free(opt->_value);
 	else {
 		struct oc_form_opt_select *sel = (void *)opt;
 		int i;
@@ -282,7 +269,7 @@ static int parse_form(struct openconnect_info *vpninfo, struct oc_auth_form *for
 
 		if (!strcmp(input_type, "hidden")) {
 			opt->type = OC_FORM_OPT_HIDDEN;
-			opt->value = (char *)xmlGetProp(xml_node, (unsigned char *)"value");
+			opt->_value = (char *)xmlGetProp(xml_node, (unsigned char *)"value");
 		} else if (!strcmp(input_type, "text")) {
 			opt->type = OC_FORM_OPT_TEXT;
 		} else if (!strcmp(input_type, "password")) {
@@ -356,7 +343,7 @@ static char *xmlnode_msg(xmlNode *xml_node)
 		if (params[nr_params]) {
 			paramlen = strlen(params[nr_params]);
 			/* Move rest of fmt string up... */
-			memmove(pct - 1 + paramlen, pct + 2, strlen(pct) - 1);
+			memmove(pct + paramlen, pct + 2, strlen(pct + 2) + 1);
 			/* ... and put the string parameter in where the '%s' was */
 			memcpy(pct, params[nr_params], paramlen);
 			pct += paramlen;
@@ -708,13 +695,13 @@ int parse_xml_response(struct openconnect_info *vpninfo, char *response, struct
 	return ret;
 }
 
-static void nuke_opt_values(struct oc_form_opt *opt)
+void nuke_opt_values(struct oc_form_opt *opt)
 {
 	for (; opt; opt = opt->next) {
 		if (opt->type == OC_FORM_OPT_TEXT ||
 		    opt->type == OC_FORM_OPT_PASSWORD) {
-			free(opt->value);
-			opt->value = NULL;
+			free(opt->_value);
+			opt->_value = NULL;
 		}
 	}
 }
@@ -757,8 +744,8 @@ retry:
 			opt->flags |= OC_FORM_OPT_IGNORE;
 		else if (!strcmp(opt->name, "secondary_username") && second_auth) {
 			if (auth_choice->secondary_username) {
-				free(opt->value);
-				opt->value = strdup(auth_choice->secondary_username);
+				free(opt->_value);
+				opt->_value = strdup(auth_choice->secondary_username);
 			}
 			if (!auth_choice->secondary_username_editable)
 				opt->flags |= OC_FORM_OPT_IGNORE;
@@ -769,9 +756,9 @@ retry:
 
 	if (ret == OC_FORM_RESULT_NEWGROUP &&
 	    form->authgroup_opt &&
-	    form->authgroup_opt->form.value) {
+	    form->authgroup_opt->form._value) {
 		free(vpninfo->authgroup);
-		vpninfo->authgroup = strdup(form->authgroup_opt->form.value);
+		vpninfo->authgroup = strdup(form->authgroup_opt->form._value);
 
 		if (!vpninfo->xmlpost)
 			goto retry;
@@ -790,7 +777,7 @@ retry:
  *  = OC_FORM_RESULT_LOGGEDIN, when form indicates that login was already successful
  */
 int handle_auth_form(struct openconnect_info *vpninfo, struct oc_auth_form *form,
-		     char *request_body, int req_len, const char **method,
+		     struct oc_text_buf *request_body, const char **method,
 		     const char **request_body_type)
 {
 	int ret;
@@ -831,12 +818,15 @@ int handle_auth_form(struct openconnect_info *vpninfo, struct oc_auth_form *form
 
 	/* tokencode generation is deferred until after username prompts and CSD */
 	ret = do_gen_tokencode(vpninfo, form);
-	if (ret)
+	if (ret) {
+		vpn_progress(vpninfo, PRG_ERR, _("Failed to generate OTP tokencode; disabling token\n"));
+		vpninfo->token_bypassed = 1;
 		return ret;
+	}
 
 	ret = vpninfo->xmlpost ?
-	      xmlpost_append_form_opts(vpninfo, form, request_body, req_len) :
-	      append_form_opts(vpninfo, form, request_body, req_len);
+	      xmlpost_append_form_opts(vpninfo, form, request_body) :
+	      append_form_opts(vpninfo, form, request_body);
 	if (!ret) {
 		*method = "POST";
 		*request_body_type = "application/x-www-form-urlencoded";
@@ -932,7 +922,7 @@ bad:
 	return NULL;
 }
 
-static int xmlpost_complete(xmlDocPtr doc, char *body, int bodylen)
+static int xmlpost_complete(xmlDocPtr doc, struct oc_text_buf *body)
 {
 	xmlChar *mem = NULL;
 	int len, ret = 0;
@@ -948,12 +938,7 @@ static int xmlpost_complete(xmlDocPtr doc, char *body, int bodylen)
 		return -ENOMEM;
 	}
 
-	if (len > bodylen)
-		ret = -E2BIG;
-	else {
-		memcpy(body, mem, len);
-		body[len] = 0;
-	}
+	buf_append_bytes(body, mem, len);
 
 	xmlFreeDoc(doc);
 	xmlFree(mem);
@@ -961,7 +946,7 @@ static int xmlpost_complete(xmlDocPtr doc, char *body, int bodylen)
 	return ret;
 }
 
-int xmlpost_initial_req(struct openconnect_info *vpninfo, char *request_body, int req_len, int cert_fail)
+int xmlpost_initial_req(struct openconnect_info *vpninfo, struct oc_text_buf *request_body, int cert_fail)
 {
 	xmlNodePtr root, node;
 	xmlDocPtr doc = xmlpost_new_query(vpninfo, "init", &root);
@@ -992,15 +977,15 @@ int xmlpost_initial_req(struct openconnect_info *vpninfo, char *request_body, in
 		if (!node)
 			goto bad;
 	}
-	return xmlpost_complete(doc, request_body, req_len);
+	return xmlpost_complete(doc, request_body);
 
 bad:
-	xmlpost_complete(doc, NULL, 0);
+	xmlpost_complete(doc, NULL);
 	return -ENOMEM;
 }
 
 static int xmlpost_append_form_opts(struct openconnect_info *vpninfo,
-				    struct oc_auth_form *form, char *body, int bodylen)
+				    struct oc_auth_form *form, struct oc_text_buf *body)
 {
 	xmlNodePtr root, node;
 	xmlDocPtr doc = xmlpost_new_query(vpninfo, "auth-reply", &root);
@@ -1024,7 +1009,7 @@ static int xmlpost_append_form_opts(struct openconnect_info *vpninfo,
 	for (opt = form->opts; opt; opt = opt->next) {
 		/* group_list: create a new <group-select> node under <config-auth> */
 		if (!strcmp(opt->name, "group_list")) {
-			if (!xmlNewTextChild(root, NULL, XCAST("group-select"), XCAST(opt->value)))
+			if (!xmlNewTextChild(root, NULL, XCAST("group-select"), XCAST(opt->_value)))
 				goto bad;
 			continue;
 		}
@@ -1033,7 +1018,7 @@ static int xmlpost_append_form_opts(struct openconnect_info *vpninfo,
 		if (!strcmp(opt->name, "answer") ||
 		    !strcmp(opt->name, "whichpin") ||
 		    !strcmp(opt->name, "new_password")) {
-			if (!xmlNewTextChild(node, NULL, XCAST("password"), XCAST(opt->value)))
+			if (!xmlNewTextChild(node, NULL, XCAST("password"), XCAST(opt->_value)))
 				goto bad;
 			continue;
 		}
@@ -1045,7 +1030,7 @@ static int xmlpost_append_form_opts(struct openconnect_info *vpninfo,
 		}
 
 		/* everything else: create <foo>user_input</foo> under <auth> */
-		if (!xmlNewTextChild(node, NULL, XCAST(opt->name), XCAST(opt->value)))
+		if (!xmlNewTextChild(node, NULL, XCAST(opt->name), XCAST(opt->_value)))
 			goto bad;
 	}
 
@@ -1053,236 +1038,13 @@ static int xmlpost_append_form_opts(struct openconnect_info *vpninfo,
 	    !xmlNewTextChild(root, NULL, XCAST("host-scan-token"), XCAST(vpninfo->csd_token)))
 		goto bad;
 
-	return xmlpost_complete(doc, body, bodylen);
+	return xmlpost_complete(doc, body);
 
 bad:
-	xmlpost_complete(doc, NULL, 0);
+	xmlpost_complete(doc, NULL);
 	return -ENOMEM;
 }
 
-/*
- * If the user clicks OK without entering any data, we will continue
- * connecting but bypass soft token generation for the duration of
- * this "obtain_cookie" session.
- *
- * If the user clicks Cancel, we will abort the connection.
- *
- * Return value:
- *  < 0, on error
- *  = 0, on success (or if the user bypassed soft token init)
- *  = 1, if the user cancelled the form submission
- */
-int prepare_stoken(struct openconnect_info *vpninfo)
-{
-#ifdef HAVE_LIBSTOKEN
-	struct oc_auth_form form;
-	struct oc_form_opt opts[3], *opt = opts;
-	char **devid = NULL, **pass = NULL, **pin = NULL;
-	int ret = 0;
-
-	memset(&form, 0, sizeof(form));
-	memset(&opts, 0, sizeof(opts));
-
-	form.opts = opts;
-	form.message = _("Enter credentials to unlock software token.");
-
-	vpninfo->token_tries = 0;
-	vpninfo->token_bypassed = 0;
-
-	if (stoken_devid_required(vpninfo->stoken_ctx)) {
-		opt->type = OC_FORM_OPT_TEXT;
-		opt->name = (char *)"devid";
-		opt->label = _("Device ID:");
-		devid = &opt->value;
-		opt++;
-	}
-	if (stoken_pass_required(vpninfo->stoken_ctx)) {
-		opt->type = OC_FORM_OPT_PASSWORD;
-		opt->name = (char *)"password";
-		opt->label = _("Password:");
-		pass = &opt->value;
-		opt++;
-	}
-	if (stoken_pin_required(vpninfo->stoken_ctx)) {
-		opt->type = OC_FORM_OPT_PASSWORD;
-		opt->name = (char *)"password";
-		opt->label = _("PIN:");
-		opt->flags = OC_FORM_OPT_NUMERIC;
-		pin = &opt->value;
-		opt++;
-	}
-
-	opts[0].next = opts[1].type ? &opts[1] : NULL;
-	opts[1].next = opts[2].type ? &opts[2] : NULL;
-
-	while (1) {
-		nuke_opt_values(opts);
-
-		if (!opts[0].type) {
-			/* don't bug the user if there's nothing to enter */
-			ret = 0;
-		} else {
-			int some_empty = 0, all_empty = 1;
-
-			/* < 0 for error; 1 if cancelled */
-			ret = process_auth_form(vpninfo, &form);
-			if (ret)
-				break;
-
-			for (opt = opts; opt; opt = opt->next) {
-				if (!opt->value || !strlen(opt->value))
-					some_empty = 1;
-				else
-					all_empty = 0;
-			}
-			if (all_empty) {
-				vpn_progress(vpninfo, PRG_INFO,
-					     _("User bypassed soft token.\n"));
-				vpninfo->token_bypassed = 1;
-				ret = 0;
-				break;
-			}
-			if (some_empty) {
-				vpn_progress(vpninfo, PRG_INFO,
-					     _("All fields are required; try again.\n"));
-				continue;
-			}
-		}
-
-		ret = stoken_decrypt_seed(vpninfo->stoken_ctx,
-					  pass ? *pass : NULL,
-					  devid ? *devid : NULL);
-		if (ret == -EIO || (ret && !devid && !pass)) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("General failure in libstoken.\n"));
-			break;
-		} else if (ret != 0) {
-			vpn_progress(vpninfo, PRG_INFO,
-				     _("Incorrect device ID or password; try again.\n"));
-			continue;
-		}
-
-		if (pin) {
-			if (stoken_check_pin(vpninfo->stoken_ctx, *pin) != 0) {
-				vpn_progress(vpninfo, PRG_INFO,
-					     _("Invalid PIN format; try again.\n"));
-				continue;
-			}
-			free(vpninfo->stoken_pin);
-			vpninfo->stoken_pin = strdup(*pin);
-			if (!vpninfo->stoken_pin) {
-				ret = -ENOMEM;
-				break;
-			}
-		}
-		vpn_progress(vpninfo, PRG_DEBUG, _("Soft token init was successful.\n"));
-		ret = 0;
-		break;
-	}
-
-	nuke_opt_values(opts);
-	return ret;
-#else
-	return -EOPNOTSUPP;
-#endif
-}
-
-/* Return value:
- *  < 0, if unable to generate a tokencode
- *  = 0, on success
- */
-static int can_gen_stoken_code(struct openconnect_info *vpninfo,
-			       struct oc_auth_form *form,
-			       struct oc_form_opt *opt)
-{
-#ifdef HAVE_LIBSTOKEN
-	if ((strcmp(opt->name, "password") && strcmp(opt->name, "answer")) ||
-	    vpninfo->token_bypassed)
-		return -EINVAL;
-	if (vpninfo->token_tries == 0) {
-		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("OK to generate INITIAL tokencode\n"));
-		vpninfo->token_time = 0;
-	} else if (vpninfo->token_tries == 1 && form->message &&
-		   strcasestr(form->message, "next tokencode")) {
-		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("OK to generate NEXT tokencode\n"));
-		vpninfo->token_time += 60;
-	} else {
-		/* limit the number of retries, to avoid account lockouts */
-		vpn_progress(vpninfo, PRG_INFO,
-			     _("Server is rejecting the soft token; switching to manual entry\n"));
-		return -ENOENT;
-	}
-	return 0;
-#else
-	return -EOPNOTSUPP;
-#endif
-}
-
-/* Return value:
- *  < 0, if unable to generate a tokencode
- *  = 0, on success
- */
-static int can_gen_totp_code(struct openconnect_info *vpninfo,
-			     struct oc_auth_form *form,
-			     struct oc_form_opt *opt)
-{
-#ifdef HAVE_LIBOATH
-	if ((strcmp(opt->name, "secondary_password") != 0) ||
-	    vpninfo->token_bypassed)
-		return -EINVAL;
-	if (vpninfo->token_tries == 0) {
-		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("OK to generate INITIAL tokencode\n"));
-		vpninfo->token_time = 0;
-	} else if (vpninfo->token_tries == 1) {
-		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("OK to generate NEXT tokencode\n"));
-		vpninfo->token_time += OATH_TOTP_DEFAULT_TIME_STEP_SIZE;
-	} else {
-		/* limit the number of retries, to avoid account lockouts */
-		vpn_progress(vpninfo, PRG_INFO,
-			     _("Server is rejecting the soft token; switching to manual entry\n"));
-		return -ENOENT;
-	}
-	return 0;
-#else
-	return -EOPNOTSUPP;
-#endif
-}
-
-/* Return value:
- *  < 0, if unable to generate a tokencode
- *  = 0, on success
- */
-static int can_gen_hotp_code(struct openconnect_info *vpninfo,
-			     struct oc_auth_form *form,
-			     struct oc_form_opt *opt)
-{
-#ifdef HAVE_LIBOATH
-	if ((strcmp(opt->name, "secondary_password") != 0) ||
-	    vpninfo->token_bypassed)
-		return -EINVAL;
-	if (vpninfo->token_tries == 0) {
-		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("OK to generate INITIAL tokencode\n"));
-	} else if (vpninfo->token_tries == 1) {
-		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("OK to generate NEXT tokencode\n"));
-		vpninfo->token_time++;
-	} else {
-		/* limit the number of retries, to avoid account lockouts */
-		vpn_progress(vpninfo, PRG_INFO,
-			     _("Server is rejecting the soft token; switching to manual entry\n"));
-		return -ENOENT;
-	}
-	return 0;
-#else
-	return -EOPNOTSUPP;
-#endif
-}
-
 /* Return value:
  *  < 0, if unable to generate a tokencode
  *  = 0, on success
@@ -1292,110 +1054,26 @@ static int can_gen_tokencode(struct openconnect_info *vpninfo,
 			     struct oc_form_opt *opt)
 {
 	switch (vpninfo->token_mode) {
+#ifdef HAVE_LIBSTOKEN
 	case OC_TOKEN_MODE_STOKEN:
 		return can_gen_stoken_code(vpninfo, form, opt);
-
+#endif
+#ifdef HAVE_LIBOATH
 	case OC_TOKEN_MODE_TOTP:
 		return can_gen_totp_code(vpninfo, form, opt);
 
 	case OC_TOKEN_MODE_HOTP:
 		return can_gen_hotp_code(vpninfo, form, opt);
-
+#endif
+#ifdef HAVE_LIBPCSCLITE
+	case OC_TOKEN_MODE_YUBIOATH:
+		return can_gen_yubikey_code(vpninfo, form, opt);
+#endif
 	default:
 		return -EINVAL;
 	}
 }
 
-static int do_gen_stoken_code(struct openconnect_info *vpninfo,
-			      struct oc_auth_form *form,
-			      struct oc_form_opt *opt)
-{
-#ifdef HAVE_LIBSTOKEN
-	char tokencode[STOKEN_MAX_TOKENCODE + 1];
-
-	if (!vpninfo->token_time)
-		vpninfo->token_time = time(NULL);
-	vpn_progress(vpninfo, PRG_INFO, _("Generating RSA token code\n"));
-
-	/* This doesn't normally fail */
-	if (stoken_compute_tokencode(vpninfo->stoken_ctx, vpninfo->token_time,
-				     vpninfo->stoken_pin, tokencode) < 0) {
-		vpn_progress(vpninfo, PRG_ERR, _("General failure in libstoken.\n"));
-		return -EIO;
-	}
-
-	vpninfo->token_tries++;
-	opt->value = strdup(tokencode);
-	return opt->value ? 0 : -ENOMEM;
-#else
-	return 0;
-#endif
-}
-
-static int do_gen_totp_code(struct openconnect_info *vpninfo,
-			    struct oc_auth_form *form,
-			    struct oc_form_opt *opt)
-{
-#ifdef HAVE_LIBOATH
-	int oath_err;
-	char tokencode[7];
-
-	if (!vpninfo->token_time)
-		vpninfo->token_time = time(NULL);
-
-	vpn_progress(vpninfo, PRG_INFO, _("Generating OATH TOTP token code\n"));
-
-	oath_err = oath_totp_generate(vpninfo->oath_secret,
-				      vpninfo->oath_secret_len,
-				      vpninfo->token_time,
-				      OATH_TOTP_DEFAULT_TIME_STEP_SIZE,
-				      OATH_TOTP_DEFAULT_START_TIME,
-				      6, tokencode);
-	if (oath_err != OATH_OK) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Unable to generate OATH TOTP token code: %s\n"),
-			     oath_strerror(oath_err));
-		return -EIO;
-	}
-
-	vpninfo->token_tries++;
-	opt->value = strdup(tokencode);
-	return opt->value ? 0 : -ENOMEM;
-#else
-	return 0;
-#endif
-}
-
-static int do_gen_hotp_code(struct openconnect_info *vpninfo,
-			    struct oc_auth_form *form,
-			    struct oc_form_opt *opt)
-{
-#ifdef HAVE_LIBOATH
-	int oath_err;
-	char tokencode[7];
-
-	vpn_progress(vpninfo, PRG_INFO, _("Generating OATH HOTP token code\n"));
-
-	oath_err = oath_hotp_generate(vpninfo->oath_secret,
-				      vpninfo->oath_secret_len,
-				      vpninfo->token_time,
-				      6, false, OATH_HOTP_DYNAMIC_TRUNCATION,
-				      tokencode);
-	if (oath_err != OATH_OK) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Unable to generate OATH HOTP token code: %s\n"),
-			     oath_strerror(oath_err));
-		return -EIO;
-	}
-
-	vpninfo->token_tries++;
-	opt->value = strdup(tokencode);
-	return opt->value ? 0 : -ENOMEM;
-#else
-	return 0;
-#endif
-}
-
 /* Return value:
  *  < 0, if unable to generate a tokencode
  *  = 0, on success
@@ -1414,15 +1092,21 @@ static int do_gen_tokencode(struct openconnect_info *vpninfo,
 	}
 
 	switch (vpninfo->token_mode) {
+#ifdef HAVE_LIBSTOKEN
 	case OC_TOKEN_MODE_STOKEN:
 		return do_gen_stoken_code(vpninfo, form, opt);
-
+#endif
+#ifdef HAVE_LIBOATH
 	case OC_TOKEN_MODE_TOTP:
 		return do_gen_totp_code(vpninfo, form, opt);
 
 	case OC_TOKEN_MODE_HOTP:
 		return do_gen_hotp_code(vpninfo, form, opt);
-
+#endif
+#ifdef HAVE_LIBPCSCLITE
+	case OC_TOKEN_MODE_YUBIOATH:
+		return do_gen_yubikey_code(vpninfo, form, opt);
+#endif
 	default:
 		return -EINVAL;
 	}
