@@ -42,9 +42,8 @@ time_t openconnect__time(time_t *t)
 }
 #endif
 
-#ifndef HAVE_ASPRINTF
-
-static int oc_vasprintf(char **strp, const char *fmt, va_list ap)
+#ifndef HAVE_VASPRINTF
+int openconnect__vasprintf(char **strp, const char *fmt, va_list ap)
 {
 	va_list ap2;
 	char *res = NULL;
@@ -102,14 +101,16 @@ static int oc_vasprintf(char **strp, const char *fmt, va_list ap)
 	*strp = res;
 	return ret;
 }
+#endif
 
+#ifndef HAVE_ASPRINTF
 int openconnect__asprintf(char **strp, const char *fmt, ...)
 {
 	va_list ap;
 	int ret;
 
 	va_start(ap, fmt);
-	ret = oc_vasprintf(strp, fmt, ap);
+	ret = vasprintf(strp, fmt, ap);
 	va_end(ap);
 	return ret;
 }
@@ -182,61 +183,99 @@ char *openconnect__strndup(const char *s, size_t n)
 }
 #endif
 
-#ifndef HAVE_SETENV
-int openconnect__setenv(const char *name, const char *value, int overwrite)
-{
-	char *buf;
-
-	if (!value) {
-		openconnect__unsetenv(name);
-		return 0;
-	}
-	/* Windows putenv() takes a copy of the string */
-	buf = alloca(strlen(name) + strlen(value) + 2);
-
-	sprintf(buf, "%s=%s", name, value);
-	putenv(buf);
-	return 0;
-}
-#endif
-
-#ifndef HAVE_UNSETENV
-void openconnect__unsetenv(const char *name)
-{
-	char *buf = alloca(strlen(name) + 2);
-
-	sprintf(buf, "%s=", name);
-	putenv(buf);
-}
-#endif
-
 #ifndef HAVE_INET_ATON
 int openconnect__inet_aton(const char *cp, struct in_addr *addr)
 {
-  addr->s_addr = inet_addr(cp);
-#if INADDR_NONE == 0xffffffff
-  if (addr->s_addr != 0xffffffff)
-	  return 0;
-  /* Is it an error, or was it really 255.255.255.255? */
-  if (!strcmp(cp, "255.255.255.255"))
-	  return 0;
-#else
-#error What is your INADDR_NONE?
-#endif
-  return 1;
+	return inet_pton(AF_INET, cp, addr);
 }
 #endif
 
 #ifdef _WIN32
-void openconnect__win32_sock_init()
+char *openconnect__win32_strerror(DWORD err)
+{
+	wchar_t *msgw;
+	char *msgutf8;
+	int nr_chars;
+
+	if (!FormatMessageW(FORMAT_MESSAGE_FROM_SYSTEM |
+			    FORMAT_MESSAGE_IGNORE_INSERTS |
+			    FORMAT_MESSAGE_ALLOCATE_BUFFER,
+			    NULL, err,
+			    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
+			    (LPWSTR)&msgw, 0, NULL)) {
+		if (asprintf(&msgutf8, _("(error 0x%x)"), err) != -1)
+			return msgutf8;
+	fail:
+		return strdup(_("(Error while describing error!)"));
+	}
+	nr_chars = wcslen(msgw);
+	if (nr_chars && msgw[nr_chars - 1] == 10)
+		msgw[--nr_chars] = 0;
+	if (nr_chars && msgw[nr_chars - 1] == 13)
+		msgw[--nr_chars] = 0;
+
+	nr_chars = WideCharToMultiByte(CP_UTF8, 0, msgw, -1, NULL, 0, NULL, NULL);
+
+	msgutf8 = malloc(nr_chars);
+	if (!msgutf8)
+		goto fail;
+
+	WideCharToMultiByte(CP_UTF8, 0, msgw, -1, msgutf8, nr_chars, NULL, NULL);
+	LocalFree(msgw);
+	return msgutf8;
+}
+
+int openconnect__win32_sock_init()
 {
 	WSADATA data;
 	if (WSAStartup (MAKEWORD(1, 1), &data) != 0) {
 		fprintf(stderr, _("ERROR: Cannot initialize sockets\n"));
-		exit(1);
+		return -EIO;
 	}
+	return 0;
 }
 
+int openconnect__win32_inet_pton(int af, const char *src, void *dst)
+{
+	union {
+		struct sockaddr_in s4;
+		struct sockaddr_in6 s6;
+	} sa;
+	int salen = sizeof(sa);
+
+	if (af != AF_INET && af != AF_INET6) {
+		errno = EAFNOSUPPORT;
+		return -1;
+	}
+
+	memset(&sa, 0, sizeof(sa));
+	sa.s4.sin_family = af;
+
+	if (WSAStringToAddressA((char *)src, af, NULL, (void *)&sa, &salen))
+		return 0;
+
+	/* For Legacy IP we need to filter out a lot of crap that
+	 * inet_aton() (and WSAStringToAddress()) will support, but
+	 * which inet_pton() should not. Not to mention the fact that
+	 * Wine's implementation will even succeed for strings like
+	 * "2001::1" (http://bugs.winehq.org/show_bug.cgi?id=36991) */
+	if (af == AF_INET) {
+		char canon[16];
+		unsigned char *a = (unsigned char *)&sa.s4.sin_addr;
+
+		snprintf(canon, sizeof(canon), "%d.%d.%d.%d",
+			 a[0], a[1], a[2], a[3]);
+
+		if (strcmp(canon, src))
+			return 0;
+
+		memcpy(dst, &sa.s4.sin_addr, sizeof(sa.s4.sin_addr));
+		return 1;
+	} else {
+		memcpy(dst, &sa.s6.sin6_addr, sizeof(sa.s6.sin6_addr));
+		return 1;
+	}
+}
 
 /* https://github.com/ncm/selectable-socketpair
 
@@ -297,13 +336,13 @@ WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  *   sockets must be closed with closesocket() regardless.
  */
 
-int dumb_socketpair(int socks[2], int make_overlapped)
+OPENCONNECT_CMD_SOCKET dumb_socketpair(OPENCONNECT_CMD_SOCKET socks[2], int make_overlapped)
 {
     union {
        struct sockaddr_in inaddr;
        struct sockaddr addr;
     } a;
-    SOCKET listener;
+    OPENCONNECT_CMD_SOCKET listener;
     int e;
     socklen_t addrlen = sizeof(a.inaddr);
     DWORD flags = (make_overlapped ? WSA_FLAG_OVERLAPPED : 0);
