@@ -322,6 +322,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 				if (dtlsmtu > mtu)
 					mtu = dtlsmtu;
 			} else if (!strcmp(buf + 7, "Session-ID")) {
+				int dtls_sessid_changed = 0;
+
 				if (strlen(colon) != 64) {
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("X-DTLS-Session-ID not 64 characters; is: \"%s\"\n"),
@@ -329,9 +331,17 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 					vpninfo->dtls_attempt_period = 0;
 					return -EINVAL;
 				}
-				for (i = 0; i < 64; i += 2)
-					vpninfo->dtls_session_id[i/2] = unhex(colon + i);
+				for (i = 0; i < 64; i += 2) {
+					unsigned char c = unhex(colon + i);
+					if (vpninfo->dtls_session_id[i/2] != c) {
+						vpninfo->dtls_session_id[i/2] = c;
+						dtls_sessid_changed = 1;
+					}
+				}
 				sessid_found = 1;
+
+				if (dtls_sessid_changed && vpninfo->dtls_state > DTLS_SLEEPING)
+					vpninfo->dtls_need_reconnect = 1;
 			}
 			continue;
 		}
@@ -368,6 +378,9 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			int cstpmtu = atol(colon);
 			if (cstpmtu > mtu)
 				mtu = cstpmtu;
+		} else if (!strcmp(buf + 7, "DynDNS")) {
+			if (!strcmp(colon, "true"))
+				vpninfo->is_dyndns = 1;
 		} else if (!strcmp(buf + 7, "Address-IP6")) {
 			vpninfo->ip_info.netmask6 = new_option->value;
 		} else if (!strcmp(buf + 7, "Address")) {
@@ -440,6 +453,12 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			     _("No IP address received. Aborting\n"));
 		return -EINVAL;
 	}
+	if (mtu < 1280 &&
+	    (vpninfo->ip_info.addr6 || vpninfo->ip_info.netmask6)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("IPv6 configuration received but MTU %d is too small.\n"),
+			     mtu);
+	}
 	if (old_addr) {
 		if (strcmp(old_addr, vpninfo->ip_info.addr)) {
 			vpn_progress(vpninfo, PRG_ERR,
@@ -489,7 +508,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	}
 	vpn_progress(vpninfo, PRG_INFO, _("CSTP connected. DPD %d, Keepalive %d\n"),
 		     vpninfo->ssl_times.dpd, vpninfo->ssl_times.keepalive);
-
+	vpn_progress(vpninfo, PRG_DEBUG, _("CSTP Ciphersuite: %s\n"),
+		     openconnect_get_cstp_cipher(vpninfo));
 
 	monitor_fd_new(vpninfo, ssl);
 
@@ -921,9 +941,9 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	do_dtls_reconnect:
 		/* succeeded, let's rekey DTLS, if it is not rekeying
 		 * itself. */
-		if (vpninfo->dtls_state != DTLS_DISABLED &&
+		if (vpninfo->dtls_state > DTLS_SLEEPING &&
 		    vpninfo->dtls_times.rekey_method == REKEY_NONE) {
-			dtls_reconnect(vpninfo);
+			vpninfo->dtls_need_reconnect = 1;
 		}
 
 		return 1;
