@@ -316,6 +316,7 @@ void dtls_shutdown(struct openconnect_info *vpninfo)
 
 #elif defined(DTLS_GNUTLS)
 #include <gnutls/dtls.h>
+#include "gnutls.h"
 
 struct {
 	const char *name;
@@ -325,16 +326,16 @@ struct {
 	const char *prio;
 } gnutls_dtls_ciphers[] = {
 	{ "AES128-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION" },
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:%COMPAT" },
 	{ "AES256-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_256_CBC, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION" },
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+RSA:%COMPAT" },
 	{ "DES-CBC3-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_3DES_CBC, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+3DES-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION" },
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+3DES-CBC:+SHA1:+RSA:%COMPAT" },
 #if GNUTLS_VERSION_NUMBER >= 0x030207 /* if DTLS 1.2 is supported (and a bug in gnutls is solved) */
 	{ "OC-DTLS1_2-AES128-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_128_GCM, GNUTLS_MAC_AEAD,
-	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-128-GCM:+AEAD:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION:+SIGN-ALL" },
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-128-GCM:+AEAD:+RSA:%COMPAT:+SIGN-ALL" },
 	{ "OC-DTLS1_2-AES256-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_256_GCM, GNUTLS_MAC_AEAD,
-	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-256-GCM:+AEAD:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION:+SIGN-ALL" },
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-256-GCM:+AEAD:+RSA:%COMPAT:+SIGN-ALL" },
 #endif
 };
 
@@ -400,6 +401,7 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 int dtls_try_handshake(struct openconnect_info *vpninfo)
 {
 	int err = gnutls_handshake(vpninfo->dtls_ssl);
+	char *str;
 
 	if (!err) {
 #ifdef HAVE_GNUTLS_DTLS_SET_DATA_MTU
@@ -426,9 +428,13 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 #endif
 
 		vpninfo->dtls_state = DTLS_CONNECTED;
-		vpn_progress(vpninfo, PRG_INFO,
-			     _("Established DTLS connection (using GnuTLS). Ciphersuite %s.\n"),
-			     vpninfo->dtls_cipher);
+		str = get_gnutls_cipher(vpninfo->dtls_ssl);
+		if (str) {
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("Established DTLS connection (using GnuTLS). Ciphersuite %s.\n"),
+				     str);
+			gnutls_free(str);
+		}
 
 		vpninfo->dtls_times.last_rekey = vpninfo->dtls_times.last_rx = 
 			vpninfo->dtls_times.last_tx = time(NULL);
@@ -445,7 +451,9 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 
 	vpn_progress(vpninfo, PRG_ERR, _("DTLS handshake failed: %s\n"),
 		     gnutls_strerror(err));
-
+	if (err == GNUTLS_E_PUSH_ERROR)
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("(Is a firewall preventing you from sending UDP packets?)\n"));
  error:
 	dtls_close(vpninfo);
 
@@ -493,7 +501,7 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 
 	dtls_fd = socket(vpninfo->peer_addr->sa_family, SOCK_DGRAM, IPPROTO_UDP);
 	if (dtls_fd < 0) {
-		perror(_("Open UDP socket for DTLS:"));
+		vpn_perror(vpninfo, _("Open UDP socket for DTLS:"));
 		return -EINVAL;
 	}
 	if (vpninfo->protect_socket)
@@ -532,14 +540,14 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 		}
 
 		if (bind(dtls_fd, (struct sockaddr *)&dtls_bind_addr, dtls_bind_addrlen)) {
-			perror(_("Bind UDP socket for DTLS"));
+			vpn_perror(vpninfo, _("Bind UDP socket for DTLS"));
 			closesocket(dtls_fd);
 			return -EINVAL;
 		}
 	}
 
 	if (connect(dtls_fd, vpninfo->dtls_addr, vpninfo->peer_addrlen)) {
-		perror(_("UDP (DTLS) connect:\n"));
+		vpn_perror(vpninfo, _("UDP (DTLS) connect:\n"));
 		closesocket(dtls_fd);
 		return -EINVAL;
 	}
@@ -578,7 +586,7 @@ void dtls_close(struct openconnect_info *vpninfo)
 	}
 }
 
-int dtls_reconnect(struct openconnect_info *vpninfo)
+static int dtls_reconnect(struct openconnect_info *vpninfo)
 {
 	dtls_close(vpninfo);
 	vpninfo->dtls_state = DTLS_SLEEPING;
@@ -677,6 +685,12 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	int work_done = 0;
 	char magic_pkt;
 
+	if (vpninfo->dtls_need_reconnect) {
+		vpninfo->dtls_need_reconnect = 0;
+		dtls_reconnect(vpninfo);
+		return 1;
+	}
+
 	if (vpninfo->dtls_state == DTLS_CONNECTING) {
 		dtls_try_handshake(vpninfo);
 		return 0;
@@ -895,11 +909,6 @@ int openconnect_setup_dtls(struct openconnect_info *vpninfo, int dtls_attempt_pe
 	return -EINVAL;
 }
 
-int dtls_reconnect(struct openconnect_info *vpninfo)
-{
-	return -EINVAL;
-}
-
 void dtls_close(struct openconnect_info *vpninfo)
 {
 }
