@@ -43,17 +43,39 @@
 #include <trousers/tss.h>
 #include <trousers/trousers.h>
 #endif
+
 #ifdef HAVE_P11KIT
 #include <p11-kit/p11-kit.h>
 #include <p11-kit/pkcs11.h>
 #include <p11-kit/pin.h>
-
-static P11KitPin *pin_callback(const char *pin_source, P11KitUri *pin_uri,
-			       const char *pin_description,
-			       P11KitPinFlags flags,
-			       void *_vpninfo);
 #endif
 
+#if defined(HAVE_P11KIT) || defined(HAVE_GNUTLS_SYSTEM_KEYS)
+static int gnutls_pin_callback(void *priv, int attempt, const char *uri,
+			       const char *token_label, unsigned int flags,
+			       char *pin, size_t pin_max);
+
+#ifndef HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION
+/* If we don't have this (3.1.0+) then we'll use p11-kit callbacks instead
+ * because the old GnuTLS callback was global rather than context-specific,
+ * which makes it basically unusable from libopenconnect. The p11-kit
+ * callback function is a simple wrapper around the GnuTLS native version. */
+typedef enum {
+        GNUTLS_PIN_USER = (1 << 0),
+        GNUTLS_PIN_SO = (1 << 1),
+        GNUTLS_PIN_FINAL_TRY = (1 << 2),
+        GNUTLS_PIN_COUNT_LOW = (1 << 3),
+        GNUTLS_PIN_CONTEXT_SPECIFIC = (1 << 4),
+        GNUTLS_PIN_WRONG = (1 << 5)
+} gnutls_pin_flag_t;
+
+static P11KitPin *p11kit_pin_callback(const char *pin_source, P11KitUri *pin_uri,
+				      const char *pin_description,
+				      P11KitPinFlags flags,
+				      void *_vpninfo);
+#endif /* !HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION */
+#endif /* HAVE_P11KIT || HAVE_GNUTLS_SYSTEM_KEYS */
+
 #include "gnutls.h"
 #include "openconnect-internal.h"
 
@@ -236,17 +258,12 @@ static int check_certificate_expiry(struct openconnect_info *vpninfo, gnutls_x50
 	return 0;
 }
 
-/* For systems that don't support O_CLOEXEC, just don't bother.
-   It's not open for long anyway. */
-#ifndef O_CLOEXEC
-#define O_CLOEXEC 0
-#endif
-
 static int load_datum(struct openconnect_info *vpninfo,
 		      gnutls_datum_t *datum, const char *fname)
 {
 	struct stat st;
 	int fd, err;
+
 #ifdef ANDROID_KEYSTORE
 	if (!strncmp(fname, "keystore:", 9)) {
 		int len;
@@ -268,9 +285,9 @@ static int load_datum(struct openconnect_info *vpninfo,
 		datum->size = len;
 		return 0;
 	}
-#endif
+#endif /* ANDROID_KEYSTORE */
 
-	fd = open(fname, O_RDONLY|O_CLOEXEC|O_BINARY);
+	fd = openconnect_open_utf8(vpninfo, fname, O_RDONLY|O_CLOEXEC|O_BINARY);
 	if (fd == -1) {
 		err = errno;
 		vpn_progress(vpninfo, PRG_ERR,
@@ -337,12 +354,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo,
 	err = gnutls_pkcs12_import(p12, datum, GNUTLS_X509_FMT_DER, 0);
 	if (err) {
 		gnutls_pkcs12_deinit(p12);
-		if (vpninfo->cert_type == CERT_TYPE_UNKNOWN)
-			return NOT_PKCS12;
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Failed to import PKCS#12 file: %s\n"),
-			     gnutls_strerror(err));
-		return -EINVAL;
+		return NOT_PKCS12;
 	}
 
 	pass = vpninfo->cert_password;
@@ -377,8 +389,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo,
 
 		/* If the first attempt, and we didn't know for sure it was PKCS#12
 		   anyway, bail out and try loading it as something different. */
-		if (pass == vpninfo->cert_password &&
-		    vpninfo->cert_type == CERT_TYPE_UNKNOWN) {
+		if (pass == vpninfo->cert_password) {
 			/* Make it non-fatal... */
 			level = PRG_DEBUG;
 			ret = NOT_PKCS12;
@@ -409,7 +420,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo,
    reference (c1ef7efb in master, 5196786c in gnutls_3_0_x-2)? */
 static int check_issuer_sanity(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer)
 {
-#if GNUTLS_VERSION_NUMBER > 0x300014
+#if GNUTLS_VERSION_NUMBER > 0x030014
 	return 0;
 #else
 	unsigned char id1[512], id2[512];
@@ -460,7 +471,7 @@ static int get_cert_name(gnutls_x509_crt_t cert, char *name, size_t namelen)
 	return 0;
 }
 
-#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS) || defined (HAVE_GNUTLS_SYSTEM_KEYS)
 #ifndef HAVE_GNUTLS_CERTIFICATE_SET_KEY
 /* For GnuTLS 2.12 even if we *have* a privkey (as we do for PKCS#11), we
    can't register it. So we have to use the cert_callback function. This
@@ -593,7 +604,7 @@ static int verify_signed_data(gnutls_pubkey_t pubkey, gnutls_privkey_t privkey,
 	return gnutls_pubkey_verify_data(pubkey, 0, data, sig);
 #endif
 }
-#endif /* (P11KIT || TROUSERS) */
+#endif /* (P11KIT || TROUSERS || SYSTEM_KEYS) */
 
 static int openssl_hash_password(struct openconnect_info *vpninfo, char *pass,
 				 gnutls_datum_t *key, gnutls_datum_t *salt)
@@ -883,13 +894,15 @@ static int load_certificate(struct openconnect_info *vpninfo)
 {
 	gnutls_datum_t fdata;
 	gnutls_x509_privkey_t key = NULL;
-#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS) || defined(HAVE_GNUTLS_SYSTEM_KEYS)
 	gnutls_privkey_t pkey = NULL;
 	gnutls_datum_t pkey_sig = {NULL, 0};
 	void *dummy_hash_data = &load_certificate;
 #endif
-#ifdef HAVE_P11KIT
+#if defined(HAVE_P11KIT) || defined(HAVE_GNUTLS_SYSTEM_KEYS)
 	char *cert_url = (char *)vpninfo->cert;
+#endif
+#ifdef HAVE_P11KIT
 	char *key_url = (char *)vpninfo->sslkey;
 	gnutls_pkcs11_privkey_t p11key = NULL;
 #endif
@@ -903,6 +916,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	int ret;
 	int i;
 	int cert_is_p11 = 0, key_is_p11 = 0;
+	int cert_is_sys = 0, key_is_sys = 0;
 	unsigned char key_id[20];
 	size_t key_id_size = sizeof(key_id);
 	char name[80];
@@ -911,7 +925,16 @@ static int load_certificate(struct openconnect_info *vpninfo)
 
 	key_is_p11 = !strncmp(vpninfo->sslkey, "pkcs11:", 7);
 	cert_is_p11 = !strncmp(vpninfo->cert, "pkcs11:", 7);
+	key_is_sys = !strncmp(vpninfo->sslkey, "system:", 7);
+	cert_is_sys = !strncmp(vpninfo->cert, "system:", 7);
 
+#ifndef HAVE_GNUTLS_SYSTEM_KEYS
+	if (key_is_sys || cert_is_sys) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("This binary built without system key support\n"));
+		return -EINVAL;
+	}
+#endif
 #ifndef HAVE_P11KIT
 	if (key_is_p11 || cert_is_p11) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -923,12 +946,13 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	if (key_is_p11 || cert_is_p11) {
 		CK_OBJECT_CLASS class;
 		CK_ATTRIBUTE attr;
-		char pin_source[40];
 		P11KitUri *uri;
+#ifndef HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION
+		char pin_source[40];
 
 		sprintf(pin_source, "openconnect:%p", vpninfo);
-		p11_kit_pin_register_callback(pin_source, pin_callback, vpninfo, NULL);
-
+		p11_kit_pin_register_callback(pin_source, p11kit_pin_callback, vpninfo, NULL);
+#endif
 		uri = p11_kit_uri_new();
 
 		attr.type = CKA_CLASS;
@@ -939,8 +963,10 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		   both certificate and key URLs, unless they already exist. */
 		if (cert_is_p11 &&
 		    !p11_kit_uri_parse(cert_url, P11_KIT_URI_FOR_ANY, uri)) {
+#ifndef HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION
 			if (!p11_kit_uri_get_pin_source(uri))
 				p11_kit_uri_set_pin_source(uri, pin_source);
+#endif
 			if (!p11_kit_uri_get_attribute(uri, CKA_CLASS)) {
 				class = CKO_CERTIFICATE;
 				p11_kit_uri_set_attribute(uri, &attr);
@@ -950,9 +976,12 @@ static int load_certificate(struct openconnect_info *vpninfo)
 
 		if (key_is_p11 &&
 		    !p11_kit_uri_parse(key_url, P11_KIT_URI_FOR_ANY, uri)) {
+#ifndef HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION
 			if (!p11_kit_uri_get_pin_source(uri))
 				p11_kit_uri_set_pin_source(uri, pin_source);
-			if (!p11_kit_uri_get_attribute(uri, CKA_CLASS)) {
+#endif
+			if (vpninfo->sslkey == vpninfo->cert ||
+			    !p11_kit_uri_get_attribute(uri, CKA_CLASS)) {
 				class = CKO_PRIVATE_KEY;
 				p11_kit_uri_set_attribute(uri, &attr);
 			}
@@ -961,31 +990,40 @@ static int load_certificate(struct openconnect_info *vpninfo)
 
 		p11_kit_uri_free(uri);
 	}
+#endif /* HAVE_PKCS11 */
 
+#if defined (HAVE_P11KIT) || defined(HAVE_GNUTLS_SYSTEM_KEYS)
 	/* Load certificate(s) first... */
-	if (cert_is_p11) {
+	if (cert_is_p11 || cert_is_sys) {
 		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("Using PKCS#11 certificate %s\n"), cert_url);
+			     cert_is_p11 ? _("Using PKCS#11 certificate %s\n") :
+			     _("Using system certificate %s\n"), cert_url);
 
 		err = gnutls_x509_crt_init(&cert);
 		if (err) {
 			ret = -ENOMEM;
 			goto out;
 		}
+#ifdef HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION
+		gnutls_x509_crt_set_pin_function(cert, gnutls_pin_callback, vpninfo);
+#endif
+		/* Yes, even for *system* URLs the only API GnuTLS offers us is
+		   ...import_pkcs11_url(). */
 		err = gnutls_x509_crt_import_pkcs11_url(cert, cert_url, 0);
 		if (err == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
 			err = gnutls_x509_crt_import_pkcs11_url(cert, cert_url,
 								GNUTLS_PKCS11_OBJ_FLAG_LOGIN);
 		if (err) {
 			vpn_progress(vpninfo, PRG_ERR,
-				     _("Error loading certificate from PKCS#11: %s\n"),
+				     cert_is_p11 ? _("Error loading certificate from PKCS#11: %s\n") :
+				     _("Error loading system certificate: %s\n"),
 				     gnutls_strerror(err));
 			ret = -EIO;
 			goto out;
 		}
 		goto got_certs;
 	}
-#endif /* HAVE_P11KIT */
+#endif /* HAVE_P11KIT || HAVE_GNUTLS_SYSTEM_KEYS */
 
 	/* OK, not a PKCS#11 certificate so it must be coming from a file... */
 	vpn_progress(vpninfo, PRG_DEBUG,
@@ -997,8 +1035,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		return ret;
 
 	/* Is it PKCS#12? */
-	if (!key_is_p11 && (vpninfo->cert_type == CERT_TYPE_PKCS12 ||
-			    vpninfo->cert_type == CERT_TYPE_UNKNOWN)) {
+	if (!key_is_p11) {
 		/* PKCS#12 should actually contain certificates *and* private key */
 		ret = load_pkcs12_certificate(vpninfo, &fdata, &key,
 					      &supporting_certs, &nr_supporting_certs,
@@ -1061,10 +1098,37 @@ static int load_certificate(struct openconnect_info *vpninfo)
  got_certs:
 	/* Now we have either a single certificate in 'cert', or an array of
 	   them in extra_certs[]. Next we look for the private key ... */
+#ifdef HAVE_GNUTLS_SYSTEM_KEYS
+	if (key_is_sys) {
+		vpn_progress(vpninfo, PRG_DEBUG,
+			     _("Using system key %s\n"), vpninfo->sslkey);
+
+		err = gnutls_privkey_init(&pkey);
+		if (err) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Error initialising private key structure: %s\n"),
+				     gnutls_strerror(err));
+			ret = -EIO;
+			goto out;
+		}
+#ifdef HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION
+		gnutls_privkey_set_pin_function(pkey, gnutls_pin_callback, vpninfo);
+#endif
+		err = gnutls_privkey_import_url(pkey, vpninfo->sslkey, 0);
+		if (err) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Error importing system key %s: %s\n"),
+				     vpninfo->sslkey, gnutls_strerror(err));
+			ret = -EIO;
+			goto out;
+		}
+		goto match_cert;
+	}
+#endif /* HAVE_GNUTLS_SYSTEM_KEYS */
 #if defined(HAVE_P11KIT)
 	if (key_is_p11) {
-		vpn_progress(vpninfo, PRG_DEBUG,
-			     _("Using PKCS#11 key %s\n"), key_url);
+		vpn_progress(vpninfo, PRG_TRACE,
+			     _("Trying PKCS#11 key URL %s\n"), key_url);
 
 		err = gnutls_pkcs11_privkey_init(&p11key);
 		if (err) {
@@ -1074,7 +1138,9 @@ static int load_certificate(struct openconnect_info *vpninfo)
 			ret = -EIO;
 			goto out;
 		}
-
+#ifdef HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION
+		gnutls_pkcs11_privkey_set_pin_function(p11key, gnutls_pin_callback, vpninfo);
+#endif
 		err = gnutls_pkcs11_privkey_import_url(p11key, key_url, 0);
 
 		/* Annoyingly, some tokens don't even admit the *existence* of
@@ -1088,7 +1154,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 			gnutls_pkcs11_obj_t crt;
 			P11KitUri *uri;
 			CK_TOKEN_INFO *token;
-			char buf[33];
+			char buf[65];
 			size_t s;
 
 			if (gnutls_pkcs11_obj_init(&crt))
@@ -1148,8 +1214,41 @@ static int load_certificate(struct openconnect_info *vpninfo)
 
 			free(key_url);
 			key_url = NULL;
-			if (!p11_kit_uri_format(uri, P11_KIT_URI_FOR_ANY, &key_url))
+			if (p11_kit_uri_format(uri, P11_KIT_URI_FOR_ANY, &key_url))
+				goto key_err_uri;
+
+			vpn_progress(vpninfo, PRG_TRACE,
+				     _("Trying PKCS#11 key URL %s\n"), key_url);
+			err = gnutls_pkcs11_privkey_import_url(p11key, key_url, 0);
+
+			/* If it still doesn't work then try dropping CKA_LABEL and adding the
+			   CKA_ID of the cert. */
+			if (err == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE &&
+			    (p11_kit_uri_get_attribute(uri, CKA_LABEL) ||
+			     !p11_kit_uri_get_attribute(uri, CKA_ID))) {
+				CK_ATTRIBUTE attr;
+
+				s = sizeof(buf);
+				if (gnutls_pkcs11_obj_get_info(crt, GNUTLS_PKCS11_OBJ_ID,
+							       buf, &s))
+					goto key_err_uri;
+
+				attr.type = CKA_ID;
+				attr.pValue = buf;
+				attr.ulValueLen = s;
+
+				p11_kit_uri_set_attribute(uri, &attr);
+				p11_kit_uri_clear_attribute(uri, CKA_LABEL);
+
+				free(key_url);
+				key_url = NULL;
+				if (p11_kit_uri_format(uri, P11_KIT_URI_FOR_ANY, &key_url))
+					goto key_err_uri;
+
+				vpn_progress(vpninfo, PRG_TRACE,
+					     _("Trying PKCS#11 key URL %s\n"), key_url);
 				err = gnutls_pkcs11_privkey_import_url(p11key, key_url, 0);
+			}
 		key_err_uri:
 			p11_kit_uri_free(uri);
 		key_err_obj:
@@ -1165,6 +1264,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 			ret = -EIO;
 			goto out;
 		}
+		vpn_progress(vpninfo, PRG_DEBUG,
+			     _("Using PKCS#11 key %s\n"), key_url);
 
 		err = gnutls_privkey_init(&pkey);
 		if (err) {
@@ -1215,9 +1316,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	}
 
 	/* Is it a PEM file with a TPM key blob? */
-	if (vpninfo->cert_type == CERT_TYPE_TPM ||
-	    (vpninfo->cert_type == CERT_TYPE_UNKNOWN &&
-	     strstr((char *)fdata.data, "-----BEGIN TSS KEY BLOB-----"))) {
+	if (strstr((char *)fdata.data, "-----BEGIN TSS KEY BLOB-----")) {
 #ifndef HAVE_TROUSERS
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("This version of OpenConnect was built without TPM support\n"));
@@ -1347,7 +1446,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	   enabled we'll fall straight through the bit at match_cert: below, and go
 	   directly to the bit where it prints the 'no match found' error and exits. */
 
-#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS) || defined(HAVE_GNUTLS_SYSTEM_KEYS)
  match_cert:
 	/* If we have a privkey from PKCS#11 or TPM, we can't do the simple comparison
 	   of key ID that we do for software keys to find which certificate is a
@@ -1401,7 +1500,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		}
 		gnutls_free(pkey_sig.data);
 	}
-#endif /* P11KIT || TROUSERS */
+#endif /* P11KIT || TROUSERS || SYSTEM_KEYS */
 
 	/* We shouldn't reach this. It means that we didn't find *any* matching cert */
 	vpn_progress(vpninfo, PRG_ERR,
@@ -1416,6 +1515,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	   a PKCS#12 file we may have a trust chain in 'supporting_certs[]' too. */
 	check_certificate_expiry(vpninfo, cert);
 	get_cert_name(cert, name, sizeof(name));
+	get_cert_md5_fingerprint(vpninfo, cert, vpninfo->local_cert_md5);
 	vpn_progress(vpninfo, PRG_INFO, _("Using client certificate '%s'\n"),
 		     name);
 
@@ -1506,7 +1606,10 @@ static int load_certificate(struct openconnect_info *vpninfo)
 						err = gnutls_x509_crt_import(issuer, &t, GNUTLS_X509_FMT_DER);
 						if (err)
 							gnutls_x509_crt_deinit(issuer);
+						else
+							free_issuer = 1;
 					}
+					gnutls_free(t.data);
 				}
 				if (err) {
 					vpn_progress(vpninfo, PRG_ERR,
@@ -1517,8 +1620,6 @@ static int load_certificate(struct openconnect_info *vpninfo)
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Got next CA '%s' from PKCS11\n"), name);
 				}
-				free_issuer = 1;
-				gnutls_free(t.data);
 			}
 #endif
 			if (err)
@@ -1576,7 +1677,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	   key and certs. GnuTLS makes us do this differently for X509 privkeys
 	   vs. TPM/PKCS#11 "generic" privkeys, and the latter is particularly
 	   'fun' for GnuTLS 2.12... */
-#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS) || defined(HAVE_GNUTLS_SYSTEM_KEYS)
 	if (pkey) {
 		err = assign_privkey(vpninfo, pkey,
 				     supporting_certs,
@@ -1624,7 +1725,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	}
 	gnutls_free(extra_certs);
 
-#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS)
+#if defined(HAVE_P11KIT) || defined(HAVE_TROUSERS) || defined(HAVE_GNUTLS_SYSTEM_KEYS)
 	if (pkey && pkey != OPENCONNECT_TPM_PKEY)
 		gnutls_privkey_deinit(pkey);
 	/* If we support arbitrary privkeys, we might have abused fdata.data
@@ -1634,6 +1735,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		gnutls_free(fdata.data);
 
 #ifdef HAVE_P11KIT
+	/* This exists in the HAVE_GNUTLS_SYSTEM_KEYS case but will never
+	   change so it's OK not to add to the #ifdef mess here. */
 	if (cert_url != vpninfo->cert)
 		free(cert_url);
 	if (key_url != vpninfo->sslkey)
@@ -1661,55 +1764,114 @@ static int get_cert_fingerprint(struct openconnect_info *vpninfo,
 }
 
 int get_cert_md5_fingerprint(struct openconnect_info *vpninfo,
-			     OPENCONNECT_X509 *cert, char *buf)
+			     void *cert, char *buf)
 {
 	return get_cert_fingerprint(vpninfo, cert, GNUTLS_DIG_MD5, buf);
 }
 
-int openconnect_get_cert_sha1(struct openconnect_info *vpninfo,
-			      OPENCONNECT_X509 *cert, char *buf)
+static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 {
-	return get_cert_fingerprint(vpninfo, cert, GNUTLS_DIG_SHA1, buf);
+	unsigned char sha1[SHA1_SIZE];
+	size_t shalen;
+	gnutls_pubkey_t pkey;
+	gnutls_datum_t d;
+	int i, err;
+
+	err = gnutls_pubkey_init(&pkey);
+	if (err)
+		return err;
+
+	err = gnutls_pubkey_import_x509(pkey, vpninfo->peer_cert, 0);
+	if (err) {
+		gnutls_pubkey_deinit(pkey);
+		return err;
+	}
+#ifdef HAVE_GNUTLS_PUBKEY_EXPORT2
+	err = gnutls_pubkey_export2(pkey, GNUTLS_X509_FMT_DER, &d);
+	if (err) {
+		gnutls_pubkey_deinit(pkey);
+		return err;
+	}
+#else
+	shalen = 0;
+	err = gnutls_pubkey_export(pkey, GNUTLS_X509_FMT_DER, NULL, &shalen);
+	if (err != GNUTLS_E_SHORT_MEMORY_BUFFER) {
+		gnutls_pubkey_deinit(pkey);
+		return err;
+	}
+	d.size = shalen;
+	d.data = gnutls_malloc(d.size);
+	if (!d.data) {
+		gnutls_pubkey_deinit(pkey);
+		return -ENOMEM;
+	}
+	err = gnutls_pubkey_export(pkey, GNUTLS_X509_FMT_DER, d.data, &shalen);
+	if (err) {
+		gnutls_free(d.data);
+		gnutls_pubkey_deinit(pkey);
+		return err;
+	}
+#endif
+	gnutls_pubkey_deinit(pkey);
+	shalen = SHA1_SIZE;
+
+	err = gnutls_fingerprint(GNUTLS_DIG_SHA1, &d, sha1, &shalen);
+	if (err) {
+		gnutls_free(d.data);
+		return err;
+	}
+
+	gnutls_free(d.data);
+
+	vpninfo->peer_cert_hash = malloc(SHA1_SIZE * 2 + 6);
+	if (vpninfo->peer_cert_hash) {
+		snprintf(vpninfo->peer_cert_hash, 6, "sha1:");
+		for (i = 0; i < shalen; i++)
+			sprintf(&vpninfo->peer_cert_hash[i*2 + 5], "%02x", sha1[i]);
+	}
+
+	return 0;
 }
 
-char *openconnect_get_cert_details(struct openconnect_info *vpninfo,
-				   OPENCONNECT_X509 *cert)
+char *openconnect_get_peer_cert_details(struct openconnect_info *vpninfo)
 {
 	gnutls_datum_t buf;
-	char *ret;
 
-	if (gnutls_x509_crt_print(cert, GNUTLS_CRT_PRINT_FULL, &buf))
+	if (gnutls_x509_crt_print(vpninfo->peer_cert, GNUTLS_CRT_PRINT_FULL, &buf))
 		return NULL;
 
-	/* Just in case gnutls_free() isn't free(), we can't steal it. */
-	ret = strdup((char *)buf.data);
-	gnutls_free(buf.data);
-
-	return ret;
+	return (char *)buf.data;
 }
 
-int openconnect_get_cert_DER(struct openconnect_info *vpninfo,
-			     OPENCONNECT_X509 *cert, unsigned char **buf)
+int openconnect_get_peer_cert_DER(struct openconnect_info *vpninfo,
+				  unsigned char **buf)
 {
 	size_t l = 0;
 	unsigned char *ret = NULL;
 
-	if (gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, ret, &l) !=
-	    GNUTLS_E_SHORT_MEMORY_BUFFER)
+	if (gnutls_x509_crt_export(vpninfo->peer_cert, GNUTLS_X509_FMT_DER,
+				   ret, &l) != GNUTLS_E_SHORT_MEMORY_BUFFER)
 		return -EIO;
 
-	ret = malloc(l);
+	ret = gnutls_malloc(l);
 	if (!ret)
 		return -ENOMEM;
 
-	if (gnutls_x509_crt_export(cert, GNUTLS_X509_FMT_DER, ret, &l)) {
-		free(ret);
+	if (gnutls_x509_crt_export(vpninfo->peer_cert, GNUTLS_X509_FMT_DER,
+				   ret, &l)) {
+		gnutls_free(ret);
 		return -EIO;
 	}
 	*buf = ret;
 	return l;
 }
 
+void openconnect_free_cert_info(struct openconnect_info *vpninfo,
+				void *buf)
+{
+	gnutls_free(buf);
+}
+
 static int verify_peer(gnutls_session_t session)
 {
 	struct openconnect_info *vpninfo = gnutls_session_get_ptr(session);
@@ -1717,12 +1879,7 @@ static int verify_peer(gnutls_session_t session)
 	gnutls_x509_crt_t cert;
 	unsigned int status, cert_list_size;
 	const char *reason = NULL;
-	int err;
-
-	if (vpninfo->peer_cert) {
-		gnutls_x509_crt_deinit(vpninfo->peer_cert);
-		vpninfo->peer_cert = NULL;
-	}
+	int err = 0;
 
 	cert_list = gnutls_certificate_get_peers(session, &cert_list_size);
 	if (!cert_list) {
@@ -1730,26 +1887,24 @@ static int verify_peer(gnutls_session_t session)
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
-	if (vpninfo->servercert) {
-		unsigned char sha1bin[SHA1_SIZE];
-		char fingerprint[(SHA1_SIZE * 2) + 1];
-		int i;
+	err = gnutls_x509_crt_init(&cert);
+	if (err) {
+		vpn_progress(vpninfo, PRG_ERR, _("Error initialising X509 cert structure\n"));
+		return GNUTLS_E_CERTIFICATE_ERROR;
+	}
 
-		err = openconnect_sha1(sha1bin, cert_list[0].data, cert_list[0].size);
-		if (err) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Could not calculate SHA1 of server's certificate\n"));
-			return GNUTLS_E_CERTIFICATE_ERROR;
-		}
-		for (i = 0; i < SHA1_SIZE; i++)
-			sprintf(&fingerprint[i*2], "%02X", sha1bin[i]);
+	err = gnutls_x509_crt_import(cert, &cert_list[0], GNUTLS_X509_FMT_DER);
+	if (err) {
+		vpn_progress(vpninfo, PRG_ERR, _("Error importing server's cert\n"));
+		gnutls_x509_crt_deinit(cert);
+		return GNUTLS_E_CERTIFICATE_ERROR;
+	}
 
-		if (strcasecmp(vpninfo->servercert, fingerprint)) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Server SSL certificate didn't match: %s\n"), fingerprint);
-			return GNUTLS_E_CERTIFICATE_ERROR;
-		}
-		return 0;
+	vpninfo->peer_cert = cert;
+	err = set_peer_cert_hash(vpninfo);
+	if (err < 0) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Could not calculate hash of server's certificate\n"));
 	}
 
 	err = gnutls_certificate_verify_peers2(session, &status);
@@ -1776,19 +1931,6 @@ static int verify_peer(gnutls_session_t session)
 		   why we don't just set a bit for that too. */
 		reason = _("signature verification failed");
 
-	err = gnutls_x509_crt_init(&cert);
-	if (err) {
-		vpn_progress(vpninfo, PRG_ERR, _("Error initialising X509 cert structure\n"));
-		return GNUTLS_E_CERTIFICATE_ERROR;
-	}
-
-	err = gnutls_x509_crt_import(cert, &cert_list[0], GNUTLS_X509_FMT_DER);
-	if (err) {
-		vpn_progress(vpninfo, PRG_ERR, _("Error importing server's cert\n"));
-		gnutls_x509_crt_deinit(cert);
-		return GNUTLS_E_CERTIFICATE_ERROR;
-	}
-
 	if (reason)
 		goto done;
 
@@ -1844,14 +1986,11 @@ static int verify_peer(gnutls_session_t session)
 			     reason);
 		if (vpninfo->validate_peer_cert)
 			err = vpninfo->validate_peer_cert(vpninfo->cbdata,
-							  cert,
 							  reason) ? GNUTLS_E_CERTIFICATE_ERROR : 0;
 		else
 			err = GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
-	vpninfo->peer_cert = cert;
-
 	return err;
 }
 
@@ -1868,7 +2007,7 @@ static int verify_peer(gnutls_session_t session)
 # define _DEFAULT_PRIO "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.0:" \
 	"%COMPAT:%DISABLE_SAFE_RENEGOTIATION:%LATEST_RECORD_VERSION"
 # if GNUTLS_VERSION_MAJOR >= 3
-#  define DEFAULT_PRIO _DEFAULT_PRIO":-CURVE-ALL"
+#  define DEFAULT_PRIO _DEFAULT_PRIO":-CURVE-ALL:-ECDHE-RSA:-ECDHE-ECDSA"
 #else
 #  define DEFAULT_PRIO _DEFAULT_PRIO
 # endif
@@ -1883,19 +2022,30 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	if (vpninfo->https_sess)
 		return 0;
 
+	if (vpninfo->peer_cert) {
+		gnutls_x509_crt_deinit(vpninfo->peer_cert);
+		vpninfo->peer_cert = NULL;
+	}
+	free(vpninfo->peer_cert_hash);
+	vpninfo->peer_cert_hash = 0;
+	gnutls_free(vpninfo->cstp_cipher);
+	vpninfo->cstp_cipher = NULL;
+
 	ssl_sock = connect_https_socket(vpninfo);
 	if (ssl_sock < 0)
 		return ssl_sock;
 
 	if (!vpninfo->https_cred) {
 		gnutls_certificate_allocate_credentials(&vpninfo->https_cred);
+		if (!vpninfo->no_system_trust) {
 #ifdef HAVE_GNUTLS_CERTIFICATE_SET_X509_SYSTEM_TRUST
-		gnutls_certificate_set_x509_system_trust(vpninfo->https_cred);
+			gnutls_certificate_set_x509_system_trust(vpninfo->https_cred);
 #else
-		gnutls_certificate_set_x509_trust_file(vpninfo->https_cred,
-						       DEFAULT_SYSTEM_CAFILE,
-						       GNUTLS_X509_FMT_PEM);
+			gnutls_certificate_set_x509_trust_file(vpninfo->https_cred,
+							       DEFAULT_SYSTEM_CAFILE,
+							       GNUTLS_X509_FMT_PEM);
 #endif
+		}
 		gnutls_certificate_set_verify_function(vpninfo->https_cred,
 						       verify_peer);
 
@@ -2017,11 +2167,18 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	vpn_progress(vpninfo, PRG_INFO, _("SSL negotiation with %s\n"),
 		     vpninfo->hostname);
 
+#ifdef GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT
+	gnutls_handshake_set_timeout(vpninfo->https_sess,
+				     GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);
+#endif
 
 	err = cstp_handshake(vpninfo, 1);
 	if (err)
 		return err;
 
+	gnutls_free(vpninfo->cstp_cipher);
+	vpninfo->cstp_cipher = get_gnutls_cipher(vpninfo->https_sess);
+
 	vpninfo->ssl_fd = ssl_sock;
 
 	vpninfo->ssl_read = openconnect_gnutls_read;
@@ -2088,10 +2245,6 @@ int cstp_handshake(struct openconnect_info *vpninfo, unsigned init)
 
 void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 {
-	if (vpninfo->peer_cert) {
-		gnutls_x509_crt_deinit(vpninfo->peer_cert);
-		vpninfo->peer_cert = NULL;
-	}
 	if (vpninfo->https_sess) {
 		gnutls_deinit(vpninfo->https_sess);
 		vpninfo->https_sess = NULL;
@@ -2109,11 +2262,11 @@ void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 #ifdef HAVE_P11KIT
 		if ((vpninfo->cert && !strncmp(vpninfo->cert, "pkcs11:", 7)) ||
 		    (vpninfo->sslkey && !strncmp(vpninfo->sslkey, "pkcs11:", 7))) {
+#ifndef HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION
 			char pin_source[40];
-
 			sprintf(pin_source, "openconnect:%p", vpninfo);
-			p11_kit_pin_unregister_callback(pin_source, pin_callback, vpninfo);
-
+			p11_kit_pin_unregister_callback(pin_source, p11kit_pin_callback, vpninfo);
+#endif
 			while (vpninfo->pin_cache) {
 				struct pin_cache *cache = vpninfo->pin_cache;
 
@@ -2167,12 +2320,30 @@ void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 	}
 }
 
-void openconnect_init_ssl(void)
+int openconnect_init_ssl(void)
 {
 #ifdef _WIN32
-	openconnect__win32_sock_init();
+	int ret = openconnect__win32_sock_init();
+	if (ret)
+		return ret;
 #endif
-	gnutls_global_init();
+	if (gnutls_global_init())
+		return -EIO;
+
+	return 0;
+}
+
+char *get_gnutls_cipher(gnutls_session_t session)
+{
+	char *str;
+#if GNUTLS_VERSION_NUMBER > 0x03010a
+	str = gnutls_session_get_desc(session);
+#else
+	str = gnutls_strdup(gnutls_cipher_suite_get_name(
+		gnutls_kx_get(session), gnutls_cipher_get(session),
+		gnutls_mac_get(session)));
+#endif
+	return str;
 }
 
 int openconnect_sha1(unsigned char *result, void *data, int datalen)
@@ -2211,49 +2382,33 @@ int openconnect_random(void *bytes, int len)
 int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
 			       char *buf)
 {
-	const gnutls_datum_t *d;
-	size_t md5len = 16;
-
-	buf[0] = 0;
-
-	d = gnutls_certificate_get_ours(vpninfo->https_sess);
-	if (!d)
-		return -EIO;
-
-	if (gnutls_fingerprint(GNUTLS_DIG_MD5, d, buf, &md5len))
-		return -EIO;
+	memcpy(buf, vpninfo->local_cert_md5, sizeof(vpninfo->local_cert_md5));
 
 	return 0;
 }
 
-#ifdef HAVE_P11KIT
-static P11KitPin *pin_callback(const char *pin_source, P11KitUri *pin_uri,
-			const char *pin_description,
-			P11KitPinFlags flags,
-			void *_vpninfo)
+#if defined(HAVE_P11KIT) || defined(HAVE_GNUTLS_SYSTEM_KEYS)
+static int gnutls_pin_callback(void *priv, int attempt, const char *uri,
+			       const char *token_label, unsigned int flags,
+			       char *pin, size_t pin_max)
 {
-	struct openconnect_info *vpninfo = _vpninfo;
+	struct openconnect_info *vpninfo = priv;
 	struct pin_cache **cache = &vpninfo->pin_cache;
 	struct oc_auth_form f;
 	struct oc_form_opt o;
 	char message[1024];
-	char *uri;
-	P11KitPin *pin;
 	int ret;
 
 	if (!vpninfo || !vpninfo->process_auth_form)
-		return NULL;
-
-	if (p11_kit_uri_format(pin_uri, P11_KIT_URI_FOR_TOKEN, &uri))
-		return NULL;
+		return -1;
 
 	while (*cache) {
 		if (!strcmp(uri, (*cache)->token)) {
-			free(uri);
-			uri = NULL;
 			if ((*cache)->pin) {
-				if ((flags & P11_KIT_PIN_FLAGS_RETRY) != P11_KIT_PIN_FLAGS_RETRY)
-					return p11_kit_pin_new_for_string((*cache)->pin);
+				if (attempt == 0) {
+					snprintf(pin, pin_max, "%s", (*cache)->pin);
+					return 0;
+				}
 				memset((*cache)->pin, 0x5a, strlen((*cache)->pin));
 				free((*cache)->pin);
 				(*cache)->pin = NULL;
@@ -2264,11 +2419,10 @@ static P11KitPin *pin_callback(const char *pin_source, P11KitUri *pin_uri,
 	}
 	if (!*cache) {
 		*cache = calloc(1, sizeof(struct pin_cache));
-		if (!*cache) {
-			free(uri);
-			return NULL;
-		}
-		(*cache)->token = uri;
+		if (!*cache)
+			return -1;
+
+		(*cache)->token = strdup(uri);
 	}
 
 	memset(&f, 0, sizeof(f));
@@ -2276,36 +2430,124 @@ static P11KitPin *pin_callback(const char *pin_source, P11KitUri *pin_uri,
 	f.opts = &o;
 
 	message[sizeof(message)-1] = 0;
-	snprintf(message, sizeof(message) - 1, _("PIN required for %s"), pin_description);
+	snprintf(message, sizeof(message) - 1, _("PIN required for %s"), token_label);
 	f.message = message;
 
+	if (flags & GNUTLS_PIN_WRONG)
+		f.error = (char *)_("Wrong PIN");
+
+	if (flags & GNUTLS_PIN_FINAL_TRY)
+		f.banner = (char *)_("This is the final try before locking!");
+	else if (flags & GNUTLS_PIN_COUNT_LOW)
+		f.banner = (char *)_("Only a few tries left before locking!");
+
+	o.next = NULL;
+	o.type = OC_FORM_OPT_PASSWORD;
+	o.name = (char *)"pkcs11_pin";
+	o.label = (char *)_("Enter PIN:");
+	o._value = NULL;
+
+	ret = process_auth_form(vpninfo, &f);
+	if (ret || !o._value)
+		return -1;
+
+	snprintf(pin, pin_max, "%s", o._value);
+	(*cache)->pin = o._value;
+
+	return 0;
+}
+
+#ifndef HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION
+static P11KitPin *p11kit_pin_callback(const char *pin_source, P11KitUri *pin_uri,
+				      const char *pin_description,
+				      P11KitPinFlags flags,
+				      void *_vpninfo)
+{
+	struct openconnect_info *vpninfo = _vpninfo;
+	char *uri;
+	P11KitPin *pin = NULL;
+	char pin_str[1024];
+	unsigned gnutls_flags = 0;
+	int attempt = 0;
+
+	if (!vpninfo || !vpninfo->process_auth_form)
+		return NULL;
+
+	if (p11_kit_uri_format(pin_uri, P11_KIT_URI_FOR_TOKEN, &uri))
+		return NULL;
+
 	/*
 	 * In p11-kit <= 0.12, these flags are *odd*.
 	 * RETRY is 0xa, FINAL_TRY is 0x14 and MANY_TRIES is 0x28.
 	 * So don't treat it like a sane bitmask. Fixed in
 	 * http://cgit.freedesktop.org/p11-glue/p11-kit/commit/?id=59774b11
 	 */
-	if ((flags & P11_KIT_PIN_FLAGS_RETRY) == P11_KIT_PIN_FLAGS_RETRY)
-		f.error = (char *)_("Wrong PIN");
-
+	if ((flags & P11_KIT_PIN_FLAGS_RETRY) == P11_KIT_PIN_FLAGS_RETRY) {
+		attempt = 1;
+		gnutls_flags |= GNUTLS_PIN_WRONG;
+	}
 	if ((flags & P11_KIT_PIN_FLAGS_FINAL_TRY) == P11_KIT_PIN_FLAGS_FINAL_TRY)
-		f.banner = (char *)_("This is the final try before locking!");
-	else if ((flags & P11_KIT_PIN_FLAGS_MANY_TRIES) == P11_KIT_PIN_FLAGS_MANY_TRIES)
-		f.banner = (char *)_("Only a few tries left before locking!");
+		gnutls_flags |= GNUTLS_PIN_FINAL_TRY;
+	if ((flags & P11_KIT_PIN_FLAGS_MANY_TRIES) == P11_KIT_PIN_FLAGS_MANY_TRIES)
+		gnutls_flags |= GNUTLS_PIN_COUNT_LOW;
 
-	o.next = NULL;
-	o.type = OC_FORM_OPT_PASSWORD;
-	o.name = (char *)"pkcs11_pin";
-	o.label = (char *)_("Enter PIN:");
-	o.value = NULL;
+	if (!gnutls_pin_callback(vpninfo, attempt, uri, pin_description,
+				gnutls_flags, pin_str, sizeof(pin_str)))
+		pin = p11_kit_pin_new_for_string(pin_str);
 
-	ret = process_auth_form(vpninfo, &f);
-	if (ret || !o.value)
-		return NULL;
-
-	pin = p11_kit_pin_new_for_string(o.value);
-	(*cache)->pin = o.value;
+	memset(pin_str, 0x5a, sizeof(pin_str));
+	free(uri);
 
 	return pin;
 }
+#endif /* !HAVE_GNUTLS_X509_CRT_SET_PIN_FUNCTION */
+#endif /* HAVE_P11KIT || HAVE_GNUTLS_SYSTEM_KEYS */
+
+#ifdef HAVE_LIBPCSCLITE
+int openconnect_hash_yubikey_password(struct openconnect_info *vpninfo,
+				      const char *password, const void *ident, int id_len)
+{
+	unsigned char U[SHA1_SIZE];
+	gnutls_hmac_hd_t dgst;
+	int ret = -EIO;
+	int i, j;
+
+	if (gnutls_hmac_init(&dgst, GNUTLS_MAC_SHA1, password, strlen(password)))
+		return -EIO;
+
+	if (gnutls_hmac(dgst, ident, id_len))
+		goto out;
+	/* This is a subset of full PBKDF2, where we know the outer loop is only
+	 * run once because our output length (16) is less than the hash output
+	 * size (20). So just hard-code the value. */
+	if (gnutls_hmac(dgst, "\x0\x0\x0\x1", 4))
+		goto out;
+	gnutls_hmac_output(dgst, U);
+
+	memcpy(vpninfo->yubikey_pwhash, U, 16);
+
+	for (i = 1; i < 1000; i++) {
+		if (gnutls_hmac(dgst, U, SHA1_SIZE))
+			goto out;
+
+		gnutls_hmac_output(dgst, U);
+
+		for (j = 0; j < 16; j++)
+			vpninfo->yubikey_pwhash[j] ^= U[j];
+	}
+
+	ret = 0;
+ out:
+	gnutls_hmac_deinit(dgst, NULL);
+	return ret;
+}
+
+int openconnect_yubikey_chalresp(struct openconnect_info *vpninfo,
+				 const void *challenge, int chall_len, void *result)
+{
+	if (gnutls_hmac_fast(GNUTLS_MAC_SHA1, vpninfo->yubikey_pwhash, 16, challenge, chall_len, result))
+		return -EIO;
+
+	return 0;
+}
 #endif
