@@ -23,6 +23,7 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
+#include <ctype.h>
 
 #ifdef HAVE_LIBSTOKEN
 #include <stoken.h>
@@ -37,7 +38,11 @@
 
 #include "openconnect-internal.h"
 
-struct openconnect_info *openconnect_vpninfo_new(char *useragent,
+#if defined(OPENCONNECT_GNUTLS)
+#include "gnutls.h"
+#endif
+
+struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 						 openconnect_validate_peer_cert_vfn validate_peer_cert,
 						 openconnect_write_new_config_vfn write_new_config,
 						 openconnect_process_auth_form_vfn process_auth_form,
@@ -45,10 +50,22 @@ struct openconnect_info *openconnect_vpninfo_new(char *useragent,
 						 void *privdata)
 {
 	struct openconnect_info *vpninfo = calloc(sizeof(*vpninfo), 1);
+#ifdef HAVE_ICONV
+	char *charset = nl_langinfo(CODESET);
+#endif
 
 	if (!vpninfo)
 		return NULL;
 
+#ifdef HAVE_ICONV
+	if (charset && strcmp(charset, "UTF-8")) {
+		vpninfo->ic_utf8_to_legacy = iconv_open(charset, "UTF-8");
+		vpninfo->ic_legacy_to_utf8 = iconv_open("UTF-8", charset);
+	} else {
+		vpninfo->ic_utf8_to_legacy = (iconv_t)-1;
+		vpninfo->ic_legacy_to_utf8 = (iconv_t)-1;
+	}
+#endif
 #ifndef _WIN32
 	vpninfo->tun_fd = -1;
 #endif
@@ -83,7 +100,8 @@ err:
 	return NULL;
 }
 
-int openconnect_set_reported_os(struct openconnect_info *vpninfo, const char *os)
+int openconnect_set_reported_os(struct openconnect_info *vpninfo,
+				const char *os)
 {
 	if (!os) {
 #if defined(__APPLE__)
@@ -107,18 +125,20 @@ int openconnect_set_reported_os(struct openconnect_info *vpninfo, const char *os
 	else
 		return -EINVAL;
 
-	vpninfo->platname = os;
+	STRDUP(vpninfo->platname, os);
 	return 0;
 }
 
-void openconnect_set_mobile_info(struct openconnect_info *vpninfo,
-				 char *mobile_platform_version,
-				 char *mobile_device_type,
-				 char *mobile_device_uniqueid)
+int openconnect_set_mobile_info(struct openconnect_info *vpninfo,
+				const char *mobile_platform_version,
+				const char *mobile_device_type,
+				const char *mobile_device_uniqueid)
 {
-	vpninfo->mobile_platform_version = mobile_platform_version;
-	vpninfo->mobile_device_type = mobile_device_type;
-	vpninfo->mobile_device_uniqueid = mobile_device_uniqueid;
+	STRDUP(vpninfo->mobile_platform_version, mobile_platform_version);
+	STRDUP(vpninfo->mobile_device_type, mobile_device_type);
+	STRDUP(vpninfo->mobile_device_uniqueid, mobile_device_uniqueid);
+
+	return 0;
 }
 
 static void free_optlist(struct oc_vpn_option *opt)
@@ -138,9 +158,17 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	openconnect_close_https(vpninfo, 1);
 	dtls_shutdown(vpninfo);
 	if (vpninfo->cmd_fd_write != -1) {
-		close(vpninfo->cmd_fd);
-		close(vpninfo->cmd_fd_write);
+		closesocket(vpninfo->cmd_fd);
+		closesocket(vpninfo->cmd_fd_write);
 	}
+
+#ifdef HAVE_ICONV
+	if (vpninfo->ic_utf8_to_legacy != (iconv_t)-1)
+		iconv_close(vpninfo->ic_utf8_to_legacy);
+
+	if (vpninfo->ic_legacy_to_utf8 != (iconv_t)-1)
+		iconv_close(vpninfo->ic_legacy_to_utf8);
+#endif
 #ifdef _WIN32
 	if (vpninfo->cmd_event)
 		CloseHandle(vpninfo->cmd_event);
@@ -150,11 +178,14 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 		CloseHandle(vpninfo->dtls_event);
 #endif
 	free(vpninfo->peer_addr);
+	free_optlist(vpninfo->csd_env);
+	free_optlist(vpninfo->script_env);
 	free_optlist(vpninfo->cookies);
 	free_optlist(vpninfo->cstp_options);
 	free_optlist(vpninfo->dtls_options);
 	cstp_free_splits(vpninfo);
 	free(vpninfo->hostname);
+	free(vpninfo->unique_hostname);
 	free(vpninfo->urlpath);
 	free(vpninfo->redirect_url);
 	free(vpninfo->cookie);
@@ -164,9 +195,14 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	free(vpninfo->proxy_pass);
 	free(vpninfo->vpnc_script);
 	free(vpninfo->cafile);
-	free(vpninfo->servercert);
 	free(vpninfo->ifname);
 	free(vpninfo->dtls_cipher);
+#ifdef OPENCONNECT_GNUTLS
+	gnutls_free(vpninfo->cstp_cipher); /* In OpenSSL this is const */
+#endif
+#ifdef DTLS_GNUTLS
+	gnutls_free(vpninfo->gnutls_dtls_cipher);
+#endif
 	free(vpninfo->dtls_addr);
 
 	if (vpninfo->csd_scriptname) {
@@ -182,6 +218,7 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	free(vpninfo->csd_starturl);
 	free(vpninfo->csd_waiturl);
 	free(vpninfo->csd_preurl);
+	free(vpninfo->platname);
 	if (vpninfo->opaque_srvdata)
 		xmlFreeNode(vpninfo->opaque_srvdata);
 	free(vpninfo->profile_url);
@@ -201,6 +238,7 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 #endif
 		vpninfo->peer_cert = NULL;
 	}
+	free(vpninfo->peer_cert_hash);
 	free(vpninfo->localname);
 	free(vpninfo->useragent);
 	free(vpninfo->authgroup);
@@ -211,31 +249,50 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 		stoken_destroy(vpninfo->stoken_ctx);
 #endif
 #ifdef HAVE_LIBOATH
-	if (vpninfo->oath_secret)
+	if (vpninfo->oath_secret) {
+#ifdef HAVE_LIBPSKC
+		if (vpninfo->pskc)
+			pskc_done(vpninfo->pskc);
+		else
+#endif /* HAVE_LIBPSKC */
+		free(vpninfo->oath_secret);
 		oath_done();
+	}
+#endif /* HAVE_LIBOATH */
+#ifdef HAVE_LIBPCSCLITE
+	if (vpninfo->token_mode == OC_TOKEN_MODE_YUBIOATH) {
+		SCardDisconnect(vpninfo->pcsc_card, SCARD_LEAVE_CARD);
+		SCardReleaseContext(vpninfo->pcsc_ctx);
+	}
+	memset(vpninfo->yubikey_pwhash, 0, sizeof(vpninfo->yubikey_pwhash));
+	free(vpninfo->yubikey_objname);
 #endif
-
 	/* These check strm->state so they are safe to call multiple times */
 	inflateEnd(&vpninfo->inflate_strm);
 	deflateEnd(&vpninfo->deflate_strm);
 
 	free(vpninfo->deflate_pkt);
+	free(vpninfo->tun_pkt);
 	free(vpninfo);
 }
 
-char *openconnect_get_hostname(struct openconnect_info *vpninfo)
+const char *openconnect_get_hostname(struct openconnect_info *vpninfo)
 {
 	return vpninfo->unique_hostname?:vpninfo->hostname;
 }
 
-void openconnect_set_hostname(struct openconnect_info *vpninfo, char *hostname)
+int openconnect_set_hostname(struct openconnect_info *vpninfo,
+			     const char *hostname)
 {
-	free(vpninfo->hostname);
-	vpninfo->hostname = hostname;
+	UTF8CHECK(hostname);
+
+	STRDUP(vpninfo->hostname, hostname);
 	free(vpninfo->unique_hostname);
 	vpninfo->unique_hostname = NULL;
 	free(vpninfo->peer_addr);
 	vpninfo->peer_addr = NULL;
+
+	return 0;
 }
 
 char *openconnect_get_urlpath(struct openconnect_info *vpninfo)
@@ -243,12 +300,17 @@ char *openconnect_get_urlpath(struct openconnect_info *vpninfo)
 	return vpninfo->urlpath;
 }
 
-void openconnect_set_urlpath(struct openconnect_info *vpninfo, char *urlpath)
+int openconnect_set_urlpath(struct openconnect_info *vpninfo,
+			    const char *urlpath)
 {
-	vpninfo->urlpath = urlpath;
+	UTF8CHECK(urlpath);
+
+	STRDUP(vpninfo->urlpath, urlpath);
+	return 0;
 }
 
-void openconnect_set_xmlsha1(struct openconnect_info *vpninfo, const char *xmlsha1, int size)
+void openconnect_set_xmlsha1(struct openconnect_info *vpninfo,
+			     const char *xmlsha1, int size)
 {
 	if (size != sizeof(vpninfo->xmlsha1))
 		return;
@@ -256,14 +318,17 @@ void openconnect_set_xmlsha1(struct openconnect_info *vpninfo, const char *xmlsh
 	memcpy(&vpninfo->xmlsha1, xmlsha1, size);
 }
 
-void openconnect_set_cafile(struct openconnect_info *vpninfo, char *cafile)
+int openconnect_set_cafile(struct openconnect_info *vpninfo, const char *cafile)
 {
-	vpninfo->cafile = cafile;
+	UTF8CHECK(cafile);
+
+	STRDUP(vpninfo->cafile, cafile);
+	return 0;
 }
 
-void openconnect_set_server_cert_sha1(struct openconnect_info *vpninfo, char *servercert)
+void openconnect_set_system_trust(struct openconnect_info *vpninfo, unsigned val)
 {
-	vpninfo->servercert = servercert;
+	vpninfo->no_system_trust = !val;
 }
 
 const char *openconnect_get_ifname(struct openconnect_info *vpninfo)
@@ -299,11 +364,14 @@ int openconnect_get_ip_info(struct openconnect_info *vpninfo,
 	return 0;
 }
 
-void openconnect_setup_csd(struct openconnect_info *vpninfo, uid_t uid, int silent, char *wrapper)
+int openconnect_setup_csd(struct openconnect_info *vpninfo, uid_t uid,
+			  int silent, const char *wrapper)
 {
 	vpninfo->uid_csd = uid;
 	vpninfo->uid_csd_given = silent ? 2 : 1;
-	vpninfo->csd_wrapper = wrapper;
+	STRDUP(vpninfo->csd_wrapper, wrapper);
+
+	return 0;
 }
 
 void openconnect_set_xmlpost(struct openconnect_info *vpninfo, int enable)
@@ -311,18 +379,25 @@ void openconnect_set_xmlpost(struct openconnect_info *vpninfo, int enable)
 	vpninfo->xmlpost = enable;
 }
 
-void openconnect_set_client_cert(struct openconnect_info *vpninfo, char *cert, char *sslkey)
+int openconnect_set_client_cert(struct openconnect_info *vpninfo,
+				const char *cert, const char *sslkey)
 {
-	vpninfo->cert = cert;
-	if (sslkey)
-		vpninfo->sslkey = sslkey;
-	else
-		vpninfo->sslkey = cert;
-}
+	UTF8CHECK(cert);
+	UTF8CHECK(sslkey);
 
-OPENCONNECT_X509 *openconnect_get_peer_cert(struct openconnect_info *vpninfo)
-{
-	return vpninfo->peer_cert;
+	/* Avoid freeing it twice if it's the same */
+	if (vpninfo->sslkey == vpninfo->cert)
+		vpninfo->sslkey = NULL;
+
+	STRDUP(vpninfo->cert, cert);
+
+	if (sslkey) {
+		STRDUP(vpninfo->sslkey, sslkey);
+	} else {
+		vpninfo->sslkey = vpninfo->cert;
+	}
+
+	return 0;
 }
 
 int openconnect_get_port(struct openconnect_info *vpninfo)
@@ -330,7 +405,7 @@ int openconnect_get_port(struct openconnect_info *vpninfo)
 	return vpninfo->port;
 }
 
-char *openconnect_get_cookie(struct openconnect_info *vpninfo)
+const char *openconnect_get_cookie(struct openconnect_info *vpninfo)
 {
 	return vpninfo->cookie;
 }
@@ -349,13 +424,16 @@ void openconnect_reset_ssl(struct openconnect_info *vpninfo)
 		free(vpninfo->peer_addr);
 		vpninfo->peer_addr = NULL;
 	}
+	openconnect_clear_cookies(vpninfo);
 }
 
-int openconnect_parse_url(struct openconnect_info *vpninfo, char *url)
+int openconnect_parse_url(struct openconnect_info *vpninfo, const char *url)
 {
 	char *scheme = NULL;
 	int ret;
 
+	UTF8CHECK(url);
+
 	openconnect_set_hostname(vpninfo, NULL);
 	free(vpninfo->urlpath);
 	vpninfo->urlpath = NULL;
@@ -379,13 +457,12 @@ int openconnect_parse_url(struct openconnect_info *vpninfo, char *url)
 }
 
 void openconnect_set_cert_expiry_warning(struct openconnect_info *vpninfo,
-					  int seconds)
+					 int seconds)
 {
 	vpninfo->cert_expire_warning = seconds;
 }
 
-void openconnect_set_pfs(struct openconnect_info *vpninfo,
-					  	unsigned val)
+void openconnect_set_pfs(struct openconnect_info *vpninfo, unsigned val)
 {
 	vpninfo->pfs = val;
 }
@@ -395,21 +472,28 @@ void openconnect_set_cancel_fd(struct openconnect_info *vpninfo, int fd)
 	vpninfo->cmd_fd = fd;
 }
 
-int openconnect_setup_cmd_pipe(struct openconnect_info *vpninfo)
+#ifdef _WIN32
+# define CMD_PIPE_ERR INVALID_SOCKET
+#else
+# define CMD_PIPE_ERR -EIO
+#endif
+
+OPENCONNECT_CMD_SOCKET openconnect_setup_cmd_pipe(struct openconnect_info *vpninfo)
 {
-	int pipefd[2];
+	OPENCONNECT_CMD_SOCKET pipefd[2];
 
 #ifdef _WIN32
 	if (dumb_socketpair(pipefd, 0))
-		return -EIO;
+		return CMD_PIPE_ERR;
 #else
 	if (pipe(pipefd) < 0)
-		return -EIO;
+		return CMD_PIPE_ERR;
 #endif
+
 	if (set_sock_nonblock(pipefd[0]) || set_sock_nonblock(pipefd[1])) {
-		close(pipefd[0]);
-		close(pipefd[1]);
-		return -EIO;
+		closesocket(pipefd[0]);
+		closesocket(pipefd[1]);
+		return CMD_PIPE_ERR;
 	}
 	vpninfo->cmd_fd = pipefd[0];
 	vpninfo->cmd_fd_write = pipefd[1];
@@ -469,100 +553,34 @@ int openconnect_has_oath_support(void)
 #endif
 }
 
-static int set_libstoken_mode(struct openconnect_info *vpninfo,
-			      const char *token_str)
+int openconnect_has_yubioath_support(void)
 {
-#ifdef HAVE_LIBSTOKEN
-	int ret;
-
-	if (!vpninfo->stoken_ctx) {
-		vpninfo->stoken_ctx = stoken_new();
-		if (!vpninfo->stoken_ctx)
-			return -EIO;
-	}
-
-	ret = token_str ?
-	      stoken_import_string(vpninfo->stoken_ctx, token_str) :
-	      stoken_import_rcfile(vpninfo->stoken_ctx, NULL);
-	if (ret)
-		return ret;
-
-	vpninfo->token_mode = OC_TOKEN_MODE_STOKEN;
-	return 0;
+#ifdef HAVE_LIBPCSCLITE
+	return 1;
 #else
-	return -EOPNOTSUPP;
+	return 0;
 #endif
 }
 
-static int set_totp_mode(struct openconnect_info *vpninfo,
-			 const char *token_str)
+int openconnect_has_system_key_support(void)
 {
-#ifdef HAVE_LIBOATH
-	int ret;
-
-	ret = oath_init();
-	if (ret != OATH_OK)
-		return -EIO;
-
-	if (strncasecmp(token_str, "base32:", strlen("base32:")) == 0) {
-		ret = oath_base32_decode(token_str + strlen("base32:"),
-					 strlen(token_str) - strlen("base32:"),
-					 &vpninfo->oath_secret,
-					 &vpninfo->oath_secret_len);
-		if (ret != OATH_OK)
-			return -EINVAL;
-	} else {
-		vpninfo->oath_secret = strdup(token_str);
-		vpninfo->oath_secret_len = strlen(token_str);
-	}
-
-	vpninfo->token_mode = OC_TOKEN_MODE_TOTP;
-	return 0;
+#ifdef HAVE_GNUTLS_SYSTEM_KEYS
+	return 1;
 #else
-	return -EOPNOTSUPP;
+	return 0;
 #endif
 }
 
-static int set_hotp_mode(struct openconnect_info *vpninfo,
-			 const char *token_str)
+int openconnect_set_token_callbacks(struct openconnect_info *vpninfo,
+				    void *tokdata,
+				    openconnect_lock_token_vfn lock,
+				    openconnect_unlock_token_vfn unlock)
 {
-#ifdef HAVE_LIBOATH
-	int ret, toklen;
-	char *p;
+	vpninfo->lock_token = lock;
+	vpninfo->unlock_token = unlock;
+	vpninfo->tok_cbdata = tokdata;
 
-	ret = oath_init();
-	if (ret != OATH_OK)
-		return -EIO;
-
-	toklen = strlen(token_str);
-	p = strrchr(token_str, ',');
-	if (p) {
-		long counter;
-		toklen = p - token_str;
-		p++;
-		counter = strtol(p, &p, 0);
-		if (counter < 0 || *p)
-			return -EINVAL;
-		vpninfo->token_time = counter;
-	}
-
-	if (strncasecmp(token_str, "base32:", strlen("base32:")) == 0) {
-		ret = oath_base32_decode(token_str + strlen("base32:"),
-					 toklen - strlen("base32:"),
-					 &vpninfo->oath_secret,
-					 &vpninfo->oath_secret_len);
-		if (ret != OATH_OK)
-			return -EINVAL;
-	} else {
-		vpninfo->oath_secret = strdup(token_str);
-		vpninfo->oath_secret_len = toklen;
-	}
-
-	vpninfo->token_mode = OC_TOKEN_MODE_HOTP;
 	return 0;
-#else
-	return -EOPNOTSUPP;
-#endif
 }
 
 /*
@@ -572,6 +590,7 @@ static int set_hotp_mode(struct openconnect_info *vpninfo,
  * read the token data from ~/.stokenrc.
  *
  * Return value:
+ *  = -EILSEQ, if token_str is not valid UTF-8
  *  = -EOPNOTSUPP, if the underlying library (libstoken, liboath) is not
  *                 available or an invalid token_mode was provided
  *  = -EINVAL, if the token string is invalid (token_str was provided)
@@ -586,19 +605,27 @@ int openconnect_set_token_mode(struct openconnect_info *vpninfo,
 {
 	vpninfo->token_mode = OC_TOKEN_MODE_NONE;
 
+	UTF8CHECK(token_str);
+
 	switch (token_mode) {
 	case OC_TOKEN_MODE_NONE:
 		return 0;
 
+#ifdef HAVE_LIBSTOKEN
 	case OC_TOKEN_MODE_STOKEN:
 		return set_libstoken_mode(vpninfo, token_str);
-
+#endif
+#ifdef HAVE_LIBOATH
 	case OC_TOKEN_MODE_TOTP:
 		return set_totp_mode(vpninfo, token_str);
 
 	case OC_TOKEN_MODE_HOTP:
 		return set_hotp_mode(vpninfo, token_str);
-
+#endif
+#ifdef HAVE_LIBPCSCLITE
+	case OC_TOKEN_MODE_YUBIOATH:
+		return set_yubikey_mode(vpninfo, token_str);
+#endif
 	default:
 		return -EOPNOTSUPP;
 	}
@@ -613,6 +640,7 @@ int openconnect_set_token_mode(struct openconnect_info *vpninfo,
  * DEPRECATED: use openconnect_set_stoken_mode() instead.
  *
  * Return value:
+ *  = -EILSEQ, if token_str is not valid UTF-8
  *  = -EOPNOTSUPP, if libstoken is not available
  *  = -EINVAL, if the token string is invalid (token_str was provided)
  *  = -ENOENT, if ~/.stokenrc is missing (token_str was NULL)
@@ -643,22 +671,131 @@ void openconnect_set_stats_handler(struct openconnect_info *vpninfo,
 }
 
 /* Set up a traditional OS-based tunnel device, optionally specified in 'ifname'. */
-int openconnect_setup_tun_device(struct openconnect_info *vpninfo, char *vpnc_script, char *ifname)
+int openconnect_setup_tun_device(struct openconnect_info *vpninfo,
+				 const char *vpnc_script, const char *ifname)
 {
 	intptr_t tun_fd;
+	char *legacy_ifname;
 
-	vpninfo->vpnc_script = vpnc_script;
-	vpninfo->ifname = ifname;
+	UTF8CHECK(vpnc_script);
+	UTF8CHECK(ifname);
 
-	set_script_env(vpninfo);
+	STRDUP(vpninfo->vpnc_script, vpnc_script);
+	STRDUP(vpninfo->ifname, ifname);
+
+	prepare_script_env(vpninfo);
 	script_config_tun(vpninfo, "pre-init");
 
 	tun_fd = os_setup_tun(vpninfo);
 	if (tun_fd < 0)
 		return tun_fd;
 
-	setenv("TUNDEV", vpninfo->ifname, 1);
+	legacy_ifname = openconnect_utf8_to_legacy(vpninfo, vpninfo->ifname);
+	script_setenv(vpninfo, "TUNDEV", legacy_ifname, 0);
+	if (legacy_ifname != vpninfo->ifname)
+		free(legacy_ifname);
 	script_config_tun(vpninfo, "connect");
 
 	return openconnect_setup_tun_fd(vpninfo, tun_fd);
 }
+
+const char *openconnect_get_dtls_cipher(struct openconnect_info *vpninfo)
+{
+#if defined(DTLS_GNUTLS)
+	if (vpninfo->dtls_state != DTLS_CONNECTED) {
+		gnutls_free(vpninfo->gnutls_dtls_cipher);
+		vpninfo->gnutls_dtls_cipher = NULL;
+		return NULL;
+	}
+	/* in DTLS rehandshakes don't switch the ciphersuite as only
+	 * one is enabled. */
+	if (vpninfo->gnutls_dtls_cipher == NULL)
+		vpninfo->gnutls_dtls_cipher = get_gnutls_cipher(vpninfo->dtls_ssl);
+	return vpninfo->gnutls_dtls_cipher;
+#else
+	return vpninfo->dtls_cipher;
+#endif
+}
+
+int openconnect_set_csd_environ(struct openconnect_info *vpninfo,
+				const char *name, const char *value)
+{
+	struct oc_vpn_option *p;
+
+	if (!name) {
+		free_optlist(vpninfo->csd_env);
+		vpninfo->csd_env = NULL;
+		return 0;
+	}
+	for (p = vpninfo->csd_env; p; p = p->next) {
+		if (!strcmp(name, p->option)) {
+			char *valdup = strdup(value);
+			if (!valdup)
+				return -ENOMEM;
+			free(p->value);
+			p->value = valdup;
+			return 0;
+		}
+	}
+	p = malloc(sizeof(*p));
+	if (!p)
+		return -ENOMEM;
+	p->option = strdup(name);
+	if (!p->option) {
+		free(p);
+		return -ENOMEM;
+	}
+	p->value = strdup(value);
+	if (!p->value) {
+		free(p->option);
+		free(p);
+		return -ENOMEM;
+	}
+	p->next = vpninfo->csd_env;
+	vpninfo->csd_env = p;
+	return 0;
+}
+
+int openconnect_check_peer_cert_hash(struct openconnect_info *vpninfo,
+				     const char *old_hash)
+{
+	char sha1_text[41];
+	const char *fingerprint;
+
+	if (strchr(old_hash, ':')) {
+		fingerprint = openconnect_get_peer_cert_hash(vpninfo);
+		if (!fingerprint)
+			return -EIO;
+	} else {
+		unsigned char *cert;
+		int len, i;
+		unsigned char sha1_bin[SHA1_SIZE];
+
+		len = openconnect_get_peer_cert_DER(vpninfo, &cert);
+		if (len < 0)
+			return len;
+
+		if (openconnect_sha1(sha1_bin, cert, len))
+			return -EIO;
+
+		for (i = 0; i < sizeof(sha1_bin); i++)
+			sprintf(&sha1_text[i*2], "%02x", sha1_bin[i]);
+
+		fingerprint = sha1_text;
+	}
+
+	if (strcasecmp(old_hash, fingerprint))
+		return 1;
+
+	return 0;
+}
+
+const char *openconnect_get_cstp_cipher(struct openconnect_info *vpninfo)
+{
+	return vpninfo->cstp_cipher;
+}
+
+const char *openconnect_get_peer_cert_hash(struct openconnect_info *vpninfo)
+{
+	return vpninfo->peer_cert_hash;
+}
