@@ -38,6 +38,7 @@
 #include <sys/types.h>
 #include <getopt.h>
 #include <time.h>
+#include <locale.h>
 
 #ifdef LIBPROXY_HDR
 #include LIBPROXY_HDR
@@ -55,13 +56,17 @@
 #include <termios.h>
 #endif
 
+#ifdef HAVE_NL_LANGINFO
+#include <langinfo.h>
+
+static const char *legacy_charset;
+#endif
+
 static int write_new_config(void *_vpninfo,
-			    char *buf, int buflen);
-static void write_progress(void *_vpninfo,
-			   int level, const char *fmt, ...);
-static int validate_peer_cert(void *_vpninfo,
-			      OPENCONNECT_X509 *peer_cert,
-			      const char *reason);
+			    const char *buf, int buflen);
+static void __attribute__ ((format(printf, 3, 4)))
+    write_progress(void *_vpninfo, int level, const char *fmt, ...);
+static int validate_peer_cert(void *_vpninfo, const char *reason);
 static int process_auth_form_cb(void *_vpninfo,
 				struct oc_auth_form *form);
 static void init_token(struct openconnect_info *vpninfo,
@@ -81,6 +86,9 @@ static int nocertcheck;
 static int non_inter;
 static int cookieonly;
 
+static char *token_filename;
+static char *server_cert = NULL;
+
 static char *username;
 static char *password;
 static char *authgroup;
@@ -91,7 +99,8 @@ static int sig_cmd_fd;
 
 #ifdef __ANDROID__
 #include <android/log.h>
-static void syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
+static void __attribute__ ((format(printf, 3, 4)))
+    syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
 {
 	static int l[4] = {
 		ANDROID_LOG_ERROR,	/* PRG_ERR   */
@@ -120,7 +129,8 @@ static void syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
  */
 #else /* !__ANDROID__ && !_WIN32 */
 #include <syslog.h>
-static void syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
+static void  __attribute__ ((format(printf, 3, 4)))
+    syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
 {
 	int priority = level ? LOG_INFO : LOG_NOTICE;
 	va_list args;
@@ -152,6 +162,7 @@ enum {
 	OPT_NO_CERT_CHECK,
 	OPT_NO_DTLS,
 	OPT_NO_HTTP_KEEPALIVE,
+	OPT_NO_SYSTEM_TRUST,
 	OPT_NO_PASSWD,
 	OPT_NO_PROXY,
 	OPT_NO_XMLPOST,
@@ -242,9 +253,292 @@ static struct option long_options[] = {
 	OPTION("os", 1, OPT_OS),
 	OPTION("no-xmlpost", 0, OPT_NO_XMLPOST),
 	OPTION("dump-http-traffic", 0, OPT_DUMP_HTTP),
+	OPTION("no-system-trust", 0, OPT_NO_SYSTEM_TRUST),
 	OPTION(NULL, 0, 0)
 };
 
+#ifdef _WIN32
+static int __attribute__ ((format(printf, 2, 0)))
+    vfprintf_utf8(FILE *f, const char *fmt, va_list args)
+{
+	HANDLE h = GetStdHandle(f == stdout ? STD_OUTPUT_HANDLE : STD_ERROR_HANDLE);
+	wchar_t wbuf[1024];
+	char buf[1024];
+	int chars, wchars;
+
+	buf[sizeof(buf) - 1] = 0;
+	chars = _vsnprintf(buf, sizeof(buf) - 1, fmt, args);
+	wchars = MultiByteToWideChar(CP_UTF8, 0, buf, -1, wbuf, sizeof(wbuf)/2);
+	WriteConsoleW(h, wbuf, wchars, NULL, NULL);
+
+	return chars;
+}
+
+static int __attribute__ ((format(printf, 2, 3)))
+    fprintf_utf8(FILE *f, const char *fmt, ...)
+{
+	va_list args;
+	int ret;
+
+	va_start(args, fmt);
+	ret = vfprintf_utf8(f, fmt, args);
+	va_end(args);
+
+	return ret;
+}
+
+static wchar_t **argv_w;
+
+/* This isn't so much "convert" the arg to UTF-8, as go grubbing
+ * around in the real UTF-16 command line and find the corresponding
+ * argument *there*, and convert *that* to UTF-8. Ick. But the
+ * alternative is to implement wgetopt(), and that's even more horrid. */
+static char *convert_arg_to_utf8(char **argv, char *arg)
+{
+	char *utf8;
+	int chars;
+	int offset;
+
+	if (!argv_w) {
+		int argc_w;
+
+		argv_w = CommandLineToArgvW(GetCommandLineW(), &argc_w);
+		if (!argv_w) {
+			char *errstr = openconnect__win32_strerror(GetLastError());
+			fprintf(stderr, _("CommandLineToArgvW() failed: %s\n"),
+				errstr);
+			free(errstr);
+			exit(1);
+		}
+	}
+
+	offset = arg - argv[optind - 1];
+
+	/* Sanity check */
+	if (offset < 0 || offset >= strlen(argv[optind - 1]) ||
+	    (offset && (argv[optind - 1][offset-1] != '=' ||
+			argv_w[optind - 1][offset - 1] != '='))) {
+		fprintf(stderr, _("Fatal error in command line handling\n"));
+		exit(1);
+	}
+
+	chars = WideCharToMultiByte(CP_UTF8, 0, argv_w[optind-1] + offset, -1,
+				    NULL, 0, NULL, NULL);
+	utf8 = malloc(chars);
+	if (!utf8)
+		return arg;
+
+	WideCharToMultiByte(CP_UTF8, 0, argv_w[optind-1] + offset, -1, utf8,
+			    chars, NULL, NULL);
+	return utf8;
+}
+
+#undef fprintf
+#undef vfprintf
+#define fprintf fprintf_utf8
+#define vfprintf vfprintf_utf8
+#define is_arg_utf8(str) (0)
+
+static void read_stdin(char **string, int hidden)
+{
+	CONSOLE_READCONSOLE_CONTROL rcc = { sizeof(rcc), 0, 13, 0 };
+	HANDLE stdinh = GetStdHandle(STD_INPUT_HANDLE);
+	DWORD cmode, nr_read;
+	wchar_t wbuf[1024];
+	char *buf;
+
+	if (hidden) {
+		GetConsoleMode(stdinh, &cmode);
+		SetConsoleMode(stdinh, cmode & (~ENABLE_ECHO_INPUT));
+	}
+
+	if (!ReadConsoleW(stdinh, wbuf, sizeof(wbuf)/2, &nr_read, &rcc)) {
+		char *errstr = openconnect__win32_strerror(GetLastError());
+		fprintf(stderr, _("ReadConsole() failed: %s\n"), errstr);
+		free(errstr);
+		goto out;
+	}
+
+	if (nr_read >= 2 && wbuf[nr_read - 1] == 10 && wbuf[nr_read - 2] == 13) {
+		wbuf[nr_read - 2] = 0;
+		nr_read -= 2;
+	}
+
+	nr_read = WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, NULL, 0, NULL, NULL);
+	if (!nr_read) {
+		char *errstr = openconnect__win32_strerror(GetLastError());
+		fprintf(stderr, _("Error converting console input: %s\n"),
+			errstr);
+		free(errstr);
+		goto out;
+	}
+	buf = malloc(nr_read);
+	if (!buf) {
+		fprintf(stderr, _("Allocation failure for string from stdin\n"));
+		exit(1);
+	}
+
+	if (!WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, buf, nr_read, NULL, NULL)) {
+		char *errstr = openconnect__win32_strerror(GetLastError());
+		fprintf(stderr, _("Error converting console input: %s\n"),
+			errstr);
+		free(errstr);
+		free(buf);
+		goto out;
+	}
+
+	*string = buf;
+
+out:
+	if (hidden) {
+		SetConsoleMode(stdinh, cmode);
+		fprintf(stderr, "\n");
+	}
+}
+
+#elif defined(HAVE_ICONV)
+#include <iconv.h>
+
+static int is_ascii(char *str)
+{
+	while (str && *str) {
+		if ((unsigned char)*str > 0x7f)
+			return 0;
+		str++;
+	}
+
+	return 1;
+}
+
+static int __attribute__ ((format(printf, 2, 0)))
+    vfprintf_utf8(FILE *f, const char *fmt, va_list args)
+{
+	char *utf8_str;
+	iconv_t ic;
+	int ret;
+	char outbuf[80];
+	ICONV_CONST char *ic_in;
+	char *ic_out;
+	size_t insize, outsize;
+
+	if (!legacy_charset)
+		return vfprintf(f, fmt, args);
+
+	ret = vasprintf(&utf8_str, fmt, args);
+	if (ret < 0)
+		return -1;
+
+	if (is_ascii(utf8_str))
+		return fwrite(utf8_str, 1, strlen(utf8_str), f);
+
+	ic = iconv_open(legacy_charset, "UTF-8");
+	if (ic == (iconv_t) -1) {
+		/* Better than nothing... */
+		ret = fprintf(f, "%s", utf8_str);
+		free(utf8_str);
+		return ret;
+	}
+
+	ic_in = utf8_str;
+	insize = strlen(utf8_str);
+	ret = 0;
+
+	while (insize) {
+		ic_out = outbuf;
+		outsize = sizeof(outbuf) - 1;
+
+		if (iconv(ic, &ic_in, &insize, &ic_out, &outsize) == (size_t)-1) {
+			if (errno == EILSEQ) {
+				do {
+					ic_in++;
+					insize--;
+				} while (insize && (ic_in[0] & 0xc0) == 0x80);
+				ic_out[0] = '?';
+				outsize--;
+			} else if (errno != E2BIG)
+				break;
+		}
+		ret += fwrite(outbuf, 1, sizeof(outbuf) - 1 - outsize, f);
+	}
+
+	iconv_close(ic);
+
+	return ret;
+}
+
+static int __attribute__ ((format(printf, 2, 3)))
+    fprintf_utf8(FILE *f, const char *fmt, ...)
+{
+	va_list args;
+	int ret;
+
+	va_start(args, fmt);
+	ret = vfprintf_utf8(f, fmt, args);
+	va_end(args);
+
+	return ret;
+}
+
+static char *convert_to_utf8(char *legacy, int free_it)
+{
+	char *utf8_str;
+	iconv_t ic;
+	ICONV_CONST char *ic_in;
+	char *ic_out;
+	size_t insize, outsize;
+
+	if (!legacy_charset || is_ascii(legacy))
+		return legacy;
+
+	ic = iconv_open("UTF-8", legacy_charset);
+	if (ic == (iconv_t) -1)
+		return legacy;
+
+	insize = strlen(legacy) + 1;
+	ic_in = legacy;
+
+	outsize = insize;
+	ic_out = utf8_str = malloc(outsize);
+	if (!utf8_str) {
+	enomem:
+		iconv_close(ic);
+		return legacy;
+	}
+
+	while (insize) {
+		if (iconv(ic, &ic_in, &insize, &ic_out, &outsize) == (size_t)-1) {
+			if (errno == E2BIG) {
+				int outlen = ic_out - utf8_str;
+				realloc_inplace(utf8_str, outlen + 10);
+				if (!utf8_str)
+					goto enomem;
+				ic_out = utf8_str + outlen;
+				outsize = 10;
+			} else {
+				/* Should never happen */
+				perror("iconv");
+				free(utf8_str);
+				goto enomem;
+			}
+		}
+	}
+
+	iconv_close(ic);
+	if (free_it)
+		free(legacy);
+	return utf8_str;
+}
+
+#define fprintf fprintf_utf8
+#define vfprintf vfprintf_utf8
+#define convert_arg_to_utf8(av, l) convert_to_utf8((l), 0)
+#define is_arg_utf8(a) (!legacy_charset || is_ascii(a))
+#else
+#define convert_to_utf8(l,f) (l)
+#define convert_arg_to_utf8(av, l) (l)
+#define is_arg_utf8(a) (1)
+#endif
+
 static void helpmessage(void)
 {
 	printf(_("For assistance with OpenConnect, please see the web page at\n"
@@ -287,6 +581,14 @@ static void print_build_opts(void)
 		printf("%sTOTP software token", sep);
 		sep = comma;
 	}
+	if (openconnect_has_yubioath_support()) {
+		printf("%sYubikey OATH", sep);
+		sep = comma;
+	}
+	if (openconnect_has_system_key_support()) {
+		printf("%sSystem keys", sep);
+		sep = comma;
+	}
 
 #ifdef HAVE_DTLS
 	printf("%sDTLS", sep);
@@ -301,24 +603,41 @@ static void print_build_opts(void)
 
 #ifndef _WIN32
 static const char default_vpncscript[] = DEFAULT_VPNCSCRIPT;
-static void disable_echo(void)
+static void read_stdin(char **string, int hidden)
 {
-	struct termios t;
+	char *c, *buf = malloc(1025);
 	int fd = fileno(stdin);
-
-	tcgetattr(fd, &t);
-	t.c_lflag &= ~ECHO;
-	tcsetattr(fd, TCSANOW, &t);
-}
-
-static void restore_echo(void)
-{
 	struct termios t;
-	int fd = fileno(stdin);
 
-	tcgetattr(fd, &t);
-	t.c_lflag |= ECHO;
-	tcsetattr(fd, TCSANOW, &t);
+	if (!buf) {
+		fprintf(stderr, _("Allocation failure for string from stdin\n"));
+		exit(1);
+	}
+
+	if (hidden) {
+		tcgetattr(fd, &t);
+		t.c_lflag &= ~ECHO;
+		tcsetattr(fd, TCSANOW, &t);
+	}
+
+	buf = fgets(buf, 1025, stdin);
+
+	if (hidden) {
+		t.c_lflag |= ECHO;
+		tcsetattr(fd, TCSANOW, &t);
+		fprintf(stderr, "\n");
+	}
+
+	if (!buf) {
+		perror(_("fgets (stdin)"));
+		exit(1);
+	}
+
+	c = strchr(buf, '\n');
+	if (c)
+		*c = 0;
+
+	*string = convert_to_utf8(buf, 1);
 }
 
 static void handle_signal(int sig)
@@ -353,7 +672,7 @@ static void set_default_vpncscript(void)
 				_pgmptr);
 			exit(1);
 		}
-		if (asprintf((char **)&default_vpncscript, "cscript %.*s%s",
+		if (asprintf((char **)&default_vpncscript, "%.*s%s",
 			     (c - _pgmptr + 1), _pgmptr, DEFAULT_VPNCSCRIPT) < 0) {
 			fprintf(stderr, _("Allocation for vpnc-script path failed\n"));
 			exit(1);
@@ -362,33 +681,6 @@ static void set_default_vpncscript(void)
 		default_vpncscript = "cscript " DEFAULT_VPNCSCRIPT;
 	}
 }
-
-static HANDLE hconin = INVALID_HANDLE_VALUE;
-static DWORD cmode;
-
-static void disable_echo(void)
-{
-	hconin = CreateFile("CONIN$", GENERIC_READ | GENERIC_WRITE,
-			    FILE_SHARE_READ, NULL, OPEN_EXISTING,
-			    FILE_ATTRIBUTE_NORMAL, NULL);
-	if (hconin == INVALID_HANDLE_VALUE)
-		return;
-	GetConsoleMode(hconin, &cmode);
-	if (!SetConsoleMode(hconin, cmode & (~ENABLE_ECHO_INPUT))) {
-		CloseHandle(hconin);
-		hconin = INVALID_HANDLE_VALUE;
-	}
-}
-
-static void restore_echo(void)
-{
-	if (hconin == INVALID_HANDLE_VALUE)
-		return;
-
-	SetConsoleMode(hconin, cmode);
-	CloseHandle(hconin);
-	hconin = INVALID_HANDLE_VALUE;
-}
 #endif
 
 static void usage(void)
@@ -456,6 +748,7 @@ static void usage(void)
 	printf("      --no-http-keepalive         %s\n", _("Disable HTTP connection re-use"));
 	printf("      --no-passwd                 %s\n", _("Disable password/SecurID authentication"));
 	printf("      --no-cert-check             %s\n", _("Do not require server SSL cert to be valid"));
+	printf("      --no-system-trust           %s\n", _("Disable default system certificate authorities"));
 	printf("      --no-xmlpost                %s\n", _("Do not attempt XML POST authentication"));
 	printf("      --non-inter                 %s\n", _("Do not expect user input; exit if it is required"));
 	printf("      --passwd-on-stdin           %s\n", _("Read password from standard input"));
@@ -466,6 +759,9 @@ static void usage(void)
 #endif
 #ifndef HAVE_LIBOATH
 	printf("                                  %s\n", _("(NOTE: liboath (TOTP,HOTP) disabled in this build)"));
+#endif
+#ifndef HAVE_LIBPCSCLITE
+	printf("                                  %s\n", _("(NOTE: Yubikey OATH disabled in this build)"));
 #endif
 	printf("      --reconnect-timeout         %s\n", _("Connection retry timeout in seconds"));
 	printf("      --servercert=FINGERPRINT    %s\n", _("Server's certificate SHA1 fingerprint"));
@@ -479,55 +775,17 @@ static void usage(void)
 }
 
 
-static void read_stdin(char **string, int hidden)
-{
-	char *c = malloc(1025), *ret;
-
-	if (!c) {
-		fprintf(stderr, _("Allocation failure for string from stdin\n"));
-		exit(1);
-	}
-
-	if (hidden) {
-		disable_echo();
-		ret = fgets(c, 1025, stdin);
-		restore_echo();
-		fprintf(stderr, "\n");
-	} else
-		ret = fgets(c, 1025, stdin);
-
-	if (!ret) {
-		perror(_("fgets (stdin)"));
-		exit(1);
-	}
-
-	*string = c;
-
-	c = strchr(*string, '\n');
-	if (c)
-		*c = 0;
-}
-
 static FILE *config_file = NULL;
 static int config_line_num = 0;
 
-/* There are three ways to handle config_arg:
- *
- * 1. We only care about it transiently and it can be lost entirely
- *    (e.g. vpninfo->reconnect_timeout = atoi(config_arg);
- * 2. We need to keep it, but it's a static string and will never be freed
- *    so when it's part of argv[] we can use it in place, but when it comes
- *    from a file we have to strdup() because otherwise it'll be overwritten.
- *    For this we use the keep_config_arg() macro below.
- * 3. It may be freed during normal operation, so we have to use strdup()
- *    even when it's an option from argv[]. (e.g. vpninfo->cert_password).
- *    For this we use the xstrdup() function below.
- */
-#define keep_config_arg() (config_file && config_arg ? strdup(config_arg) : config_arg)
-
 static char *xstrdup(const char *arg)
 {
-	char *ret = strdup(arg);
+	char *ret;
+
+	if (!arg)
+		return NULL;
+
+	ret = strdup(arg);
 
 	if (!ret) {
 		fprintf(stderr, _("Failed to allocate string\n"));
@@ -536,6 +794,28 @@ static char *xstrdup(const char *arg)
 	return ret;
 }
 
+/* There are three ways to handle config_arg:
+ *
+ * 1. We only care about it transiently and it can be lost entirely
+ *    (e.g. vpninfo->reconnect_timeout = atoi(config_arg);
+ * 2. We need to keep it, but it's a static string and will never be freed
+ *    so when it's part of argv[] we can use it in place (unless it needs
+ *    converting to UTF-8), but when it comes from a file we have to strdup()
+ *    because otherwise it'll be overwritten.
+ *    For this we use the keep_config_arg() macro below.
+ * 3. It may be freed during normal operation, so we have to use strdup()
+ *    or convert_arg_to_utf8() even when it's an option from argv[].
+ *    (e.g. vpninfo->cert_password).
+ *    For this we use the dup_config_arg() macro below.
+ */
+
+#define keep_config_arg() \
+	(config_file ? xstrdup(config_arg) : convert_arg_to_utf8(argv, config_arg))
+
+#define dup_config_arg() \
+	((config_file || is_arg_utf8(config_arg)) ? xstrdup(config_arg) : \
+	 convert_arg_to_utf8(argv, config_arg))
+
 static int next_option(int argc, char **argv, char **config_arg)
 {
 	/* These get re-used */
@@ -649,10 +929,14 @@ int main(int argc, char **argv)
 	int use_dtls = 1;
 	FILE *fp = NULL;
 	char *config_arg;
+	char *config_filename;
 	char *token_str = NULL;
 	oc_token_mode_t token_mode = OC_TOKEN_MODE_NONE;
 	int reconnect_timeout = 300;
 	int ret;
+#ifdef HAVE_NL_LANGINFO
+	char *charset;
+#endif
 #ifndef _WIN32
 	struct sigaction sa;
 	struct utsname utsbuf;
@@ -663,9 +947,24 @@ int main(int argc, char **argv)
 
 #ifdef ENABLE_NLS
 	bindtextdomain("openconnect", LOCALEDIR);
-	setlocale(LC_ALL, "");
 #endif
 
+	setlocale(LC_ALL, "");
+
+#ifdef HAVE_NL_LANGINFO
+	charset = nl_langinfo(CODESET);
+	if (charset && strcmp(charset, "UTF-8"))
+		legacy_charset = strdup(charset);
+
+#ifndef HAVE_ICONV
+	if (legacy_charset)
+		fprintf(stderr,
+			_("WARNING: This version of openconnect was built without iconv\n"
+			  "         support but you appear to be using the legacy character\n"
+			  "         set \"%s\". Expect strangeness.\n"), legacy_charset);
+#endif /* !HAVE_ICONV */
+#endif /* HAVE_NL_LANGINFO */
+
 	if (strcmp(openconnect_version_str, openconnect_binary_version)) {
 		fprintf(stderr, _("WARNING: This version of openconnect is %s but\n"
 				  "         the libopenconnect library is %s\n"),
@@ -745,7 +1044,10 @@ int main(int argc, char **argv)
 				fprintf(stderr, _("Cannot use 'config' option inside config file\n"));
 				exit(1);
 			}
-			config_file = fopen(config_arg, "r");
+			config_filename = keep_config_arg(); /* Convert to UTF-8 */
+			config_file = openconnect_fopen_utf8(vpninfo, config_filename, "r");
+			if (config_filename != config_arg)
+				free(config_filename);
 			if (!config_file) {
 				fprintf(stderr, _("Cannot open config file '%s': %s\n"),
 					config_arg, strerror(errno));
@@ -755,7 +1057,7 @@ int main(int argc, char **argv)
 			/* The next option will come from the file... */
 			break;
 		case OPT_CAFILE:
-			openconnect_set_cafile(vpninfo, xstrdup(config_arg));
+			openconnect_set_cafile(vpninfo, dup_config_arg());
 			break;
 		case OPT_PIDFILE:
 			pidfile = keep_config_arg();
@@ -764,7 +1066,8 @@ int main(int argc, char **argv)
 			openconnect_set_pfs(vpninfo, 1);
 			break;
 		case OPT_SERVERCERT:
-			openconnect_set_server_cert_sha1(vpninfo, xstrdup(config_arg));
+			server_cert = keep_config_arg();
+			openconnect_set_system_trust(vpninfo, 0);
 			break;
 		case OPT_NO_DTLS:
 			use_dtls = 0;
@@ -806,16 +1109,16 @@ int main(int argc, char **argv)
 			authgroup = keep_config_arg();
 			break;
 		case 'C':
-			vpninfo->cookie = strdup(config_arg);
+			vpninfo->cookie = dup_config_arg();
 			break;
 		case 'c':
-			vpninfo->cert = strdup(config_arg);
+			vpninfo->cert = dup_config_arg();
 			break;
 		case 'e':
 			vpninfo->cert_expire_warning = 86400 * atoi(config_arg);
 			break;
 		case 'k':
-			vpninfo->sslkey = strdup(config_arg);
+			vpninfo->sslkey = dup_config_arg();
 			break;
 		case 'd':
 			vpninfo->deflate = 1;
@@ -825,12 +1128,12 @@ int main(int argc, char **argv)
 			break;
 		case 'g':
 			free(urlpath);
-			urlpath = strdup(config_arg);
+			urlpath = dup_config_arg();
 			break;
 		case 'h':
 			usage();
 		case 'i':
-			ifname = xstrdup(config_arg);
+			ifname = dup_config_arg();
 			break;
 		case 'm': {
 			int mtu = atol(config_arg);
@@ -856,12 +1159,15 @@ int main(int argc, char **argv)
 			autoproxy = 0;
 			break;
 		case OPT_PROXY_AUTH:
-			openconnect_set_proxy_auth(vpninfo, xstrdup(config_arg));
+			openconnect_set_proxy_auth(vpninfo, dup_config_arg());
 			break;
 		case OPT_NO_PROXY:
 			autoproxy = 0;
 			proxy = NULL;
 			break;
+		case OPT_NO_SYSTEM_TRUST:
+			openconnect_set_system_trust(vpninfo, 0);
+			break;
 		case OPT_LIBPROXY:
 			autoproxy = 1;
 			proxy = NULL;
@@ -876,11 +1182,11 @@ int main(int argc, char **argv)
 			nocertcheck = 1;
 			break;
 		case 's':
-			vpnc_script = xstrdup(config_arg);
+			vpnc_script = dup_config_arg();
 			break;
 		case 'u':
 			free(username);
-			username = strdup(config_arg);
+			username = dup_config_arg();
 			break;
 		case OPT_DISABLE_IPV6:
 			vpninfo->disable_ipv6 = 1;
@@ -914,7 +1220,7 @@ int main(int argc, char **argv)
 			break;
 		case OPT_USERAGENT:
 			free(vpninfo->useragent);
-			vpninfo->useragent = strdup(config_arg);
+			vpninfo->useragent = dup_config_arg();
 			break;
 		case OPT_FORCE_DPD:
 			openconnect_set_dpd(vpninfo, atoi(config_arg));
@@ -929,6 +1235,8 @@ int main(int argc, char **argv)
 				token_mode = OC_TOKEN_MODE_TOTP;
 			} else if (strcasecmp(config_arg, "hotp") == 0) {
 				token_mode = OC_TOKEN_MODE_HOTP;
+			} else if (strcasecmp(config_arg, "yubioath") == 0) {
+				token_mode = OC_TOKEN_MODE_YUBIOATH;
 			} else {
 				fprintf(stderr, _("Invalid software token mode \"%s\"\n"),
 					config_arg);
@@ -948,7 +1256,7 @@ int main(int argc, char **argv)
 				/* generic defaults */
 				openconnect_set_mobile_info(vpninfo,
 					xstrdup("1.0"),
-					xstrdup(config_arg),
+					dup_config_arg(),
 					xstrdup("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"));
 			}
 			break;
@@ -1048,11 +1356,8 @@ int main(int argc, char **argv)
 		/* --authenticate */
 		printf("COOKIE='%s'\n", vpninfo->cookie);
 		printf("HOST='%s'\n", openconnect_get_hostname(vpninfo));
-		if (vpninfo->peer_cert) {
-			char buf[41] = {0, };
-			openconnect_get_cert_sha1(vpninfo, vpninfo->peer_cert, buf);
-			printf("FINGERPRINT='%s'\n", buf);
-		}
+		printf("FINGERPRINT='%s'\n",
+		       openconnect_get_peer_cert_hash(vpninfo));
 		openconnect_vpninfo_free(vpninfo);
 		exit(0);
 	} else if (cookieonly) {
@@ -1125,7 +1430,7 @@ int main(int argc, char **argv)
 		   more sanely. It's *possible* that we'll fail to write to
 		   it, but very unlikely. */
 		if (pidfile != NULL) {
-			fp = fopen(pidfile, "w");
+			fp = openconnect_fopen_utf8(vpninfo, pidfile, "w");
 			if (!fp) {
 				fprintf(stderr, _("Failed to open '%s' for write: %s\n"),
 					pidfile, strerror(errno));
@@ -1170,7 +1475,7 @@ int main(int argc, char **argv)
 		ret = 1;
 		break;
 	case -EINTR:
-		vpn_progress(vpninfo, PRG_INFO, _("User canceled (SIGINT); exiting.\n"));
+		vpn_progress(vpninfo, PRG_INFO, _("User cancelled (SIGINT); exiting.\n"));
 		ret = 0;
 		break;
 	case -ECONNABORTED:
@@ -1187,13 +1492,14 @@ int main(int argc, char **argv)
 	exit(ret);
 }
 
-static int write_new_config(void *_vpninfo, char *buf, int buflen)
+static int write_new_config(void *_vpninfo, const char *buf, int buflen)
 {
 	struct openconnect_info *vpninfo = _vpninfo;
 	int config_fd;
 	int err;
 
-	config_fd = open(vpninfo->xmlconfig, O_WRONLY|O_TRUNC|O_CREAT, 0644);
+	config_fd = openconnect_open_utf8(vpninfo, vpninfo->xmlconfig,
+					  O_WRONLY|O_TRUNC|O_CREAT|O_BINARY);
 	if (config_fd < 0) {
 		err = errno;
 		fprintf(stderr, _("Failed to open %s for write: %s\n"),
@@ -1214,7 +1520,8 @@ static int write_new_config(void *_vpninfo, char *buf, int buflen)
 	return 0;
 }
 
-void write_progress(void *_vpninfo, int level, const char *fmt, ...)
+static void __attribute__ ((format(printf, 3, 4)))
+    write_progress(void *_vpninfo, int level, const char *fmt, ...)
 {
 	FILE *outf = level ? stdout : stderr;
 	va_list args;
@@ -1240,35 +1547,49 @@ void write_progress(void *_vpninfo, int level, const char *fmt, ...)
 
 struct accepted_cert {
 	struct accepted_cert *next;
-	char fingerprint[SHA1_SIZE * 2 + 1];
-	char host[0];
+	char *fingerprint;
+	char *host;
+	int port;
 } *accepted_certs;
 
-static int validate_peer_cert(void *_vpninfo, OPENCONNECT_X509 *peer_cert,
-			      const char *reason)
+static int validate_peer_cert(void *_vpninfo, const char *reason)
 {
 	struct openconnect_info *vpninfo = _vpninfo;
-	char fingerprint[SHA1_SIZE * 2 + 1];
+	const char *fingerprint;
 	struct accepted_cert *this;
-	int ret;
+
+	if (server_cert) {
+		int err = openconnect_check_peer_cert_hash(vpninfo, server_cert);
+
+		if (!err)
+			return 0;
+
+		if (err < 0)
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Could not calculate hash of server's certificate\n"));
+		else
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Server SSL certificate didn't match: %s\n"),
+				     openconnect_get_peer_cert_hash(vpninfo));
+
+		return -EINVAL;
+	}
 
 	if (nocertcheck)
 		return 0;
 
-	ret = openconnect_get_cert_sha1(vpninfo, peer_cert, fingerprint);
-	if (ret)
-		return ret;
+	fingerprint = openconnect_get_peer_cert_hash(vpninfo);
 
 	for (this = accepted_certs; this; this = this->next) {
 		if (!strcasecmp(this->host, vpninfo->hostname) &&
-		    !strcasecmp(this->fingerprint, fingerprint))
+		    this->port == vpninfo->port &&
+		    !openconnect_check_peer_cert_hash(vpninfo, this->fingerprint))
 			return 0;
 	}
 
 	while (1) {
-		char buf[80];
 		char *details;
-		char *p;
+		char *response = NULL;
 
 		fprintf(stderr, _("\nCertificate from VPN server \"%s\" failed verification.\n"
 			 "Reason: %s\n"), vpninfo->hostname, reason);
@@ -1278,30 +1599,33 @@ static int validate_peer_cert(void *_vpninfo, OPENCONNECT_X509 *peer_cert,
 
 		fprintf(stderr, _("Enter '%s' to accept, '%s' to abort; anything else to view: "),
 		       _("yes"), _("no"));
-		if (!fgets(buf, sizeof(buf), stdin))
-			return -EINVAL;
-		p = strchr(buf, '\n');
-		if (p)
-			*p = 0;
 
-		if (!strcasecmp(buf, _("yes"))) {
-			struct accepted_cert *newcert = malloc(sizeof(*newcert) +
-							       strlen(vpninfo->hostname) + 1);
+		read_stdin(&response, 0);
+		if (!response)
+			return -EINVAL;
+
+		if (!strcasecmp(response, _("yes"))) {
+			struct accepted_cert *newcert = malloc(sizeof(*newcert));
 			if (newcert) {
 				newcert->next = accepted_certs;
 				accepted_certs = newcert;
-				strcpy(newcert->fingerprint, fingerprint);
-				strcpy(newcert->host, vpninfo->hostname);
+				newcert->fingerprint = strdup(fingerprint);
+				newcert->host = strdup(vpninfo->hostname);
+				newcert->port = vpninfo->port;
 			}
+			free(response);
 			return 0;
 		}
-		if (!strcasecmp(buf, _("no")))
+		if (!strcasecmp(response, _("no"))) {
+			free(response);
 			return -EINVAL;
+		}
+		free(response);
 
-		details = openconnect_get_cert_details(vpninfo, peer_cert);
+		details = openconnect_get_peer_cert_details(vpninfo);
 		fputs(details, stderr);
-		free(details);
-		fprintf(stderr, _("SHA1 fingerprint: %s\n"), fingerprint);
+		openconnect_free_cert_info(vpninfo, details);
+		fprintf(stderr, _("Server key hash: %s\n"), fingerprint);
 	}
 }
 
@@ -1321,7 +1645,7 @@ static int match_choice_label(struct openconnect_info *vpninfo,
 
 		if (!strncasecmp(label, choice->label, input_len)) {
 			if (strlen(choice->label) == input_len) {
-				select_opt->form.value = choice->name;
+				select_opt->form._value = choice->name;
 				return 0;
 			} else {
 				match = choice->name;
@@ -1331,7 +1655,7 @@ static int match_choice_label(struct openconnect_info *vpninfo,
 	}
 
 	if (partial_matches == 1) {
-		select_opt->form.value = match;
+		select_opt->form._value = match;
 		return 0;
 	} else if (partial_matches > 1) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -1460,26 +1784,26 @@ static int process_auth_form_cb(void *_vpninfo,
 		} else if (opt->type == OC_FORM_OPT_TEXT) {
 			if (username &&
 			    !strcmp(opt->name, "username")) {
-				opt->value = username;
+				opt->_value = username;
 				username = NULL;
 			} else {
-				opt->value = prompt_for_input(opt->label, vpninfo, 0);
+				opt->_value = prompt_for_input(opt->label, vpninfo, 0);
 			}
 
-			if (!opt->value)
+			if (!opt->_value)
 				goto err;
 			empty = 0;
 
 		} else if (opt->type == OC_FORM_OPT_PASSWORD) {
 			if (password &&
 			    !strcmp(opt->name, "password")) {
-				opt->value = password;
+				opt->_value = password;
 				password = NULL;
 			} else {
-				opt->value = prompt_for_input(opt->label, vpninfo, 1);
+				opt->_value = prompt_for_input(opt->label, vpninfo, 1);
 			}
 
-			if (!opt->value)
+			if (!opt->_value)
 				goto err;
 			empty = 0;
 		}
@@ -1496,13 +1820,83 @@ static int process_auth_form_cb(void *_vpninfo,
 	return OC_FORM_RESULT_ERR;
 }
 
+static int lock_token(void *tokdata)
+{
+	struct openconnect_info *vpninfo = tokdata;
+	char *file_token;
+	int err;
+
+	/* FIXME: Actually lock the file */
+	err = read_file_into_string(vpninfo, token_filename, &file_token);
+	if (err < 0)
+	    return err;
+
+	err = openconnect_set_token_mode(vpninfo, vpninfo->token_mode, file_token);
+	free(file_token);
+
+	return 0;
+}
+
+static int unlock_token(void *tokdata, const char *new_tok)
+{
+	struct openconnect_info *vpninfo = tokdata;
+	int tok_fd;
+	int err;
+
+	if (!new_tok)
+		return 0;
+
+	tok_fd = openconnect_open_utf8(vpninfo, token_filename,
+				       O_WRONLY|O_TRUNC|O_CREAT|O_BINARY);
+	if (tok_fd < 0) {
+		err = errno;
+		fprintf(stderr, _("Failed to open token file for write: %s\n"),
+			strerror(err));
+		return -err;
+	}
+
+	/* FIXME: We should actually write to a new tempfile, then rename */
+	if (write(tok_fd, new_tok, strlen(new_tok)) != strlen(new_tok)) {
+		err = errno;
+		fprintf(stderr, _("Failed to write token: %s\n"),
+			strerror(err));
+		close(tok_fd);
+		return -err;
+	}
+
+	close(tok_fd);
+	return 0;
+}
+
 static void init_token(struct openconnect_info *vpninfo,
 		       oc_token_mode_t token_mode, const char *token_str)
 {
 	int ret;
+	char *file_token = NULL;
 
-	ret = openconnect_set_token_mode(vpninfo, token_mode, token_str);
+	if (token_str) {
+		switch(token_str[0]) {
+		case '@':
+			token_str++;
+		case '/':
+			if (read_file_into_string(vpninfo, token_str,
+						  &file_token) < 0)
+				exit(1);
+			break;
+		default:
+			/* Use token_str as raw data */
+			break;
+		}
+	}
 
+	ret = openconnect_set_token_mode(vpninfo, token_mode,
+					 file_token ? : token_str);
+	if (file_token) {
+		token_filename = strdup(token_str);
+		openconnect_set_token_callbacks(vpninfo, vpninfo,
+						lock_token, unlock_token);
+		free(file_token);
+	}
 	switch (token_mode) {
 	case OC_TOKEN_MODE_STOKEN:
 		switch (ret) {
@@ -1542,6 +1936,21 @@ static void init_token(struct openconnect_info *vpninfo,
 
 		break;
 
+	case OC_TOKEN_MODE_YUBIOATH:
+		switch(ret) {
+		case 0:
+			return;
+		case -ENOENT:
+			fprintf(stderr, _("Yubikey token not found\n"));
+			exit(1);
+		case -EOPNOTSUPP:
+			fprintf(stderr, _("OpenConnect was not built with Yubikey support\n"));
+			exit(1);
+		default:
+			fprintf(stderr, _("General Yubikey failure: %s\n"), strerror(-ret));
+			exit(1);
+		}
+
 	case OC_TOKEN_MODE_NONE:
 		/* No-op */
 		break;
