@@ -51,15 +51,15 @@ static int ntlm_sspi(struct openconnect_info *vpninfo, struct oc_text_buf *buf,
         ULONG ret_flags;
 
 	if (challenge) {
-		int token_len;
+		int token_len = -EINVAL;
 
 		input_desc.cBuffers = 1;
 		input_desc.pBuffers = &in_token;
 		input_desc.ulVersion = SECBUFFER_VERSION;
 
 		in_token.BufferType = SECBUFFER_TOKEN;
-		token_len = openconnect_base64_decode(&in_token.pvBuffer, challenge);
-		if (token_len < 0)
+		in_token.pvBuffer = openconnect_base64_decode(&token_len, challenge);
+		if (!in_token.pvBuffer)
 			return token_len;
 		in_token.cbBuffer = token_len;
 	}
@@ -72,10 +72,15 @@ static int ntlm_sspi(struct openconnect_info *vpninfo, struct oc_text_buf *buf,
         out_token.cbBuffer = 0;
         out_token.pvBuffer = NULL;
 
-	status = InitializeSecurityContext(&vpninfo->ntlm_sspi_cred, challenge ? &vpninfo->ntlm_sspi_ctx : NULL, (SEC_CHAR *)"",
-					   ISC_REQ_ALLOCATE_MEMORY | ISC_REQ_CONFIDENTIALITY | ISC_REQ_REPLAY_DETECT | ISC_REQ_CONNECTION,
-					   0, SECURITY_NETWORK_DREP, challenge ? &input_desc : NULL, 0, &vpninfo->ntlm_sspi_ctx,
-					   &output_desc, &ret_flags, NULL);
+	status = InitializeSecurityContextW(&vpninfo->ntlm_sspi_cred,
+					    challenge ? &vpninfo->ntlm_sspi_ctx : NULL,
+					    (SEC_WCHAR *)L"",
+					    ISC_REQ_ALLOCATE_MEMORY | ISC_REQ_CONFIDENTIALITY | ISC_REQ_REPLAY_DETECT | ISC_REQ_CONNECTION,
+					    0, SECURITY_NETWORK_DREP,
+					    challenge ? &input_desc : NULL,
+					    0, &vpninfo->ntlm_sspi_ctx,
+					    &output_desc, &ret_flags, NULL);
+
 	if (status != SEC_E_OK && status != SEC_I_CONTINUE_NEEDED) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("InitializeSecurityContext() failed: %lx\n"), status);
@@ -96,8 +101,10 @@ static int ntlm_helper_spawn(struct openconnect_info *vpninfo, struct oc_text_bu
         SECURITY_STATUS status;
 	int ret;
 
-	status = AcquireCredentialsHandle(NULL, (SEC_CHAR *)"NTLM", SECPKG_CRED_OUTBOUND,
-					  NULL, NULL, NULL, NULL, &vpninfo->ntlm_sspi_cred, NULL);
+	status = AcquireCredentialsHandleW(NULL, (SEC_WCHAR *)L"NTLM",
+					   SECPKG_CRED_OUTBOUND, NULL, NULL,
+					   NULL, NULL,
+					   &vpninfo->ntlm_sspi_cred, NULL);
 	if (status != SEC_E_OK) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("AcquireCredentialsHandle() failed: %lx\n"), status);
@@ -225,12 +232,16 @@ static int ntlm_helper_challenge(struct openconnect_info *vpninfo, struct oc_tex
 		  strlen(vpninfo->auth[AUTH_TYPE_NTLM].challenge)) != strlen(vpninfo->auth[AUTH_TYPE_NTLM].challenge) ||
 	    write(vpninfo->ntlm_helper_fd, "\n", 1) != 1) {
 	err:
+		vpn_progress(vpninfo, PRG_ERR, _("Error communicating with ntlm_auth helper\n"));
 		close(vpninfo->ntlm_helper_fd);
 		vpninfo->ntlm_helper_fd = -1;
 		return -EAGAIN;
 	}
 	len = read(vpninfo->ntlm_helper_fd, helperbuf, sizeof(helperbuf));
-	if (len < 4 || helperbuf[0] != 'K' || helperbuf[1] != 'K' ||
+	/* Accept both 'KK' and 'AF'. It should be the latter but see
+	   https://bugzilla.samba.org/show_bug.cgi?id=10691 */
+	if (len < 4 || (!(helperbuf[0] == 'K' && helperbuf[1] == 'K') &&
+			!(helperbuf[0] == 'A' && helperbuf[1] == 'F')) ||
 	    helperbuf[2] != ' ' || helperbuf[len - 1] != '\n') {
 		goto err;
 	}
@@ -246,7 +257,8 @@ void cleanup_ntlm_auth(struct openconnect_info *vpninfo)
 {
 	if (vpninfo->auth[AUTH_TYPE_NTLM].state == NTLM_SSO_REQ) {
 		close(vpninfo->ntlm_helper_fd);
-		vpninfo->ntlm_helper_fd = -1;}
+		vpninfo->ntlm_helper_fd = -1;
+	}
 }
 #endif /* !_WIN32 */
 
@@ -735,68 +747,6 @@ static void setup_schedule (const unsigned char *key_56, DES_KS ks)
 	deskey (ks, key, 0);
 }
 
-static int buf_append_utf16le(struct oc_text_buf *buf, const char *utf8)
-{
-	int len = 0;
-	unsigned char c;
-	unsigned long utfchar;
-	int min;
-	int nr_extra;
-
-	/* Ick. Now I'm implementing my own UTF8 handling too. Perhaps it's
-	   time to bite the bullet and start requiring something like glib? */
-	while (*utf8) {
-		c = *(utf8++);
-		if (c < 128) {
-			utfchar = c;
-			nr_extra = 0;
-			min = 0;
-		} else if ((c & 0xe0) == 0xc0) {
-			utfchar = c & 0x1f;
-			nr_extra = 1;
-			min = 0x80;
-		} else if ((c & 0xf0) == 0xe0) {
-			utfchar = c & 0x0f;
-			nr_extra = 2;
-			min = 0x800;
-		} else if ((c & 0xf8) == 0xf0) {
-			utfchar = c & 0x07;
-			nr_extra = 3;
-			min = 0x10000;
-		} else {
-			return -EINVAL;
-		}
-
-		while (nr_extra--) {
-			c = *(utf8++);
-			if ((c & 0xc0) != 0x80)
-				return -EINVAL;
-			utfchar <<= 6;
-			utfchar |= (c & 0x3f);
-		}
-		if (utfchar > 0x10ffff || utfchar < min)
-			return -EINVAL;
-
-		if (utfchar >= 0x10000) {
-			utfchar -= 0x10000;
-			if (buf_ensure_space(buf, 4))
-				return -ENOMEM;
-			buf->data[buf->pos++] = (utfchar >> 10) & 0xff;
-			buf->data[buf->pos++] = 0xd8 | ((utfchar >> 18) & 3);
-			buf->data[buf->pos++] = utfchar & 0xff;
-			buf->data[buf->pos++] = 0xdc | ((utfchar >> 8) & 3);
-			len += 4;
-		} else {
-			if (buf_ensure_space(buf, 2))
-				return -ENOMEM;
-			buf->data[buf->pos++] = utfchar & 0xff;
-			buf->data[buf->pos++] = utfchar >> 8;
-			len += 2;
-		}
-	}
-	return len;
-}
-
 #define LM_PASSWORD_MAGIC "\x4B\x47\x53\x21\x40\x23\x24\x25" \
                           "\x4B\x47\x53\x21\x40\x23\x24\x25" \
 			  "\x00\x00\x00\x00\x00"
@@ -954,7 +904,7 @@ static int ntlm_manual_challenge(struct openconnect_info *vpninfo, struct oc_tex
 	char *user;
 	unsigned char nonce[8], hash[21], lm_resp[24], nt_resp[24];
 	unsigned char *token;
-	int token_len;
+	int token_len = -EINVAL;
 	int ntlmver;
 
 	if (!vpninfo->auth[AUTH_TYPE_NTLM].challenge)
@@ -963,9 +913,9 @@ static int ntlm_manual_challenge(struct openconnect_info *vpninfo, struct oc_tex
 	if (ntlm_nt_hash (vpninfo->proxy_pass, (char *) hash))
 		return -EINVAL;
 
-	token_len = openconnect_base64_decode((void **)&token,
-					      vpninfo->auth[AUTH_TYPE_NTLM].challenge);
-	if (token_len < 0)
+	token = openconnect_base64_decode(&token_len,
+					  vpninfo->auth[AUTH_TYPE_NTLM].challenge);
+	if (!token)
 		return token_len;
 
 	if (token_len < NTLM_CHALLENGE_NONCE_OFFSET + 8 || token[0] != 'N' ||
@@ -1079,7 +1029,12 @@ int ntlm_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *buf
 		/* Clean up after it. We're done here, whether it worked or not */
 		cleanup_ntlm_auth(vpninfo);
 		vpninfo->auth[AUTH_TYPE_NTLM].state = NTLM_MANUAL;
-		if (!ret || ret == -EAGAIN)
+		if (ret == -EAGAIN) {
+			/* Don't let it reset our state when it reconnects */
+			vpninfo->proxy_close_during_auth = 1;
+			return ret;
+		}
+		if (!ret)
 			return ret;
 	}
 	if (vpninfo->auth[AUTH_TYPE_NTLM].state == NTLM_MANUAL && vpninfo->proxy_user &&
