@@ -62,6 +62,11 @@
 #endif
 #endif
 
+#ifdef HAVE_ICONV
+#include <langinfo.h>
+#include <iconv.h>
+#endif
+
 #include <zlib.h>
 #include <stdint.h>
 #include <sys/time.h>
@@ -80,8 +85,20 @@
 #include GSSAPI_HDR
 #endif
 
+#ifdef HAVE_LIBPSKC
+#include <pskc/pskc.h>
+#endif
+
+#ifdef HAVE_LIBPCSCLITE
+#ifdef __APPLE__
+#include <PCSC/wintypes.h>
+#include <PCSC/winscard.h>
+#else
+#include <winscard.h>
+#endif
+#endif
+
 #ifdef ENABLE_NLS
-#include <locale.h>
 #include <libintl.h>
 #define _(s) dgettext("openconnect", s)
 #else
@@ -146,11 +163,6 @@ struct oc_text_buf {
 #define RECONNECT_INTERVAL_MIN	10
 #define RECONNECT_INTERVAL_MAX	100
 
-#define CERT_TYPE_UNKNOWN	0
-#define CERT_TYPE_PEM		1
-#define CERT_TYPE_PKCS12	2
-#define CERT_TYPE_TPM		3
-
 #define REDIR_TYPE_NONE		0
 #define REDIR_TYPE_NEWHOST	1
 #define REDIR_TYPE_LOCAL	2
@@ -175,12 +187,16 @@ struct proxy_auth_state {
 };
 
 struct openconnect_info {
+#ifdef HAVE_ICONV
+	iconv_t ic_legacy_to_utf8;
+	iconv_t ic_utf8_to_legacy;
+#endif
 	char *redirect_url;
 	int redirect_type;
 
 	const char *csd_xmltag;
 	int csd_nostub;
-	const char *platname;
+	char *platname;
 	char *mobile_platform_version;
 	char *mobile_device_type;
 	char *mobile_device_uniqueid;
@@ -206,6 +222,7 @@ struct openconnect_info {
 	int proxy_fd;
 	char *proxy_user;
 	char *proxy_pass;
+	int proxy_close_during_auth;
 	struct proxy_auth_state auth[MAX_AUTH_TYPES];
 #ifdef HAVE_GSSAPI
 	gss_name_t gss_target_name;
@@ -216,7 +233,7 @@ struct openconnect_info {
 	CtxtHandle ntlm_sspi_ctx;
 	CredHandle sspi_cred;
 	CtxtHandle sspi_ctx;
-	char *sspi_target_name;
+	SEC_WCHAR *sspi_target_name;
 #else
 	int ntlm_helper_fd;
 #endif
@@ -228,12 +245,11 @@ struct openconnect_info {
 	int port;
 	char *urlpath;
 	int cert_expire_warning;
-	const char *cert;
-	const char *sslkey;
-	int cert_type;
+	char *cert;
+	char *sslkey;
 	char *cert_password;
 	char *cafile;
-	char *servercert;
+	unsigned no_system_trust;
 	const char *xmlconfig;
 	char xmlsha1[(SHA1_SIZE * 2) + 1];
 	char *authgroup;
@@ -253,19 +269,45 @@ struct openconnect_info {
 #ifdef HAVE_LIBSTOKEN
 	struct stoken_ctx *stoken_ctx;
 	char *stoken_pin;
+	int stoken_concat_pin;
+	int stoken_interval;
+#endif
+#ifdef HAVE_LIBPSKC
+	pskc_t *pskc;
+	pskc_key_t *pskc_key;
 #endif
 #ifdef HAVE_LIBOATH
 	char *oath_secret;
 	size_t oath_secret_len;
+	enum {
+		HOTP_SECRET_BASE32 = 1,
+		HOTP_SECRET_RAW,
+		HOTP_SECRET_HEX,
+		HOTP_SECRET_PSKC,
+	} hotp_secret_format; /* We need to give it back in the same form */
 #endif
+#ifdef HAVE_LIBPCSCLITE
+	SCARDHANDLE pcsc_ctx, pcsc_card;
+	char *yubikey_objname;
+	unsigned char yubikey_pwhash[16];
+	int yubikey_pw_set;
+	int yubikey_mode;
+#endif
+	openconnect_lock_token_vfn lock_token;
+	openconnect_unlock_token_vfn unlock_token;
+	void *tok_cbdata;
 
-	OPENCONNECT_X509 *peer_cert;
+	void *peer_cert;
+	char *peer_cert_hash;
 
 	char *cookie; /* Pointer to within cookies list */
 	struct oc_vpn_option *cookies;
 	struct oc_vpn_option *cstp_options;
 	struct oc_vpn_option *dtls_options;
 
+	struct oc_vpn_option *script_env;
+	struct oc_vpn_option *csd_env;
+
 	unsigned pfs;
 #if defined(OPENCONNECT_OPENSSL)
 	X509 *cert_x509;
@@ -274,6 +316,7 @@ struct openconnect_info {
 #elif defined(OPENCONNECT_GNUTLS)
 	gnutls_session_t https_sess;
 	gnutls_certificate_credentials_t https_cred;
+	char local_cert_md5[MD5_SIZE * 2 + 1]; /* For CSD */
 	struct pin_cache *pin_cache;
 #ifdef HAVE_TROUSERS
 	TSS_HCONTEXT tpm_context;
@@ -298,6 +341,8 @@ struct openconnect_info {
 	struct pkt *current_ssl_pkt;
 	struct pkt *pending_deflated_pkt;
 
+	struct pkt *tun_pkt;
+
 	z_stream inflate_strm;
 	uint32_t inflate_adler32;
 	z_stream deflate_strm;
@@ -318,8 +363,12 @@ struct openconnect_info {
 	   NULL or not or pass it to DTLS_SEND/DTLS_RECV. This way we
 	   have fewer ifdefs and accessor macros for it. */
 	gnutls_session_t dtls_ssl;
+	char *gnutls_dtls_cipher; /* cached for openconnect_get_dtls_cipher() */
 #endif
+	char *cstp_cipher;
+
 	int dtls_state;
+	int dtls_need_reconnect;
 	struct keepalive_info dtls_times;
 	unsigned char dtls_session_id[32];
 	unsigned char dtls_secret[48];
@@ -378,6 +427,7 @@ struct openconnect_info {
 	int dtls_local_port;
 
 	int deflate;
+	int is_dyndns; /* Attempt to redo DNS lookup on each CSTP reconnect */
 	char *useragent;
 
 	const char *quit_reason;
@@ -437,6 +487,7 @@ struct openconnect_info {
 #define AC_PKT_TERM_SERVER	9	/* Server kick */
 
 #define vpn_progress(vpninfo, ...) (vpninfo)->progress((vpninfo)->cbdata, __VA_ARGS__)
+#define vpn_perror(vpninfo, msg) vpn_progress((vpninfo), PRG_ERR, "%s: %s", (msg), strerror(errno));
 
 /****************************************************************************/
 /* Oh Solaris how we hate thee! */
@@ -444,6 +495,10 @@ struct openconnect_info {
 #define time(x) openconnect__time(x)
 time_t openconnect__time(time_t *t);
 #endif
+#ifndef HAVE_VASPRINTF
+#define vasprintf openconnect__vasprintf
+int openconnect__vasprintf(char **strp, const char *fmt, va_list ap);
+#endif
 #ifndef HAVE_ASPRINTF
 #define asprintf openconnect__asprintf
 int openconnect__asprintf(char **strp, const char *fmt, ...);
@@ -461,14 +516,6 @@ char *openconnect__strcasestr(const char *haystack, const char *needle);
 #define strndup openconnect__strndup
 char *openconnect__strndup(const char *s, size_t n);
 #endif
-#ifndef HAVE_SETENV
-#define setenv openconnect__setenv
-int openconnect__setenv(const char *name, const char *value, int overwrite);
-#endif
-#ifndef HAVE_UNSETENV
-#define unsetenv openconnect__unsetenv
-void openconnect__unsetenv(const char *name);
-#endif
 
 #ifndef HAVE_INET_ATON
 #define inet_aton openconnect__inet_aton
@@ -495,15 +542,27 @@ static inline int set_fd_cloexec(int fd)
 
 #ifdef _WIN32
 #define pipe(fds) _pipe(fds, 4096, O_BINARY)
-void openconnect__win32_sock_init();
-int dumb_socketpair(int socks[2], int make_overlapped);
+int openconnect__win32_sock_init();
+char *openconnect__win32_strerror(DWORD err);
+#undef inet_pton
+#define inet_pton openconnect__win32_inet_pton
+int openconnect__win32_inet_pton(int af, const char *src, void *dst);
+#define OPENCONNECT_CMD_SOCKET SOCKET
+OPENCONNECT_CMD_SOCKET dumb_socketpair(OPENCONNECT_CMD_SOCKET socks[2], int make_overlapped);
 #else
 #define closesocket close
+#define OPENCONNECT_CMD_SOCKET int
 #ifndef O_BINARY
 #define O_BINARY 0
 #endif
 #endif
 
+/* For systems that don't support O_CLOEXEC, just don't bother.
+   We don't keep files open for long anyway. */
+#ifndef O_CLOEXEC
+#define O_CLOEXEC 0
+#endif
+
 /* I always coded as if it worked like this. Now it does. */
 #define realloc_inplace(p, size) do {			\
 	void *__realloc_old = p;			\
@@ -514,10 +573,21 @@ int dumb_socketpair(int socks[2], int make_overlapped);
 
 /****************************************************************************/
 
+/* iconv.c */
+#ifdef HAVE_ICONV
+char *openconnect_utf8_to_legacy(struct openconnect_info *vpninfo, const char *utf8);
+char *openconnect_legacy_to_utf8(struct openconnect_info *vpninfo, const char *legacy);
+#else
+#define openconnect_utf8_to_legacy(v, str) ((char *)str)
+#define openconnect_legacy_to_utf8(v, str) ((char *)str)
+#endif
+
 /* script.c */
-int setenv_int(const char *opt, int value);
-void set_script_env(struct openconnect_info *vpninfo);
+int script_setenv(struct openconnect_info *vpninfo, const char *opt, const char *val, int append);
+int script_setenv_int(struct openconnect_info *vpninfo, const char *opt, int value);
+void prepare_script_env(struct openconnect_info *vpninfo);
 int script_config_tun(struct openconnect_info *vpninfo, const char *reason);
+int apply_script_env(struct oc_vpn_option *envs);
 
 /* tun.c / tun-win32.c */
 void os_shutdown_tun(struct openconnect_info *vpninfo);
@@ -532,7 +602,6 @@ int dtls_try_handshake(struct openconnect_info *vpninfo);
 int connect_dtls_socket(struct openconnect_info *vpninfo);
 void dtls_close(struct openconnect_info *vpninfo);
 void dtls_shutdown(struct openconnect_info *vpninfo);
-int dtls_reconnect(struct openconnect_info *vpninfo);
 
 /* cstp.c */
 int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout);
@@ -542,7 +611,8 @@ void cstp_free_splits(struct openconnect_info *vpninfo);
 /* ssl.c */
 unsigned string_is_hostname(const char* str);
 int connect_https_socket(struct openconnect_info *vpninfo);
-int request_passphrase(struct openconnect_info *vpninfo, const char *label,
+int __attribute__ ((format(printf, 4, 5)))
+    request_passphrase(struct openconnect_info *vpninfo, const char *label,
 		       char **response, const char *fmt, ...);
 int  __attribute__ ((format (printf, 2, 3)))
     openconnect_SSL_printf(struct openconnect_info *vpninfo, const char *fmt, ...);
@@ -559,18 +629,28 @@ void cmd_fd_set(struct openconnect_info *vpninfo, fd_set *fds, int *maxfd);
 void check_cmd_fd(struct openconnect_info *vpninfo, fd_set *fds);
 int is_cancel_pending(struct openconnect_info *vpninfo, fd_set *fds);
 void poll_cmd_fd(struct openconnect_info *vpninfo, int timeout);
+int openconnect_open_utf8(struct openconnect_info *vpninfo,
+			  const char *fname, int mode);
+FILE *openconnect_fopen_utf8(struct openconnect_info *vpninfo,
+			     const char *fname, const char *mode);
+
+void openconnect_clear_cookies(struct openconnect_info *vpninfo);
 
 /* {gnutls,openssl}.c */
 int openconnect_open_https(struct openconnect_info *vpninfo);
 void openconnect_close_https(struct openconnect_info *vpninfo, int final);
 int cstp_handshake(struct openconnect_info *vpninfo, unsigned init);
-int get_cert_md5_fingerprint(struct openconnect_info *vpninfo, OPENCONNECT_X509 *cert,
+int get_cert_md5_fingerprint(struct openconnect_info *vpninfo, void *cert,
 			     char *buf);
 int openconnect_sha1(unsigned char *result, void *data, int len);
 int openconnect_md5(unsigned char *result, void *data, int len);
 int openconnect_random(void *bytes, int len);
 int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
 			       char *buf);
+int openconnect_yubikey_chalresp(struct openconnect_info *vpninfo,
+				 const void *challenge, int chall_len, void *result);
+int openconnect_hash_yubikey_password(struct openconnect_info *vpninfo,
+				      const char *password, const void *ident, int id_len);
 #if defined(OPENCONNECT_OPENSSL)
 #define openconnect_https_connected(_v) ((_v)->https_ssl)
 #elif defined (OPENCONNECT_GNUTLS)
@@ -585,19 +665,58 @@ int keepalive_action(struct keepalive_info *ka, int *timeout);
 int ka_stalled_action(struct keepalive_info *ka, int *timeout);
 
 /* xml.c */
+ssize_t read_file_into_string(struct openconnect_info *vpninfo, const char *fname,
+			      char **ptr);
 int config_lookup_host(struct openconnect_info *vpninfo, const char *host);
 
+/* oath.c */
+int set_totp_mode(struct openconnect_info *vpninfo, const char *token_str);
+int set_hotp_mode(struct openconnect_info *vpninfo, const char *token_str);
+int can_gen_totp_code(struct openconnect_info *vpninfo,
+		      struct oc_auth_form *form,
+		      struct oc_form_opt *opt);
+int can_gen_hotp_code(struct openconnect_info *vpninfo,
+		      struct oc_auth_form *form,
+		      struct oc_form_opt *opt);
+int do_gen_totp_code(struct openconnect_info *vpninfo,
+		     struct oc_auth_form *form,
+		     struct oc_form_opt *opt);
+int do_gen_hotp_code(struct openconnect_info *vpninfo,
+		     struct oc_auth_form *form,
+		     struct oc_form_opt *opt);
+
+/* stoken.c */
+int prepare_stoken(struct openconnect_info *vpninfo);
+int set_libstoken_mode(struct openconnect_info *vpninfo, const char *token_str);
+int can_gen_stoken_code(struct openconnect_info *vpninfo,
+			struct oc_auth_form *form,
+			struct oc_form_opt *opt);
+int do_gen_stoken_code(struct openconnect_info *vpninfo,
+		       struct oc_auth_form *form,
+		       struct oc_form_opt *opt);
+
+/* yubikey.c */
+int set_yubikey_mode(struct openconnect_info *vpninfo, const char *token_str);
+int can_gen_yubikey_code(struct openconnect_info *vpninfo,
+			 struct oc_auth_form *form,
+			 struct oc_form_opt *opt);
+int do_gen_yubikey_code(struct openconnect_info *vpninfo,
+			struct oc_auth_form *form,
+			struct oc_form_opt *opt);
+
 /* auth.c */
+void nuke_opt_values(struct oc_form_opt *opt);
 int parse_xml_response(struct openconnect_info *vpninfo, char *response,
 		       struct oc_auth_form **form, int *cert_rq);
 int process_auth_form(struct openconnect_info *vpninfo,
 		      struct oc_auth_form *form);
-int handle_auth_form(struct openconnect_info *vpninfo, struct oc_auth_form *form,
-		     char *request_body, int req_len, const char **method,
+int handle_auth_form(struct openconnect_info *vpninfo,
+		     struct oc_auth_form *form,
+		     struct oc_text_buf *request_body, const char **method,
 		     const char **request_body_type);
 void free_auth_form(struct oc_auth_form *form);
-int xmlpost_initial_req(struct openconnect_info *vpninfo, char *request_body, int req_len, int cert_fail);
-int prepare_stoken(struct openconnect_info *vpninfo);
+int xmlpost_initial_req(struct openconnect_info *vpninfo,
+			struct oc_text_buf *request_body, int cert_fail);
 
 /* http.c */
 struct oc_text_buf *buf_alloc(void);
@@ -606,12 +725,17 @@ void  __attribute__ ((format (printf, 2, 3)))
 	buf_append(struct oc_text_buf *buf, const char *fmt, ...);
 void buf_append_bytes(struct oc_text_buf *buf, const void *bytes, int len);
 void buf_append_base64(struct oc_text_buf *buf, const void *bytes, int len);
-int openconnect_base64_decode(void **out, const char *in);
+int buf_append_utf16le(struct oc_text_buf *buf, const char *utf8);
+void buf_append_from_utf16le(struct oc_text_buf *buf, const void *utf16);
+void *openconnect_base64_decode(int *len, const char *in);
+void buf_truncate(struct oc_text_buf *buf);
+void buf_append_urlencoded(struct oc_text_buf *buf, char *str);
 int buf_error(struct oc_text_buf *buf);
 int buf_free(struct oc_text_buf *buf);
 char *openconnect_create_useragent(const char *base);
+void cleanup_proxy_auth(struct openconnect_info *vpninfo);
 int process_proxy(struct openconnect_info *vpninfo, int ssl_sock);
-int internal_parse_url(char *url, char **res_proto, char **res_host,
+int internal_parse_url(const char *url, char **res_proto, char **res_host,
 		       int *res_port, char **res_path, int default_port);
 
 /* ntlm.c */
@@ -629,4 +753,29 @@ int digest_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *b
 /* version.c */
 extern const char *openconnect_version_str;
 
+#define STRDUP(res, arg) \
+	do {							\
+		free(res);					\
+		if (arg) {					\
+			res = strdup(arg);			\
+			if (res == NULL) return -ENOMEM;	\
+		} else res = NULL;				\
+	} while(0)
+
+#define UTF8CHECK(arg) \
+	if ((arg) && buf_append_utf16le(NULL, (arg))) { \
+		vpn_progress(vpninfo, PRG_ERR,				\
+			     _("ERROR: %s() called with invalid UTF-8 for '%s' argument\n"),\
+			     __func__, #arg);				\
+		return -EILSEQ;						\
+	}
+
+#define UTF8CHECK_VOID(arg) \
+	if ((arg) && buf_append_utf16le(NULL, (arg))) { \
+		vpn_progress(vpninfo, PRG_ERR,				\
+			     _("ERROR: %s() called with invalid UTF-8 for '%s' argument\n"),\
+			     __func__, #arg);				\
+		return;							\
+	}
+
 #endif /* __OPENCONNECT_INTERNAL_H__ */
