@@ -57,14 +57,14 @@ int openconnect_md5(unsigned char *result, void *data, int len)
 	return 0;
 }
 
-int openconnect_get_cert_DER(struct openconnect_info *vpninfo,
-			     OPENCONNECT_X509 *cert, unsigned char **buf)
+int openconnect_get_peer_cert_DER(struct openconnect_info *vpninfo,
+				  unsigned char **buf)
 {
 	BIO *bp = BIO_new(BIO_s_mem());
 	BUF_MEM *certinfo;
 	size_t l;
 
-	if (!i2d_X509_bio(bp, cert)) {
+	if (!i2d_X509_bio(bp, vpninfo->peer_cert)) {
 		BIO_free(bp);
 		return -EIO;
 	}
@@ -282,8 +282,9 @@ static int ui_write(UI *ui, UI_STRING *uis)
 		break;
 
 	default:
-		fprintf(stderr, "Unhandled SSL UI request type %d\n",
-			UI_get_string_type(uis));
+		vpn_progress(ui_data->vpninfo, PRG_ERR,
+			     _("Unhandled SSL UI request type %d\n"),
+			     UI_get_string_type(uis));
 		return 0;
 	}
 	return 1;
@@ -302,8 +303,8 @@ static int ui_flush(UI *ui)
 
 	for (opt = (struct ui_form_opt *)ui_data->form.opts; opt;
 	     opt = (struct ui_form_opt *)opt->opt.next) {
-		if (opt->opt.value && opt->uis)
-			UI_set_result(ui, opt->uis, opt->opt.value);
+		if (opt->opt._value && opt->uis)
+			UI_set_result(ui, opt->uis, opt->opt._value);
 	}
 	return 1;
 }
@@ -316,8 +317,8 @@ static int ui_close(UI *ui)
 	opt = (struct ui_form_opt *)ui_data->form.opts;
 	while (opt) {
 		next_opt = (struct ui_form_opt *)opt->opt.next;
-		if (opt->opt.value)
-			free(opt->opt.value);
+		if (opt->opt._value)
+			free(opt->opt._value);
 		free(opt);
 		opt = next_opt;
 	}
@@ -400,6 +401,38 @@ static int pem_pw_cb(char *buf, int len, int w, void *v)
 	return plen;
 }
 
+static int install_extra_certs(struct openconnect_info *vpninfo, const char *source,
+			       STACK_OF(X509) *ca)
+{
+	X509 *cert = vpninfo->cert_x509;
+	int i;
+
+ next:
+	for (i = 0; i < sk_X509_num(ca); i++) {
+		X509 *cert2 = sk_X509_value(ca, i);
+		if (X509_check_issued(cert2, cert) == X509_V_OK) {
+			char buf[200];
+
+			if (cert2 == cert)
+				break;
+			if (X509_check_issued(cert2, cert2) == X509_V_OK)
+				break;
+
+			X509_NAME_oneline(X509_get_subject_name(cert2),
+					  buf, sizeof(buf));
+			vpn_progress(vpninfo, PRG_DEBUG,
+				     _("Extra cert from %s: '%s'\n"), source, buf);
+			CRYPTO_add(&cert2->references, 1, CRYPTO_LOCK_X509);
+			SSL_CTX_add_extra_chain_cert(vpninfo->https_ctx, cert2);
+			cert = cert2;
+			goto next;
+		}
+	}
+	sk_X509_pop_free(ca, X509_free);
+
+	return 0;
+}
+
 static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12)
 {
 	EVP_PKEY *pkey = NULL;
@@ -466,32 +499,8 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 		ret = -EINVAL;
 	}
 
-	/* Only include supporting certificates which are actually necessary */
-	if (ca) {
-		int i;
-	next:
-		for (i = 0; i < sk_X509_num(ca); i++) {
-			X509 *cert2 = sk_X509_value(ca, i);
-			if (X509_check_issued(cert2, cert) == X509_V_OK) {
-				char buf[200];
-
-				if (cert2 == cert)
-					break;
-				if (X509_check_issued(cert2, cert2) == X509_V_OK)
-					break;
-
-				X509_NAME_oneline(X509_get_subject_name(cert2),
-						  buf, sizeof(buf));
-				vpn_progress(vpninfo, PRG_DEBUG,
-					     _("Extra cert from PKCS#12: '%s'\n"), buf);
-				CRYPTO_add(&cert2->references, 1, CRYPTO_LOCK_X509);
-				SSL_CTX_add_extra_chain_cert(vpninfo->https_ctx, cert2);
-				cert = cert2;
-				goto next;
-			}
-		}
-		sk_X509_pop_free(ca, X509_free);
-	}
+	if (ca)
+		install_extra_certs(vpninfo, _("PKCS#12"), ca);
 
 	PKCS12_free(p12);
 	return ret;
@@ -564,31 +573,100 @@ static int load_tpm_certificate(struct openconnect_info *vpninfo)
 }
 #endif
 
-static int reload_pem_cert(struct openconnect_info *vpninfo)
+/* This is a reimplementation of SSL_CTX_use_certificate_chain_file().
+ * We do this for three reasons:
+ *
+ * - Firstly, we have no way to obtain the primary X509 certificate
+ *   after SSL_CTX_use_certificate_chain_file() has loaded it, and we
+ *   need to inspect it to check for expiry and report its name etc.
+ *   So in the past we've opened the cert file again and read the cert
+ *   again in a reload_pem_cert() function which was a partial
+ *   reimplementation anyway.
+ *
+ * - Secondly, on Windows, OpenSSL only partially handles UTF-8 filenames.
+ *   Specifically, BIO_new_file() will convert UTF-8 to UTF-16 and attempt
+ *   to use _wfopen() to open the file, but BIO_read_filename() will not.
+ *   It is BIO_read_filename() which the SSL_CTX_*_file functions use, and
+ *   thus they don't work with UTF-8 file names. This is filed as RT#3479:
+ *   http://rt.openssl.org/Ticket/Display.html?id=3479
+ *
+ * - Finally, and least importantly, it does actually matter which supporting
+ *   certs we offer on the wire because of RT#1942. Doing this for ourselves
+ *   allows us to explicitly print the supporting certs that we're using,
+ *   which may assist in diagnosing problems.
+ */
+static int load_cert_chain_file(struct openconnect_info *vpninfo)
 {
-	BIO *b = BIO_new(BIO_s_file_internal());
+	BIO *b;
+	FILE *f = openconnect_fopen_utf8(vpninfo, vpninfo->cert, "rb");
+	STACK_OF(X509) *extra_certs = NULL;
 	char buf[200];
 
-	if (!b)
-		return -ENOMEM;
-
-	if (BIO_read_filename(b, vpninfo->cert) <= 0) {
-	err:
-		BIO_free(b);
+	if (!f) {
 		vpn_progress(vpninfo, PRG_ERR,
-			     _("Failed to reload X509 cert for expiry check\n"));
+			     _("Failed to open certificate file %s: %s\n"),
+			     vpninfo->cert, strerror(errno));
+		return -ENOENT;
+	}
+
+	b = BIO_new_fp(f, 1);
+	if (!b) {
+		fclose(f);
+	err:
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Loading certificate failed\n"));
 		openconnect_report_ssl_errors(vpninfo);
 		return -EIO;
 	}
 	vpninfo->cert_x509 = PEM_read_bio_X509_AUX(b, NULL, NULL, NULL);
-	BIO_free(b);
-	if (!vpninfo->cert_x509)
+	if (!vpninfo->cert_x509) {
+		BIO_free(b);
 		goto err;
+	}
 
 	X509_NAME_oneline(X509_get_subject_name(vpninfo->cert_x509), buf, sizeof(buf));
 	vpn_progress(vpninfo, PRG_INFO,
 			     _("Using client certificate '%s'\n"), buf);
 
+	if (!SSL_CTX_use_certificate(vpninfo->https_ctx, vpninfo->cert_x509)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to install certificate in OpenSSL context\n"));
+		openconnect_report_ssl_errors(vpninfo);
+		BIO_free(b);
+		return -EIO;
+	}
+
+	while (1) {
+		X509 *x = PEM_read_bio_X509(b, NULL, NULL, NULL);
+		if (!x) {
+			unsigned long err = ERR_peek_last_error();
+			if (ERR_GET_LIB(err) == ERR_LIB_PEM &&
+			    ERR_GET_REASON(err) == PEM_R_NO_START_LINE)
+				ERR_clear_error();
+			else
+				goto err_extra;
+			break;
+		}
+		if (!extra_certs)
+			extra_certs = sk_X509_new_null();
+		if (!extra_certs) {
+		err_extra:
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to process all supporting certs. Trying anyway...\n"));
+			openconnect_report_ssl_errors(vpninfo);
+			X509_free(x);
+			/* It might work without... */
+			break;
+		}
+		if (!sk_X509_push(extra_certs, x))
+			goto err_extra;
+	}
+
+	BIO_free(b);
+
+	if (extra_certs)
+		install_extra_certs(vpninfo, _("PEM file"), extra_certs);
+
 	return 0;
 }
 
@@ -653,6 +731,9 @@ static int is_pem_password_error(struct openconnect_info *vpninfo)
 
 static int load_certificate(struct openconnect_info *vpninfo)
 {
+	FILE *f;
+	char buf[256];
+
 	if (!strncmp(vpninfo->sslkey, "pkcs11:", 7) ||
 	    !strncmp(vpninfo->cert, "pkcs11:", 7)) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -663,13 +744,10 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	vpn_progress(vpninfo, PRG_DEBUG,
 		     _("Using certificate file %s\n"), vpninfo->cert);
 
-	if (strncmp(vpninfo->cert, "keystore:", 9) &&
-	    (vpninfo->cert_type == CERT_TYPE_PKCS12 ||
-	     vpninfo->cert_type == CERT_TYPE_UNKNOWN)) {
-		FILE *f;
+	if (strncmp(vpninfo->cert, "keystore:", 9)) {
 		PKCS12 *p12;
 
-		f = fopen(vpninfo->cert, "rb");
+		f = openconnect_fopen_utf8(vpninfo, vpninfo->cert, "rb");
 		if (!f) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to open certificate file %s: %s\n"),
@@ -681,13 +759,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		if (p12)
 			return load_pkcs12_certificate(vpninfo, p12);
 
-		/* Not PKCS#12 */
-		if (vpninfo->cert_type == CERT_TYPE_PKCS12) {
-			vpn_progress(vpninfo, PRG_ERR, _("Read PKCS#12 failed\n"));
-			openconnect_report_ssl_errors(vpninfo);
-			return -EINVAL;
-		}
-		/* Clear error and fall through to see if it's a PEM file... */
+		/* Not PKCS#12. Clear error and fall through to see if it's a PEM file... */
 		ERR_clear_error();
 	}
 
@@ -716,18 +788,10 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	} else
 #endif /* ANDROID_KEYSTORE */
 	{
-		if (!SSL_CTX_use_certificate_chain_file(vpninfo->https_ctx,
-							vpninfo->cert)) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Loading certificate failed\n"));
-			openconnect_report_ssl_errors(vpninfo);
-			return -EINVAL;
-		}
-
-		/* Ew, we can't get it back from the OpenSSL CTX in any sane fashion */
-		reload_pem_cert(vpninfo);
+		int ret = load_cert_chain_file(vpninfo);
+		if (ret)
+			return ret;
 	}
-
 #ifdef ANDROID_KEYSTORE
 	if (!strncmp(vpninfo->sslkey, "keystore:", 9)) {
 		EVP_PKEY *key;
@@ -756,57 +820,57 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	}
 #endif /* ANDROID_KEYSTORE */
 
-	if (vpninfo->cert_type == CERT_TYPE_UNKNOWN) {
-		FILE *f = fopen(vpninfo->sslkey, "rb");
-		char buf[256];
+	f = openconnect_fopen_utf8(vpninfo, vpninfo->sslkey, "rb");
+	if (!f) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to open private key file %s: %s\n"),
+			     vpninfo->cert, strerror(errno));
+		return -ENOENT;
+	}
 
-		if (!f) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Failed to open private key file %s: %s\n"),
-				     vpninfo->cert, strerror(errno));
-			return -ENOENT;
-		}
+	buf[255] = 0;
+	while (fgets(buf, 255, f)) {
+		if (!strcmp(buf, "-----BEGIN TSS KEY BLOB-----\n")) {
+			fclose(f);
+			return load_tpm_certificate(vpninfo);
+		} else if (!strcmp(buf, "-----BEGIN RSA PRIVATE KEY-----\n") ||
+			   !strcmp(buf, "-----BEGIN DSA PRIVATE KEY-----\n") ||
+			   !strcmp(buf, "-----BEGIN ENCRYPTED PRIVATE KEY-----\n") ||
+			   !strcmp(buf, "-----BEGIN PRIVATE KEY-----\n")) {
+			RSA *key;
+			BIO *b = BIO_new_fp(f, BIO_CLOSE);
 
-		buf[255] = 0;
-		while (fgets(buf, 255, f)) {
-			if (!strcmp(buf, "-----BEGIN TSS KEY BLOB-----\n")) {
-				vpninfo->cert_type = CERT_TYPE_TPM;
-				break;
-			} else if (!strcmp(buf, "-----BEGIN RSA PRIVATE KEY-----\n") ||
-				   !strcmp(buf, "-----BEGIN DSA PRIVATE KEY-----\n") ||
-				   !strcmp(buf, "-----BEGIN ENCRYPTED PRIVATE KEY-----\n") ||
-				   !strcmp(buf, "-----BEGIN PRIVATE KEY-----\n")) {
-				vpninfo->cert_type = CERT_TYPE_PEM;
-				break;
+			if (!b) {
+				fclose(f);
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Loading private key failed\n"));
+				openconnect_report_ssl_errors(vpninfo);
 			}
-		}
-		fclose(f);
-		if (vpninfo->cert_type == CERT_TYPE_UNKNOWN) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Failed to identify private key type in '%s'\n"),
-				     vpninfo->sslkey);
-			return -EINVAL;
+		again:
+			fseek(f, 0, SEEK_SET);
+			key = PEM_read_bio_RSAPrivateKey(b, NULL, pem_pw_cb, vpninfo);
+			if (!key) {
+				if (is_pem_password_error(vpninfo))
+					goto again;
+				BIO_free(b);
+				return -EINVAL;
+			}
+			SSL_CTX_use_RSAPrivateKey(vpninfo->https_ctx, key);
+			RSA_free(key);
+			BIO_free(b);
+			return 0;
 		}
 	}
+	fclose(f);
 
-	if (vpninfo->cert_type == CERT_TYPE_TPM)
-		return load_tpm_certificate(vpninfo);
-
-	/* Standard PEM certificate */
-	SSL_CTX_set_default_passwd_cb(vpninfo->https_ctx, pem_pw_cb);
-	SSL_CTX_set_default_passwd_cb_userdata(vpninfo->https_ctx, vpninfo);
- again:
-	if (!SSL_CTX_use_RSAPrivateKey_file(vpninfo->https_ctx, vpninfo->sslkey,
-					    SSL_FILETYPE_PEM)) {
-		if (is_pem_password_error(vpninfo))
-			goto again;
-		return -EINVAL;
-	}
-	return 0;
+	vpn_progress(vpninfo, PRG_ERR,
+		     _("Failed to identify private key type in '%s'\n"),
+		     vpninfo->sslkey);
+	return -EINVAL;
 }
 
 static int get_cert_fingerprint(struct openconnect_info *vpninfo,
-				OPENCONNECT_X509 *cert, const EVP_MD *type,
+				X509 *cert, const EVP_MD *type,
 				char *buf)
 {
 	unsigned char md[EVP_MAX_MD_SIZE];
@@ -822,31 +886,44 @@ static int get_cert_fingerprint(struct openconnect_info *vpninfo,
 }
 
 int get_cert_md5_fingerprint(struct openconnect_info *vpninfo,
-			     OPENCONNECT_X509 *cert, char *buf)
+			     void *cert, char *buf)
 {
 	return get_cert_fingerprint(vpninfo, cert, EVP_md5(), buf);
 }
 
-int openconnect_get_cert_sha1(struct openconnect_info *vpninfo,
-			      OPENCONNECT_X509 *cert, char *buf)
+static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 {
-	return get_cert_fingerprint(vpninfo, cert, EVP_sha1(), buf);
-}
+	unsigned char sha1[SHA1_SIZE];
+	EVP_PKEY *pkey;
+	BIO *bp = BIO_new(BIO_s_mem());
+	BUF_MEM *keyinfo;
+	int i;
 
-static int check_server_cert(struct openconnect_info *vpninfo, X509 *cert)
-{
-	char fingerprint[EVP_MAX_MD_SIZE * 2 + 1];
-	int ret;
+	/* We can't use X509_pubkey_digest() because it only hashes the
+	   subjectPublicKey BIT STRING, and not the whole of the
+	   SubjectPublicKeyInfo SEQUENCE. */
+	pkey = X509_get_pubkey(vpninfo->peer_cert);
 
-	ret = openconnect_get_cert_sha1(vpninfo, cert, fingerprint);
-	if (ret)
-		return ret;
-
-	if (strcasecmp(vpninfo->servercert, fingerprint)) {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Server SSL certificate didn't match: %s\n"), fingerprint);
-		return -EINVAL;
+	if (!i2d_PUBKEY_bio(bp, pkey)) {
+		EVP_PKEY_free(pkey);
+		BIO_free(bp);
+		return -ENOMEM;
 	}
+	EVP_PKEY_free(pkey);
+
+	BIO_get_mem_ptr(bp, &keyinfo);
+
+	openconnect_sha1(sha1, keyinfo->data, keyinfo->length);
+
+	BIO_free(bp);
+
+	vpninfo->peer_cert_hash = malloc(SHA1_SIZE * 2 + 6);
+	if (vpninfo->peer_cert_hash) {
+		snprintf(vpninfo->peer_cert_hash, 6, "sha1:");
+		for (i = 0; i < sizeof(sha1); i++)
+			sprintf(&vpninfo->peer_cert_hash[i*2 + 5], "%02x", sha1[i]);
+	}
+
 	return 0;
 }
 
@@ -1126,40 +1203,28 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 
 static int verify_peer(struct openconnect_info *vpninfo, SSL *https_ssl)
 {
-	X509 *peer_cert;
 	int ret;
+	int vfy = SSL_get_verify_result(https_ssl);
+	const char *err_string = NULL;
 
-	peer_cert = SSL_get_peer_certificate(https_ssl);
+	if (vfy != X509_V_OK)
+		err_string = X509_verify_cert_error_string(vfy);
+	else if (match_cert_hostname(vpninfo, vpninfo->peer_cert))
+		err_string = _("certificate does not match hostname");
 
-	if (vpninfo->servercert) {
-		/* If given a cert fingerprint on the command line, that's
-		   all we look for */
-		ret = check_server_cert(vpninfo, peer_cert);
+	if (err_string) {
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("Server certificate verify failed: %s\n"),
+			     err_string);
+
+		if (vpninfo->validate_peer_cert)
+			ret = vpninfo->validate_peer_cert(vpninfo->cbdata,
+							  err_string);
+		else
+			ret = -EINVAL;
 	} else {
-		int vfy = SSL_get_verify_result(https_ssl);
-		const char *err_string = NULL;
-
-		if (vfy != X509_V_OK)
-			err_string = X509_verify_cert_error_string(vfy);
-		else if (match_cert_hostname(vpninfo, peer_cert))
-			err_string = _("certificate does not match hostname");
-
-		if (err_string) {
-			vpn_progress(vpninfo, PRG_INFO,
-				     _("Server certificate verify failed: %s\n"),
-				     err_string);
-
-			if (vpninfo->validate_peer_cert)
-				ret = vpninfo->validate_peer_cert(vpninfo->cbdata,
-								  peer_cert,
-								  err_string);
-			else
-				ret = -EINVAL;
-		} else {
-			ret = 0;
-		}
+		ret = 0;
 	}
-	X509_free(peer_cert);
 
 	return ret;
 }
@@ -1255,6 +1320,7 @@ static int check_certificate_expiry(struct openconnect_info *vpninfo)
 	}
 	return 0;
 }
+
 int openconnect_open_https(struct openconnect_info *vpninfo)
 {
 	method_const SSL_METHOD *ssl3_method;
@@ -1270,6 +1336,9 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		X509_free(vpninfo->peer_cert);
 		vpninfo->peer_cert = NULL;
 	}
+	free (vpninfo->peer_cert_hash);
+	vpninfo->peer_cert_hash = NULL;
+	vpninfo->cstp_cipher = NULL;
 
 	ssl_sock = connect_https_socket(vpninfo);
 	if (ssl_sock < 0)
@@ -1310,8 +1379,9 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		SSL_CTX_set_cert_verify_callback(vpninfo->https_ctx,
 						 ssl_app_verify_callback, NULL);
 #endif
-		SSL_CTX_set_default_verify_paths(vpninfo->https_ctx);
-		
+		if (!vpninfo->no_system_trust)
+			SSL_CTX_set_default_verify_paths(vpninfo->https_ctx);
+
 		if (vpninfo->pfs)
 			SSL_CTX_set_cipher_list(vpninfo->https_ctx, "HIGH:!aNULL:!eNULL:-RSA");
 
@@ -1356,7 +1426,16 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		} else
 #endif
 		if (vpninfo->cafile) {
-			if (!SSL_CTX_load_verify_locations(vpninfo->https_ctx, vpninfo->cafile, NULL)) {
+			/* OpenSSL does actually manage to cope with UTF-8 for
+			   this one, under Windows. So only convert for legacy
+			   UNIX. */
+			char *cafile = openconnect_utf8_to_legacy(vpninfo,
+								  vpninfo->cafile);
+			err = SSL_CTX_load_verify_locations(vpninfo->https_ctx,
+							    cafile, NULL);
+			if (cafile != vpninfo->cafile)
+				free(cafile);
+			if (!err) {
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Failed to open CA file '%s'\n"),
 					     vpninfo->cafile);
@@ -1417,6 +1496,10 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		}
 	}
 
+	vpninfo->cstp_cipher = (char *)SSL_get_cipher_name(https_ssl);
+	vpninfo->peer_cert = SSL_get_peer_certificate(https_ssl);
+	set_peer_cert_hash(vpninfo);
+
 	if (verify_peer(vpninfo, https_ssl)) {
 		SSL_free(https_ssl);
 		closesocket(ssl_sock);
@@ -1430,9 +1513,6 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	vpninfo->ssl_write = openconnect_openssl_write;
 	vpninfo->ssl_gets = openconnect_openssl_gets;
 
-	/* Stash this now, because it might not be available later if the
-	   server has disconnected. */
-	vpninfo->peer_cert = SSL_get_peer_certificate(vpninfo->https_ssl);
 
 	vpn_progress(vpninfo, PRG_INFO, _("Connected to HTTPS on %s\n"),
 		     vpninfo->hostname);
@@ -1447,10 +1527,6 @@ int cstp_handshake(struct openconnect_info *vpninfo, unsigned init)
 
 void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 {
-	if (vpninfo->peer_cert) {
-		X509_free(vpninfo->peer_cert);
-		vpninfo->peer_cert = NULL;
-	}
 	if (vpninfo->https_ssl) {
 		SSL_free(vpninfo->https_ssl);
 		vpninfo->https_ssl = NULL;
@@ -1474,26 +1550,28 @@ void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 	}
 }
 
-void openconnect_init_ssl(void)
+int openconnect_init_ssl(void)
 {
 #ifdef _WIN32
-	openconnect__win32_sock_init();
+	int ret = openconnect__win32_sock_init();
+	if (ret)
+		return ret;
 #endif
 	SSL_library_init();
 	ERR_clear_error();
 	SSL_load_error_strings();
 	OpenSSL_add_all_algorithms();
+	return 0;
 }
 
-char *openconnect_get_cert_details(struct openconnect_info *vpninfo,
-				   OPENCONNECT_X509 *cert)
+char *openconnect_get_peer_cert_details(struct openconnect_info *vpninfo)
 {
 	BIO *bp = BIO_new(BIO_s_mem());
 	BUF_MEM *certinfo;
 	char zero = 0;
 	char *ret;
 
-	X509_print_ex(bp, cert, 0, 0);
+	X509_print_ex(bp, vpninfo->peer_cert, 0, 0);
 	BIO_write(bp, &zero, 1);
 	BIO_get_mem_ptr(bp, &certinfo);
 
@@ -1502,6 +1580,11 @@ char *openconnect_get_cert_details(struct openconnect_info *vpninfo,
 	return ret;
 }
 
+void openconnect_free_cert_info(struct openconnect_info *vpninfo,
+				void *buf)
+{
+	free(buf);
+}
 
 int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
 			       char *buf)
@@ -1516,3 +1599,26 @@ int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
 
 	return 0;
 }
+
+#ifdef HAVE_LIBPCSCLITE
+int openconnect_hash_yubikey_password(struct openconnect_info *vpninfo,
+				      const char *password, const void *ident, int id_len)
+{
+	if (!PKCS5_PBKDF2_HMAC_SHA1(password, strlen(password), ident, id_len, 1000, 16,
+				    vpninfo->yubikey_pwhash))
+		return -EIO;
+
+	return 0;
+}
+
+int openconnect_yubikey_chalresp(struct openconnect_info *vpninfo,
+				  const void *challenge, int chall_len, void *result)
+{
+	unsigned int mdlen = SHA1_SIZE;
+
+	if (!HMAC(EVP_sha1(), vpninfo->yubikey_pwhash, 16, challenge, chall_len, result, &mdlen))
+		return -EIO;
+
+	return 0;
+}
+#endif
