@@ -106,6 +106,23 @@ unsigned string_is_hostname(const char *str)
 	return 1;
 }
 
+static int match_sockaddr(struct sockaddr *a, struct sockaddr *b)
+{
+	if (a->sa_family == AF_INET) {
+		struct sockaddr_in *a4 = (void *)a;
+		struct sockaddr_in *b4 = (void *)b;
+
+		return (a4->sin_addr.s_addr == b4->sin_addr.s_addr) &&
+			(a4->sin_port == b4->sin_port);
+	} else if (a->sa_family == AF_INET6) {
+		struct sockaddr_in6 *a6 = (void *)a;
+		struct sockaddr_in6 *b6 = (void *)b;
+		return !memcmp(&a6->sin6_addr, &b6->sin6_addr, sizeof(a6->sin6_addr)) &&
+		       a6->sin6_port == b6->sin6_port;
+	} else
+		return 0;
+}
+
 int connect_https_socket(struct openconnect_info *vpninfo)
 {
 	int ssl_sock = -1;
@@ -114,7 +131,11 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 	if (!vpninfo->port)
 		vpninfo->port = 443;
 
-	if (vpninfo->peer_addr) {
+	/* If we're talking to a server which told us it has dynamic DNS, don't
+	   just re-use its previous IP address. If we're talking to a proxy, we
+	   can use *its* previous IP address. We expect it'll re-do the DNS
+	   lookup for the server anyway. */
+	if (vpninfo->peer_addr && (!vpninfo->is_dyndns || vpninfo->proxy)) {
 	reconnect:
 #ifdef SOCK_CLOEXEC
 		ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM | SOCK_CLOEXEC, IPPROTO_IP);
@@ -139,7 +160,8 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			}
 			if (ssl_sock >= 0)
 				closesocket(ssl_sock);
-			return -EINVAL;
+			ssl_sock = -EINVAL;
+			goto out;
 		}
 	} else {
 		struct addrinfo hints, *result, *rp;
@@ -176,8 +198,10 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			else
 				i = asprintf(&url, "https://%s:%d/%s", vpninfo->hostname,
 					     vpninfo->port, vpninfo->urlpath?:"");
-			if (i == -1)
-				return -ENOMEM;
+			if (i == -1) {
+				ssl_sock = -ENOMEM;
+				goto out;
+			}
 
 			proxies = px_proxy_factory_get_proxies(vpninfo->proxy_factory,
 							       url);
@@ -211,8 +235,10 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 
 		if (hostname[0] == '[' && hostname[strlen(hostname)-1] == ']') {
 			hostname = strndup(hostname + 1, strlen(hostname) - 2);
-			if (!hostname)
-				return -ENOMEM;
+			if (!hostname) {
+				ssl_sock = -ENOMEM;
+				goto out;
+			}
 			hints.ai_flags |= AI_NUMERICHOST;
 		}
 
@@ -224,7 +250,15 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 				     hostname, gai_strerror(err));
 			if (hints.ai_flags & AI_NUMERICHOST)
 				free(hostname);
-			return -EINVAL;
+			ssl_sock = -EINVAL;
+			/* If we were just retrying for dynamic DNS, reconnct using
+			   the previously-known IP address */
+			if (vpninfo->peer_addr) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Reconnecting to DynDNS server using previously cached IP address\n"));
+				goto reconnect;
+			}
+			goto out;
 		}
 		if (hints.ai_flags & AI_NUMERICHOST)
 			free(hostname);
@@ -251,12 +285,15 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			if (cancellable_connect(vpninfo, ssl_sock, rp->ai_addr, rp->ai_addrlen) >= 0) {
 				/* Store the peer address we actually used, so that DTLS can
 				   use it again later */
+				free(vpninfo->peer_addr);
+				vpninfo->peer_addrlen = 0;
 				vpninfo->peer_addr = malloc(rp->ai_addrlen);
 				if (!vpninfo->peer_addr) {
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Failed to allocate sockaddr storage\n"));
 					closesocket(ssl_sock);
-					return -ENOMEM;
+					ssl_sock = -ENOMEM;
+					goto out;
 				}
 				vpninfo->peer_addrlen = rp->ai_addrlen;
 				memcpy(vpninfo->peer_addr, rp->ai_addr, rp->ai_addrlen);
@@ -281,6 +318,17 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			}
 			closesocket(ssl_sock);
 			ssl_sock = -1;
+
+			/* If we're in DynDNS mode but this *was* the cached IP address,
+			 * don't bother falling back to it if it didn't work. */
+			if (vpninfo->peer_addr && vpninfo->peer_addrlen == rp->ai_addrlen &&
+			    match_sockaddr(vpninfo->peer_addr, rp->ai_addr)) {
+				vpn_progress(vpninfo, PRG_TRACE,
+					     _("Forgetting non-functional previous peer address\n"));
+				free(vpninfo->peer_addr);
+				vpninfo->peer_addr = 0;
+				vpninfo->peer_addrlen = 0;
+			}
 		}
 		freeaddrinfo(result);
 
@@ -288,7 +336,13 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to connect to host %s\n"),
 				     vpninfo->proxy?:vpninfo->hostname);
-			return -EINVAL;
+			ssl_sock = -EINVAL;
+			if (vpninfo->peer_addr) {
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Reconnecting to DynDNS server using previously cached IP address\n"));
+				goto reconnect;
+			}
+			goto out;
 		}
 	}
 
@@ -302,10 +356,13 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 					     _("Reconnecting to proxy %s\n"), vpninfo->proxy);
 				goto reconnect;
 			}
-			return err;
+			ssl_sock = err;
 		}
 	}
-
+ out:
+	/* If proxy processing returned -EAGAIN to reconnect before attempting
+	   further auth, and we failed to reconnect, we have to clean up here. */
+	cleanup_proxy_auth(vpninfo);
 	return ssl_sock;
 }
 
@@ -324,7 +381,8 @@ int  __attribute__ ((format (printf, 2, 3)))
 
 }
 
-int request_passphrase(struct openconnect_info *vpninfo, const char *label,
+int __attribute__ ((format(printf, 4, 5)))
+    request_passphrase(struct openconnect_info *vpninfo, const char *label,
 		       char **response, const char *fmt, ...)
 {
 	struct oc_auth_form f;
@@ -346,11 +404,11 @@ int request_passphrase(struct openconnect_info *vpninfo, const char *label,
 	o.type = OC_FORM_OPT_PASSWORD;
 	o.name = (char *)label;
 	o.label = buf;
-	o.value = NULL;
+	o._value = NULL;
 
 	ret = process_auth_form(vpninfo, &f);
 	if (!ret) {
-		*response = o.value;
+		*response = o._value;
 		return 0;
 	}
 
@@ -361,38 +419,63 @@ int request_passphrase(struct openconnect_info *vpninfo, const char *label,
 int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 {
 	struct statvfs buf;
+	char *sslkey = openconnect_utf8_to_legacy(vpninfo, vpninfo->sslkey);
+	int err = 0;
 
-	if (statvfs(vpninfo->sslkey, &buf)) {
-		int err = errno;
+	if (statvfs(sslkey, &buf)) {
+		err = -errno;
 		vpn_progress(vpninfo, PRG_ERR, _("statvfs: %s\n"),
 			     strerror(errno));
-		return -err;
-	}
-	if (asprintf(&vpninfo->cert_password, "%lx", buf.f_fsid))
-		return -ENOMEM;
-	return 0;
+	} else if (asprintf(&vpninfo->cert_password, "%lx", buf.f_fsid) == -1)
+		err = -ENOMEM;
+
+	if (sslkey != vpninfo->sslkey)
+		free(sslkey);
+	return err;
 }
 #elif defined(_WIN32)
 #include <fileapi.h>
+typedef BOOL WINAPI (*GVIBH)(HANDLE, LPWSTR, DWORD, LPDWORD, LPDWORD, LPDWORD, LPWSTR, DWORD);
+
 int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 {
 	HANDLE h;
 	DWORD serial;
+	HINSTANCE kernlib;
+	GVIBH func = NULL;
 	int success;
+	int fd;
 
-	h = CreateFile(vpninfo->sslkey, 0, 0, NULL, OPEN_EXISTING,
-		       FILE_ATTRIBUTE_NORMAL, NULL);
-	if (h == INVALID_HANDLE_VALUE)
-		return -EIO;
+	/* Some versions of Windows don't have this so don't use standard
+	   load-time linking or it'll cause failures. */
+	kernlib = LoadLibraryA("Kernel32.dll");
+	if (!kernlib) {
+	notsupp:
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Could not obtain file system ID for passphrase\n"));
+		return -EOPNOTSUPP;
+	}
+	func = (GVIBH)GetProcAddress(kernlib, "GetVolumeInformationByHandleW");
+	FreeLibrary(kernlib);
+	if (!func)
+		goto notsupp;
 
-	success = GetVolumeInformationByHandleW(h, NULL, 0, &serial, NULL,
-						NULL, NULL, 0);
-	CloseHandle(h);
+	fd = openconnect_open_utf8(vpninfo, vpninfo->sslkey, O_RDONLY);
+	if (fd == -1) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to open private key file '%s': %s\n"),
+			     vpninfo->sslkey, strerror(errno));
+		return -ENOENT;
+	}
+
+	h = (HANDLE)_get_osfhandle(fd);
+	success = func(h, NULL, 0, &serial, NULL, NULL, NULL, 0);
+	close(fd);
 
 	if (!success)
 		return -EIO;
 
-	if (asprintf(&vpninfo->cert_password, "%lx", serial))
+	if (asprintf(&vpninfo->cert_password, "%lx", serial) == -1)
 		return -ENOMEM;
 
 	return 0;
@@ -400,21 +483,28 @@ int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 #else
 int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 {
+	char *sslkey = openconnect_utf8_to_legacy(vpninfo, vpninfo->sslkey);
 	struct statfs buf;
 	unsigned *fsid = (unsigned *)&buf.f_fsid;
 	unsigned long long fsid64;
+	int err = 0;
 
-	if (statfs(vpninfo->sslkey, &buf)) {
-		int err = errno;
+	if (statfs(sslkey, &buf)) {
+		err = -errno;
 		vpn_progress(vpninfo, PRG_ERR, _("statfs: %s\n"),
 			     strerror(errno));
 		return -err;
-	}
-	fsid64 = ((unsigned long long)fsid[0] << 32) | fsid[1];
+	} else {
+		fsid64 = ((unsigned long long)fsid[0] << 32) | fsid[1];
 
-	if (asprintf(&vpninfo->cert_password, "%llx", fsid64))
-		return -ENOMEM;
-	return 0;
+		if (asprintf(&vpninfo->cert_password, "%llx", fsid64) == -1)
+			err = -ENOMEM;
+	}
+
+	if (sslkey != vpninfo->sslkey)
+		free(sslkey);
+
+	return err;
 }
 #endif
 
@@ -585,9 +675,13 @@ void check_cmd_fd(struct openconnect_info *vpninfo, fd_set *fds)
 		return;
 	}
 
+#ifdef _WIN32
+	if (recv(vpninfo->cmd_fd, &cmd, 1, 0) != 1)
+		return;
+#else
 	if (read(vpninfo->cmd_fd, &cmd, 1) != 1)
 		return;
-
+#endif
 	switch (cmd) {
 	case OC_CMD_CANCEL:
 	case OC_CMD_DETACH:
@@ -628,3 +722,72 @@ void poll_cmd_fd(struct openconnect_info *vpninfo, int timeout)
 		check_cmd_fd(vpninfo, &rd_set);
 	}
 }
+
+#ifdef _WIN32
+#include <io.h>
+#include <sys/stat.h>
+int openconnect_open_utf8(struct openconnect_info *vpninfo, const char *fname, int mode)
+{
+	wchar_t *fname_w;
+	int nr_chars = MultiByteToWideChar(CP_UTF8, 0, fname, -1, NULL, 0);
+	int fd;
+
+	if (!nr_chars) {
+		errno = EINVAL;
+		return -1;
+	}
+	fname_w = malloc(nr_chars * sizeof(wchar_t));
+	if (!fname_w) {
+		errno = ENOMEM;
+		return -1;
+	}
+	MultiByteToWideChar(CP_UTF8, 0, fname, -1, fname_w, nr_chars);
+
+	fd = _wopen(fname_w, mode, _S_IREAD | _S_IWRITE);
+	free(fname_w);
+
+	return fd;
+}
+#else
+int openconnect_open_utf8(struct openconnect_info *vpninfo, const char *fname, int mode)
+{
+	char *legacy_fname = openconnect_utf8_to_legacy(vpninfo, fname);
+	int fd;
+
+	fd = open(legacy_fname, mode, 0644);
+	if (legacy_fname != fname)
+		free(legacy_fname);
+
+	return fd;
+}
+#endif
+
+FILE *openconnect_fopen_utf8(struct openconnect_info *vpninfo, const char *fname,
+			     const char *mode)
+{
+	int fd;
+	int flags;
+
+	if (!strcmp(mode, "r"))
+		flags = O_RDONLY|O_CLOEXEC;
+	else if (!strcmp(mode, "rb"))
+		flags = O_RDONLY|O_CLOEXEC|O_BINARY;
+	else if (!strcmp(mode, "w"))
+		flags = O_WRONLY|O_CLOEXEC|O_CREAT|O_TRUNC;
+	else if (!strcmp(mode, "wb"))
+		flags = O_WRONLY|O_CLOEXEC|O_CREAT|O_TRUNC|O_BINARY;
+	else {
+		/* This should never happen, but if we forget and start using other
+		   modes without implementing proper mode->flags conversion, complain! */
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("openconnect_fopen_utf8() used with unsupported mode '%s'\n"),
+			     mode);
+		return NULL;
+	}
+
+	fd = openconnect_open_utf8(vpninfo, fname, flags);
+	if (fd == -1)
+		return NULL;
+
+	return fdopen(fd, mode);
+}
